/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005   Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*----------------------------------------------------------\
|                      EngineClass.cpp                      |
|    Contient toutes les classes nécessaires au moteur 3D   |
| et au moteur physique.                                    |
|                                                           |
\----------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"					// Pour la lecture des fichiers 3D
#include "cob.h"					// Pour la lecture et l'éxecution des scripts
#include "tdf.h"					// Pour la gestion des éléments du jeu
#include "EngineClass.h"

float	player_color[30]={	0.11f,	0.28f,	0.91f,
							0.83f,	0.17f,	0.0f,
							1.0f,	1.0f,	1.0f,
							0.11f,	0.62f,	0.07f,
							0.03f,	0.12f,	0.48f,
							0.5f,	0.34f,	0.62f,
							1.0f,	1.0f,	0.0f,
							0.0f,	0.0f,	0.0f,
							0.61f,	0.8f,	0.87f,
							0.67f,	0.67f,	0.51f };
INGAME_UNITS units;
PLAYERS	players;		// Objet contenant les données sur les joueurs

int NB_PLAYERS;

void MAP_OTA::load(char *data,int ota_size)
{
	destroy();
	set_uformat(U_ASCII);
	char *pos=data;
	char *ligne=NULL;
	int nb=0;
	int index=0;
	char *limit=data+ota_size;
	char *f;
	int n_pos=0;
	do {
		nb++;
		if(ligne)
			delete[] ligne;
		ligne=get_line(pos);
		strlwr(ligne);
		while(pos[0]!=0 && pos[0]!=13 && pos[0]!=10)	{	pos++;	n_pos++;	}
		while(pos[0]==13 || pos[0]==10)	{	pos++;	n_pos++;	}

		if(strstr(ligne,"//") || strstr(ligne,"/*") || strstr(ligne,"{")) { }		// Saute les commentaires
		else if(f=strstr(ligne,"missionname=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			missionname=strdup(f+12);
			}
		else if(f=strstr(ligne,"planet=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			planet=strdup(f+7);
			}
		else if(f=strstr(ligne,"missiondescription=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			missiondescription=strdup(f+19);
			char *prec=missiondescription;
			char *cur=prec;
			int nb=0;
			while(cur[0]) {
				if(cur[0]==' ')	prec=cur;
				nb++;
				if(nb>34) {
					prec[0]='\n';
					nb=cur-prec;
					}
				cur++;
				}
			}
		else if(f=strstr(ligne,"tidalstrength=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			tidalstrength=atoi(f+14);
			}
		else if(f=strstr(ligne,"solarstrength=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			solarstrength=atoi(f+14);
			}
		else if(f=strstr(ligne,"lavaworld=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			lavaworld=(f[10]=='1');
			}
		else if(f=strstr(ligne,"killmul=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			killmul=atoi(f+8);
			}
		else if(f=strstr(ligne,"minwindspeed=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			minwindspeed=atoi(f+13);
			}
		else if(f=strstr(ligne,"maxwindspeed=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			maxwindspeed=atoi(f+13);
			}
		else if(f=strstr(ligne,"gravity=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			gravity=atoi(f+8)*0.1f;
			}
		else if(f=strstr(ligne,"numplayers=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			numplayers=strdup(f+11);
			}
		else if(f=strstr(ligne,"size=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			map_size=strdup(f+5);
			}
		else if(f=strstr(ligne,"surfacemetal=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			SurfaceMetal=atoi(f+13);
			}
		else if(f=strstr(ligne,"mohometal=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			MohoMetal=atoi(f+10);
			}
		else if(f=strstr(ligne,"specialwhat=startpos")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			index=atoi(f+20)-1;
			}
		else if(f=strstr(ligne,"xpos=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			startX[index]=atoi(f+5);
			}
		else if(f=strstr(ligne,"zpos=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			startZ[index]=atoi(f+5);
			}
		else if(f=strstr(ligne,"waterdoesdamage=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			waterdoesdamage=atoi(f+16)>0;
			}
		else if(f=strstr(ligne,"waterdamage=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			waterdamage=atoi(f+12);
			}
		else if(f=strstr(ligne,"missionhint=")) {}
		else if(f=strstr(ligne,"brief=")) {}
		else if(f=strstr(ligne,"narration=")) {}
		else if(f=strstr(ligne,"glamour=")) {}
		else if(f=strstr(ligne,"lineofsight=")) {}
		else if(f=strstr(ligne,"mapping=")) {}
		else if(f=strstr(ligne,"timemul=")) {}
		else if(f=strstr(ligne,"memory=")) {}
		else if(f=strstr(ligne,"useonlyunits=")) {}
		else if(f=strstr(ligne,"schemacount=")) {}
		else if(f=strstr(ligne,"type=")) {}
		else if(f=strstr(ligne,"aiprofile=")) {}
		else if(f=strstr(ligne,"humanmetal=")) {}
		else if(f=strstr(ligne,"computermetal=")) {}
		else if(f=strstr(ligne,"humanenergy=")) {}
		else if(f=strstr(ligne,"computerenergy=")) {}
		else if(f=strstr(ligne,"meteorweapon=")) {}
		else if(f=strstr(ligne,"meteorradius=")) {}
		else if(f=strstr(ligne,"meteordensity=")) {}
		else if(f=strstr(ligne,"meteorduration=")) {}
		else if(f=strstr(ligne,"meteorinterval=")) {}
		else if(f=strstr(ligne,"killenemycommander=")) {}
		else if(f=strstr(ligne,"destroyallunits=")) {}
		else if(f=strstr(ligne,"allunitskilled=")) {}
		else if(f=strstr(ligne,"featurename=")) {}
		else {
			if(strlen(ligne)>1 && strstr(ligne,"{")==NULL && strstr(ligne,"}")==NULL && strstr(ligne,"[")==NULL)
				Console->AddEntry("(OTA) inconnu: %s",ligne);
			}

	} while(nb<1000 && pos<limit);
	delete[] ligne;
	ligne=NULL;
	if(waterdamage==0)
		waterdoesdamage=false;
}

void MAP::draw_mini(int x1,int y1,int w,int h,CAMERA *cam)			// Dessine la mini-carte
{
	if(!mini) return;		// Vérifie si la mini-carte existe
	int rw=w*mini_w/252;
	int rh=h*mini_h/252;
	x1+=w-rw>>1;
	y1+=h-rh>>1;
	float lw=mini_w/252.0f;
	float lh=mini_h/252.0f;
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,glmini);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f);	glVertex2f(x1,y1);
		glTexCoord2f(lw,0.0f);		glVertex2f(x1+rw,y1);
		glTexCoord2f(lw,lh);		glVertex2f(x1+rw,y1+rh);
		glTexCoord2f(0.0f,lh);		glVertex2f(x1,y1+rh);
	glEnd();

	if(!cam) return;

	VECTOR A,B,C,D,P;
	A=cam->Dir+1.0f*(-cam->Side-0.75f*cam->Up);
	B=cam->Dir+1.0f*(cam->Side-0.75f*cam->Up);
	C=cam->Dir+1.0f*(cam->Side+0.75f*cam->Up);
	D=cam->Dir+1.0f*(-cam->Side+0.75f*cam->Up);
	const int nmax=64;
	float cx[4*nmax+4],cy[4*nmax+4];
	if(A.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(A.y)*A;	cx[0]=P.x;	cy[0]=P.z; }
	else {
		P=cam->Pos+10000.0f*A;	cx[0]=P.x;	cy[0]=P.z; }
	if(B.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(B.y)*B;	cx[1]=P.x;	cy[1]=P.z; }
	else {
		P=cam->Pos+10000.0f*B;	cx[1]=P.x;	cy[1]=P.z; }
	if(C.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(C.y)*C;	cx[2]=P.x;	cy[2]=P.z; }
	else {
		P=cam->Pos+10000.0f*C;	cx[2]=P.x;	cy[2]=P.z; }
	if(D.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(D.y)*D;	cx[3]=P.x;	cy[3]=P.z; }
	else {
		P=cam->Pos+10000.0f*D;	cx[3]=P.x;	cy[3]=P.z; }

	int i;
	for(i=0;i<4;i++) {
		cx[i]=(cx[i]+0.5f*map_w)*rw/map_w;
		cy[i]=(cy[i]+0.5f*map_h)*rh/map_h;
		}
	for(i=0;i<4;i++) {
		for(int e=0;e<nmax;e++) {
			cx[i*nmax+e+4]=(cx[i]*(nmax-e)+cx[(i+1)%4]*(e+1))/(nmax+1);
			cy[i*nmax+e+4]=(cy[i]*(nmax-e)+cy[(i+1)%4]*(e+1))/(nmax+1);
			}
		}
	for(i=0;i<4+(nmax<<2);i++) {
		if(cx[i]<0.0f) cx[i]=0.0f;
		else if(cx[i]>rw) cx[i]=rw;
		if(cy[i]<0.0f) cy[i]=0.0f;
		else if(cy[i]>rh) cy[i]=rh;
		}

	glDisable(GL_TEXTURE_2D);
	glColor3f(1.0f,1.0f,1.0f);
	glBegin(GL_LINE_STRIP);
		for(i=0;i<4;i++) {
			glVertex2f(cx[i]+x1,cy[i]+y1);
			for(int e=0;e<nmax;e++)
				glVertex2f(x1+cx[i*nmax+e+4],y1+cy[i*nmax+e+4]);
			}
		glVertex2f(cx[0]+x1,cy[0]+y1);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void MAP::update_player_visibility(int player_id,int px,int py,int r,bool remove)
{
	byte mask=1<<player_id;
	int r2=r*r;
	if(player_id==players.local_human_id)
		for(int y=0;y<=r;y++) {
			int x=(int)(0.5f+sqrt((float)(r2-y*y)));
			int ry=py-y;
			if(ry>=0 && ry<bloc_h_db) {
				int rx=px-x;
				int lx=px+x;
				if(rx<0)	rx=0;
				if(lx>bloc_w_db)	lx=bloc_w<<1;
				if(remove)
					for(;rx<lx;rx++) {
						map_data[ry][rx].view|=mask;
						map_data[ry][rx].timer--;
						}
				else
					for(;rx<lx;rx++) {
						map_data[ry][rx].view|=mask;
						map_data[ry][rx].timer++;
						}
				}
			if(y!=0) {
				ry=py+y;
				if(ry>=0 && ry<bloc_h_db) {
					int rx=px-x;
					int lx=px+x;
					if(rx<0)	rx=0;
					if(lx>bloc_w_db)	lx=bloc_w<<1;
					if(remove)
						for(;rx<lx;rx++) {
							map_data[ry][rx].view|=mask;
							map_data[ry][rx].timer--;
							}
					else
						for(;rx<lx;rx++) {
							map_data[ry][rx].view|=mask;
							map_data[ry][rx].timer++;
							}
					}
				}
			}
	else
		for(int y=0;y<=r;y++) {
			int x=(int)(0.5f+sqrt((float)(r2-y*y)));
			int ry=py-y;
			if(ry>=0 && ry<bloc_h_db) {
				int rx=px-x;
				int lx=px+x;
				if(rx<0)	rx=0;
				if(lx>bloc_w_db)	lx=bloc_w<<1;
				for(;rx<lx;rx++)
					map_data[ry][rx].view|=mask;
				}
			if(y!=0) {
				ry=py+y;
				if(ry>=0 && ry<bloc_h_db) {
					int rx=px-x;
					int lx=px+x;
					if(rx<0)	rx=0;
					if(lx>bloc_w_db)	lx=bloc_w<<1;
					for(;rx<lx;rx++)
						map_data[ry][rx].view|=mask;
					}
				}
			}
}

	void MAP::draw(CAMERA *cam,byte player_mask,bool FLAT,float niv,float t,float dt,bool depth_only,bool check_visibility,bool draw_uw)
	{
		cam->SetView();
		if(FLAT && !water)	return;
		if(FLAT)
			glTranslatef(0.0f,0.0f,sea_dec);
		int i,x,y;
		glDisable(GL_CULL_FACE);
		glDisable(GL_LIGHTING);
		if(!FLAT) {
			if(ntex>0) {
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				for(i=0;i<8;i++) {
					glActiveTextureARB(GL_TEXTURE0_ARB+i);
					ReInitTexSys();
					glDisable(GL_TEXTURE_2D);
					}
				}
			else
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			}
		VECTOR A,B,C,D,P;
		A=cam->Dir+1.0f*(-cam->Side-0.75f*cam->Up);
		B=cam->Dir+1.0f*(cam->Side-0.75f*cam->Up);
		C=cam->Dir+1.0f*(cam->Side+0.75f*cam->Up);
		D=cam->Dir+1.0f*(-cam->Side+0.75f*cam->Up);
		float cx[4],cy[4];
		if(A.y<0.0f) {
			P=cam->Pos+cam->Pos.y/fabs(A.y)*A;	cx[0]=P.x;	cy[0]=P.z; }
		else {
			P=cam->Pos+10000.0f*A;	cx[0]=P.x;	cy[0]=P.z; }
		if(B.y<0.0f) {
			P=cam->Pos+cam->Pos.y/fabs(B.y)*B;	cx[1]=P.x;	cy[1]=P.z; }
		else {
			P=cam->Pos+10000.0f*B;	cx[1]=P.x;	cy[1]=P.z; }
		if(C.y<0.0f) {
			P=cam->Pos+cam->Pos.y/fabs(C.y)*C;	cx[2]=P.x;	cy[2]=P.z; }
		else {
			P=cam->Pos+10000.0f*C;	cx[2]=P.x;	cy[2]=P.z; }
		if(D.y<0.0f) {
			P=cam->Pos+cam->Pos.y/fabs(D.y)*D;	cx[3]=P.x;	cy[3]=P.z; }
		else {
			P=cam->Pos+10000.0f*D;	cx[3]=P.x;	cy[3]=P.z; }

		int minx=bloc_w<<1,maxx=0;
		int miny=bloc_h<<1,maxy=0;
		for(i=0;i<4;i++) {
			cx[i]=(cx[i]+0.5f*map_w)/16.0f;
			cy[i]=(cy[i]+0.5f*map_h)/16.0f;
			if(cx[i]<minx) minx=(int)cx[i];
			if(cx[i]>maxx) maxx=(int)cx[i];
			if(cy[i]<miny) miny=(int)cy[i];
			if(cy[i]>maxy) maxy=(int)cy[i];
			}
		int y1=bloc_h,y2=0,x1=bloc_w,x2=0,mx,my;
		float limit=cam->zfar*sqrt(2.0f);
		x1=(int)((cam->Pos.x+0.5f*map_w-limit)/16.0f);
		y1=(int)((cam->Pos.z+0.5f*map_h-limit)/16.0f);
		x2=(int)((cam->Pos.x+0.5f*map_w+limit)/16.0f);
		y2=(int)((cam->Pos.z+0.5f*map_h+limit)/16.0f);
		mx=(int)((cam->Pos.x+0.5f*map_w)/16.0f);
		my=(int)((cam->Pos.z+0.5f*map_h)/16.0f);
		if(x1<minx)
			x1=minx;
		if(x2>maxx)
			x2=maxx;
		if(y1<miny)
			y1=miny;
		if(y2>maxy)
			y2=maxy;
		if(x1>mx) x1=mx;
		if(y1>my) y1=my;
		if(x2<mx) x2=mx;
		if(y2<my) y2=my;

		if((abs(my-y2)<<4)>cam->zfar+64.0f)
			y2 = my>y2 ? my - (int)(cam->zfar/16.0f)-4 : my + 4 + (int)(cam->zfar/16.0f);
		if((abs(my-y1)<<4)>cam->zfar+64.0f)
			y1 = my>y1 ? my - (int)(cam->zfar/16.0f)-4 : my + 4 + (int)(cam->zfar/16.0f);
		if((abs(mx-x2)<<4)>cam->zfar+64.0f)
			x2 = mx>x2 ? mx - (int)(cam->zfar/16.0f)-4 : mx + 4 + (int)(cam->zfar/16.0f);
		if((abs(mx-x1)<<4)>cam->zfar+64.0f)
			x1 = mx>x1 ? mx - (int)(cam->zfar/16.0f)-4 : mx + 4 + (int)(cam->zfar/16.0f);

		if(y1<0) y1=0;
		if(y2<0) y2=0;
		if(y1>=bloc_h) y1=bloc_h-1;
		if(y2>=bloc_h) y2=bloc_h-1;
		if(x1<0) x1=0;
		if(x2<0) x2=0;
		if(x1>=bloc_w) x1=bloc_w-1;
		if(x2>=bloc_w) x2=bloc_w-1;

		y1-=3;
		if(y1<0) y1=0;

		A=(cam->Dir+1.0f*(0.75f*cam->Up-cam->Side));	A.Unit();
		float ref=0.95f*(A%cam->Dir);
		float dhm=0.5f*map_h;
		float dwm=0.5f*map_w;

		if(!FLAT)
			glColor4f(1.0f,1.0f,1.0f,1.0f);

		POINTF flat[9];
		if(FLAT) {
			flat[0].x=0.0f;		flat[0].y=niv+cos(t)*0.5f;			flat[0].z=0.0f;
			flat[1].x=8.0f;		flat[1].y=niv+cos(t+1.0f)*0.5f;		flat[1].z=0.0f;
			flat[2].x=16.0f;	flat[2].y=flat[0].y;				flat[2].z=0.0f;
			flat[3].x=0.0f;		flat[3].y=niv+cos(t+1.5f)*0.5f;		flat[3].z=8.0f;
			flat[4].x=8.0f;		flat[4].y=niv+cos(t+2.5f)*0.5f;		flat[4].z=8.0f;
			flat[5].x=16.0f;	flat[5].y=flat[3].y;				flat[5].z=8.0f;
			flat[6].x=0.0f;		flat[6].y=flat[0].y;				flat[6].z=16.0f;
			flat[7].x=8.0f;		flat[7].y=flat[1].y;				flat[7].z=16.0f;
			flat[8].x=16.0f;	flat[8].y=flat[0].y;				flat[8].z=16.0f;
			}

		if(ntex>0 && !depth_only) {
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			}

		if(FLAT)
			glTranslatef(cos(t),0.0f,sin(t));
		int old_i=-1;
		GLuint old_tex=bloc[0].tex;
		if(!depth_only)
			glBindTexture(GL_TEXTURE_2D,old_tex);
		if(!FLAT && check_visibility) {
			for(y=oy1;y<=oy2;y++)
				memset(view[y]+ox1,0,ox2-ox1+1);
			features.min_idx=features.nb_features-1;
			features.max_idx=0;
			features.list_size=0;
			ox1=x1;	ox2=x2;
			oy1=y1;	oy2=y2;
			}
		else if(!check_visibility) {
			x1=ox1;	x2=ox2;
			y1=oy1;	y2=oy2;
			for(int i=0;i<features.list_size;i++)
				features.feature[features.list[i]].draw=true;
			}
		int lavaprob=(int)(1000*dt);
		POINTF	buf_p[900];				// Tampon qui accumule les blocs pour les dessiner en chaîne
		float	buf_t[1800];
		float	buf_c[3600];
		short	buf_size=0;					// En blocs
		glEnableClientState(GL_COLOR_ARRAY);		// Couleurs(pour le brouillard de guerre)
		glColorPointer(4,GL_FLOAT,0,buf_c);
		glVertexPointer( 3, GL_FLOAT, 0, buf_p);
		glTexCoordPointer(2, GL_FLOAT, 0, buf_t);
		int	ox=x1;

		bool low_def_view = cam->RPos.y>300.0f;		// Low detail map for mega zoom
		if(low_def_view) {							// draw the low detail map
			i=0;
			for(y=0;y<=low_h;y++) {
				int Y=y*(bloc_h_db-2)/low_h;
				for(x=0;x<=low_w;x++) {
					int X=x*(bloc_w_db-2)/low_w;
					int Z;
					Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X].view&player_mask)==0)	low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=0.1f;
					else if(!map_data[Z][X].timer)				low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=0.5f;
					else										low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=1.0f;
					i++;
					}
				}
			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer(4,GL_FLOAT,0,low_col);
			if(FLAT) {
				glTranslatef(0.0f,niv,0.0f);
				glVertexPointer( 3, GL_FLOAT, 0, low_vtx_flat);
				}
			else
				glVertexPointer( 3, GL_FLOAT, 0, low_vtx);
			glTexCoordPointer(2, GL_FLOAT, 0, low_tcoord);
			glBindTexture(GL_TEXTURE_2D,low_tex);
			glDrawElements(GL_TRIANGLE_STRIP, low_nb_idx,GL_UNSIGNED_INT,low_index);		// draw this map
			}

		if(cam->RPos.y>=900.0f) {
			memset(view[0],1,bloc_w*bloc_h);
			ox1=0;
			ox2=bloc_w-1;
			oy1=0;
			oy2=bloc_h-1;
			}

		if(cam->RPos.y<900.0f)
		for(y=y1;y<=y2;y++) {				// Balaye les blocs susceptibles d'être visibles pour dessiner ceux qui le sont
			int pre_y=y<<4;
			int Y=y<<1;
			int pre_y2=y*bloc_w;
			VECTOR T;
			T.x=-dwm;	T.y=0.0f;	T.z=pre_y-dhm;
			buf_size=0;
			ox=x1;
			for(x=x1;x<=x2;x++) {
				int X=x<<1;
				if(!FLAT && check_visibility) {
					VECTOR	V;
					V.x=(x<<4)-dwm;
					V.y=h_map[Y|1][X|1];
					V.z=pre_y-dhm+get_zdec(X|1,Y|1);
					V=V-cam->Pos;
					if(fabs(V%cam->Dir)>cam->zfar) {
						view[y][x]=0;
						continue;
						}
					float d=V.Sq();
					if(d>16384.0f)
						if((V%cam->Dir)<ref*sqrt(d)) {
							view[y][x]=0;
							continue;
							}
					if((map_data[Y][X].view&player_mask)==0 && (map_data[Y][X|1].view&player_mask)==0 && (map_data[Y|1][X].view&player_mask)==0 && (map_data[Y|1][X|1].view&player_mask)==0) {
						if(water) {
							if(map_data[Y][X].underwater || map_data[Y|1][X].underwater || map_data[Y][X|1].underwater || map_data[Y|1][X|1].underwater)
								view[y][x]=2;
							else
								view[y][x]=3;
							}
						}
					else {
						if(map_data[Y][X].timer==0 && map_data[Y][X|1].timer==0 && map_data[Y|1][X].timer==0 && map_data[Y|1][X|1].timer==0) {
							if(map_data[Y][X].underwater || map_data[Y|1][X].underwater || map_data[Y][X|1].underwater || map_data[Y|1][X|1].underwater)
								view[y][x]=2;
							else
								view[y][x]=3;
							}
						else
							view[y][x]=1;
						if(map_data[Y][X].stuff>=0 && map_data[Y][X].stuff<features.nb_features) {		// Indique comme affichables les objets présents sur le bloc
							if(features.feature[map_data[Y][X].stuff].type<0)
								map_data[Y][X].stuff=-1;
							else {
								features.feature[map_data[Y][X].stuff].draw=true;
								features.feature[map_data[Y][X].stuff].grey=(view[y][x]&2)==2;
								features.list[features.list_size++]=map_data[Y][X].stuff;
								}
							}
						if(map_data[Y][X|1].stuff>=0 && map_data[Y][X|1].stuff<features.nb_features) {
							if(features.feature[map_data[Y][X|1].stuff].type<0)
								map_data[Y][X|1].stuff=-1;
							else {
								features.feature[map_data[Y][X|1].stuff].draw=true;
								features.feature[map_data[Y][X|1].stuff].grey=(view[y][x]&2)==2;
								features.list[features.list_size++]=map_data[Y][X|1].stuff;
								}
							}
						if(map_data[Y|1][X].stuff>=0 && map_data[Y|1][X].stuff<features.nb_features) {
							if(features.feature[map_data[Y|1][X].stuff].type<0)
								map_data[Y|1][X].stuff=-1;
							else {
								features.feature[map_data[Y|1][X].stuff].draw=true;
								features.feature[map_data[Y|1][X].stuff].grey=(view[y][x]&2)==2;
								features.list[features.list_size++]=map_data[Y|1][X].stuff;
								}
							}
						if(map_data[Y|1][X|1].stuff>=0 && map_data[Y|1][X|1].stuff<features.nb_features) {
							if(features.feature[map_data[Y|1][X|1].stuff].type<0)
								map_data[Y|1][X|1].stuff=-1;
							else {
								features.feature[map_data[Y|1][X|1].stuff].draw=true;
								features.feature[map_data[Y|1][X|1].stuff].grey=(view[y][x]&2)==2;
								features.list[features.list_size++]=map_data[Y|1][X|1].stuff;
								}
							}
						}
					}
				else {
					if(view[y][x]==0)
						continue;
					if(view[y][x]==2 && !draw_uw)	continue;		// Jump this if it is under water and don't have to be drawn
					if(view[y][x]==3)
						view[y][x]=2;
					if(view[y][x]==2 && FLAT)
						view[y][x]=0;
					}
				if(low_def_view)	continue;
								// Si le joueur ne peut pas voir ce morceau, on ne le dessine pas en clair
				i=bmap[y][x];
				if(FLAT) {
					bloc[i].point=lvl[pre_y2+x];
					if(bloc[i].point==NULL || bloc[i].point[0].y<niv || bloc[i].point[1].y<niv || bloc[i].point[2].y<niv ||
						bloc[i].point[3].y<niv || bloc[i].point[4].y<niv || bloc[i].point[5].y<niv ||
						bloc[i].point[6].y<niv || bloc[i].point[7].y<niv || bloc[i].point[8].y<niv)
						bloc[i].point=flat;
					else
						continue;
					}
				else {
					if(check_visibility && bloc[i].lava && (rand_from_table()%1000000)<=lavaprob) {		// Lava emiting code moved here because of lava effect using fragment program
						POINTF POS;
						VECTOR V;
						POS.x=(x<<4)-dwm+8.0f;
						POS.z=pre_y-dhm+8.0f;
						POS.y=sealvl-10.0f;
						V.x=((rand_from_table()%201)-100);
						V.y=((rand_from_table()%51)+50);
						V.z=((rand_from_table()%201)-100);
						V.Unit();
						particle_engine.emit_lava(POS,V,1,10,(rand_from_table()%1000)*0.01f+30.0f);
						}
					bloc[i].point=lvl[pre_y2+x];
					if(bloc[i].point==NULL) {
						lvl[pre_y2+x]=bloc[i].point=(POINTF*) malloc(sizeof(POINTF)*9);
						if(tnt) {
							bloc[i].point[0].x=0.0f;	bloc[i].point[0].z=get_zdec(X,Y);
							bloc[i].point[1].x=8.0f;	bloc[i].point[1].z=get_zdec(X|1,Y);
							bloc[i].point[2].x=16.0f;	bloc[i].point[2].z=get_zdec(X+2,Y);
							bloc[i].point[3].x=0.0f;	bloc[i].point[3].z=8.0f+get_zdec(X,Y|1);
							bloc[i].point[4].x=8.0f;	bloc[i].point[4].z=8.0f+get_zdec(X|1,Y|1);
							bloc[i].point[5].x=16.0f;	bloc[i].point[5].z=8.0f+get_zdec(X+2,Y|1);
							bloc[i].point[6].x=0.0f;	bloc[i].point[6].z=16.0f+get_zdec(X,Y+2);
							bloc[i].point[7].x=8.0f;	bloc[i].point[7].z=16.0f+get_zdec(X|1,Y+2);
							bloc[i].point[8].x=16.0f;	bloc[i].point[8].z=16.0f+get_zdec(X+2,Y+2);
							bloc[i].point[0].y=get_nh(X,Y);
							bloc[i].point[1].y=get_nh(X|1,Y);
							bloc[i].point[2].y=get_nh(X+2,Y);
							bloc[i].point[3].y=get_nh(X,Y|1);
							bloc[i].point[4].y=get_nh(X|1,Y|1);
							bloc[i].point[5].y=get_nh(X+2,Y|1);
							bloc[i].point[6].y=get_nh(X,Y+2);
							bloc[i].point[7].y=get_nh(X|1,Y+2);
							bloc[i].point[8].y=get_nh(X+2,Y+2);
							}
						else {
							bloc[i].point[0].x=0.0f;	bloc[i].point[0].z=0.0f;
							bloc[i].point[1].x=8.0f;	bloc[i].point[1].z=0.0f;
							bloc[i].point[2].x=16.0f;	bloc[i].point[2].z=0.0f;
							bloc[i].point[3].x=0.0f;	bloc[i].point[3].z=8.0f;
							bloc[i].point[4].x=8.0f;	bloc[i].point[4].z=8.0f;
							bloc[i].point[5].x=16.0f;	bloc[i].point[5].z=8.0f;
							bloc[i].point[6].x=0.0f;	bloc[i].point[6].z=16.0f;
							bloc[i].point[7].x=8.0f;	bloc[i].point[7].z=16.0f;
							bloc[i].point[8].x=16.0f;	bloc[i].point[8].z=16.0f;
							bloc[i].point[0].y=get_h(X,Y);
							bloc[i].point[1].y=get_h(X|1,Y);
							bloc[i].point[2].y=get_h(X+2,Y);
							bloc[i].point[3].y=get_h(X,Y|1);
							bloc[i].point[4].y=get_h(X|1,Y|1);
							bloc[i].point[5].y=get_h(X+2,Y|1);
							bloc[i].point[6].y=get_h(X,Y+2);
							bloc[i].point[7].y=get_h(X|1,Y+2);
							bloc[i].point[8].y=get_h(X+2,Y+2);
							}
						}
					}

				if(bloc[i].tex!=old_tex || buf_size>=100 || ox+1<x) {
					glDrawElements(GL_TRIANGLE_STRIP, buf_size*13,GL_UNSIGNED_SHORT,buf_i);		// dessine le tout
					buf_size=0;
					old_tex=bloc[i].tex;
					glBindTexture(GL_TEXTURE_2D,bloc[i].tex);
					}
				ox=x;

				short buf_pos=buf_size*9;
				T.x+=x<<4;
				if(!FLAT)
					for(byte e=0;e<9;e++) {					// Copie le bloc
						buf_p[buf_pos+e].x=bloc[i].point[e].x+T.x;
						buf_p[buf_pos+e].y=bloc[i].point[e].y;
						buf_p[buf_pos+e].z=bloc[i].point[e].z+T.z;
						}
				else
					for(byte e=0;e<9;e++) {					// Copie le bloc
						buf_p[buf_pos+e].x=flat[e].x+T.x;
						buf_p[buf_pos+e].y=flat[e].y;
						buf_p[buf_pos+e].z=flat[e].z+T.z;
						}
				T.x-=x<<4;
//				memcpy(buf_t+(buf_pos<<1),bloc[i].texcoord,sizeof(float)*18);		// texture
				memcpy(buf_t+(buf_pos<<1),bloc[i].texcoord,72);		// texture
				float *color=buf_c+(buf_pos<<2);
				for(i=0;i<36;i+=4) {
					color[i]=color[i|1]=color[i|2]=1.0f;
					color[i|3]=FLAT ? 0.75f : 1.0f;
					}
				int Z;
				Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X].view&player_mask)==0)	color[16]=color[17]=color[18]=0.1f;
				else if(!map_data[Z][X].timer)				color[16]=color[17]=color[18]=0.5f;
				Z=Y+get_zdec_notest(X|1,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X|1].view&player_mask)==0)	color[20]=color[21]=color[22]=0.1f;
				else if(!map_data[Z][X|1].timer)			color[20]=color[21]=color[22]=0.5f;
				Z=Y+1+get_zdec_notest(X,Y|1);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X].view&player_mask)==0)	color[28]=color[29]=color[30]=0.1f;
				else if(!map_data[Z][X].timer)				color[28]=color[29]=color[30]=0.5f;
				Z=Y+1+get_zdec_notest(X|1,Y|1);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X|1].view&player_mask)==0)	color[32]=color[33]=color[34]=0.1f;
				else if(!map_data[Z][X|1].timer)			color[32]=color[33]=color[34]=0.5f;
				if(X>0) {
					Z=Y+get_zdec_notest(X-1,Y);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X-1].view&player_mask)==0)		color[12]=color[13]=color[14]=0.1f;
					else if(!map_data[Z][X-1].timer)				color[12]=color[13]=color[14]=0.5f;
					Z=Y+1+get_zdec_notest(X-1,Y|1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X-1].view&player_mask)==0)		color[24]=color[25]=color[26]=0.1f;
					else if(!map_data[Z][X-1].timer)				color[24]=color[25]=color[26]=0.5f;
					if(Y>0)
						Z=Y-1+get_zdec_notest(X-1,Y-1);		if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
						if((map_data[Z][X-1].view&player_mask)==0)		color[0]=color[1]=color[2]=0.1f;
						else if(!map_data[Z][X-1].timer)				color[0]=color[1]=color[2]=0.5f;
					}
				if(Y>0) {
					Z=Y-1+get_zdec_notest(X,Y-1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X].view&player_mask)==0)	color[4]=color[5]=color[6]=0.1f;
					else if(!map_data[Z][X].timer)				color[4]=color[5]=color[6]=0.5f;
					Z=Y-1+get_zdec_notest(X|1,Y-1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X|1].view&player_mask)==0)		color[8]=color[9]=color[10]=0.1f;
					else if(!map_data[Z][X|1].timer)				color[8]=color[9]=color[10]=0.5f;
					}
/*				Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				Z&=0xFFFFFE;
				X&=0xFFFFFE;
				if(path[Z][X] || path[Z+1][X] || path[Z][X+1] || path[Z+1][X+1])		// Affiche les résultats du pathfinding
					for(i=0;i<9;i++) {
						color[i<<2]=color[(i<<2)+1]=color[(i<<2)+2]=color[(i<<2)+3]=0.0f;
						color[(i<<2)+1]=1.0f;
						}*/
				buf_size++;
				}
			if(buf_size>0) {
				glDrawElements(GL_TRIANGLE_STRIP, buf_size*13,GL_UNSIGNED_SHORT,buf_i);		// dessine le tout
				buf_size=0;
				}
			}
		glDisableClientState(GL_COLOR_ARRAY);		// Couleurs(pour le brouillard de guerre)
		if(depth_only) {
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			}
	}

	VECTOR MAP::hit(VECTOR Pos,VECTOR Dir,bool water)			// Calcule l'intersection d'un rayon avec la carte(le rayon partant du dessus de la carte)
	{
		if(Dir.x==0.0f && Dir.z==0.0f) {			// Solution triviale
			VECTOR P=Pos;
			P.y=get_unit_h(P.x,P.z);
			return P;
			}
		if(get_unit_h(Pos.x,Pos.z)>Pos.y)		// Cas non traité
			return Pos;
		float step=1.0f;
		if(Dir.x!=0.0f && Dir.z!=0.0f) {
			if(fabs(Dir.x)<fabs(Dir.z))
				step=1.0f/fabs(Dir.x);
			else
				step=1.0f/fabs(Dir.z);
			}
		int nb=0;
		float dwm=map_w_d;
		float dhm=map_h_d;
		Dir=(4.0f*step)*Dir;
		while(((sealvl<Pos.y && water) || !water) && get_unit_h(Pos.x,Pos.z)<Pos.y) {
			if(nb>=1000)
				return Pos;
			nb++;
			Pos=Pos+Dir;
			if(fabs(Pos.x)>dwm || fabs(Pos.z)>dhm)		// Pas de résultat
				return Pos;
			}
		for(byte i=0;i<7;i++) {
			Pos=Pos-Dir;		// On recommence la dernière opération mais avec plus de précision
			Dir=0.5f*Dir;
			nb=0;
			while(((sealvl<Pos.y && water) || !water) && get_unit_h(Pos.x,Pos.z)<Pos.y) {
				if(nb>=2)
					return Pos;
				nb++;
				Pos=Pos+Dir;
				}
			}
		return Pos;		// Meilleure solution approximative trouvée
	}

	void WATER::draw(float t,float X,float Y,bool flat)
	{
		glActiveTextureARB(GL_TEXTURE0_ARB);
		if(flat) {
			glTranslatef(cos(t),0.0f,sin(t));
			glBegin(GL_QUADS);
				glTexCoord2f(-map_w/w+0.5f,-map_h/w+0.5f);		glVertex3f(-map_w,0.0f,-map_h);
				glTexCoord2f(map_w/w+0.5f,-map_h/w+0.5f);		glVertex3f(map_w,0.0f,-map_h);
				glTexCoord2f(map_w/w+0.5f,map_h/w+0.5f);		glVertex3f(map_w,0.0f,map_h);
				glTexCoord2f(-map_w/w+0.5f,map_h/w+0.5f);		glVertex3f(-map_w,0.0f,map_h);
			glEnd();
			return;
			}
		glTranslatef(X+cos(t),0.0f,Y+sin(t));
		glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glTranslatef(X/w,Y/w,0.0f);
		glMatrixMode(GL_MODELVIEW);
		if(!flat) {
			int i=0;
			float cv=w/s;
			for(int y=0;y<=s;y++) {
				float phaseY=y*cv;
				for(int x=0;x<=s;x++) {
					float phaseX=x*cv;
					point[i++].y=(cos(t+(X+Y+phaseX+phaseY)*0.1f)+sin(1.1f*t+(X-Y+phaseX-phaseY)*0.1f))*0.25f;
					}
				}
			normalise();
			}

		glEnableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);
		glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
		glVertexPointer( 3, GL_FLOAT, 0, point);
		glNormalPointer( GL_FLOAT, 0, N);
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, texcoord);

		glDrawElements(GL_TRIANGLES, nb_idx,GL_UNSIGNED_INT,index);		// dessine le tout

		glDisableClientState(GL_NORMAL_ARRAY);
	}

	void SKY::draw()
	{
		glDisableClientState(GL_COLOR_ARRAY);
		glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
		glVertexPointer( 3, GL_FLOAT, 0, point);
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, texcoord);

		glDrawElements(GL_TRIANGLES, nb_idx,GL_UNSIGNED_SHORT,index);		// dessine le tout
	}

	void UNIT::add_mission(int mission_type,VECTOR *target,bool step,int dat,void *pointer,PATH_NODE *path,byte flags)
	{
		MISSION *new_mission=(MISSION*) malloc(sizeof(MISSION));
		new_mission->next=NULL;
		new_mission->mission=mission_type;
		new_mission->step=step;
		new_mission->time=0.0f;
		new_mission->data=dat;
		new_mission->p=pointer;
		new_mission->path=path;
		new_mission->last_d=9999999.0f;
		new_mission->flags=flags;

		if(target)
			new_mission->target=*target;

		MISSION *stop=(MISSION*) malloc(sizeof(MISSION));
		stop->mission=MISSION_STOP;
		stop->step=true;
		stop->time=0.0f;
		stop->p=NULL;
		stop->data=0;
		stop->path=NULL;
		stop->last_d=9999999.0f;
		stop->flags=flags;
		if(step) {
			stop->next=NULL;
			new_mission->next=stop;
			}
		else {
			stop->next=new_mission;
			new_mission=stop;
			}

		if(step && mission) {
//			new_mission->next=mission;
			stop->next=mission;
			mission=new_mission;
			start_mission_script(mission->mission);
			}
		else {
			if(mission) {				// Ajoute l'ordre aux autres
				MISSION *cur=mission;
				while(cur->next!=NULL)	cur=cur->next;
				cur->next=new_mission;
				}
			else {
				mission=new_mission;
				start_mission_script(mission->mission);
				}
			}
	}

	void UNIT::set_mission(int mission_type,VECTOR *target,bool step,int dat,bool stopit,void *pointer,PATH_NODE *path,byte flags)
	{
		int old_mission=-1;
		if(mission!=NULL)	old_mission=mission->mission;
		while(mission!=NULL) 	clear_mission();			// Efface les ordres précédents

		switch(old_mission)		// Commandes de fin de mission
		{
		case MISSION_REPAIR:
		case MISSION_RECLAIM:
		case MISSION_BUILD_2:
			launch_script(get_script_index(SCRIPT_stopbuilding));
			launch_script(get_script_index(SCRIPT_stop));
			break;
		case MISSION_ATTACK:
		case MISSION_AIM:
		case MISSION_SHOOT:
			deactivate();
			break;
		};

		mission=(MISSION*) malloc(sizeof(MISSION));
		mission->next=NULL;
		mission->mission=mission_type;
		mission->step=step;
		mission->time=0.0f;
		mission->data=dat;
		mission->p=pointer;
		mission->path=path;
		mission->last_d=9999999.0f;
		mission->flags=flags;
		if(target)
			mission->target=*target;

		if(stopit) {
			MISSION *stop=(MISSION*) malloc(sizeof(MISSION));
			stop->next=mission;
			stop->mission=MISSION_STOP;
			stop->step=true;
			stop->time=0.0f;
			stop->p=NULL;
			stop->data=0;
			stop->path=NULL;
			stop->last_d=9999999.0f;
			stop->flags=flags;
			mission=stop;
			}
		else
			start_mission_script(mission->mission);
		c_time=0.0f;
	}

	void UNIT::draw(CAMERA *cam,MAP *map,bool cullface,bool height_line)
	{
		visible=false;
		if(model==NULL) return;		// S'il n'y a pas de modèle associé, on quitte la fonction

		int px=cur_px>>1;
		int py=cur_py>>1;
		if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	return;	// Unité hors de la carte
		if(map->view[py][px]!=1) return;	// Unité non visible

		VECTOR D;
		D=Pos-cam->Pos;		// Vecteur "viseur unité" partant de la caméra vers l'unité

		float dist=D.Sq();
		if(dist>=16384.0f && (D%cam->Dir)<=0.0f)		return;
		if((D%cam->Dir)>cam->zfar2)	return;		// Si l'objet est hors champ on ne le dessine pas

		visible=true;

		MATRIX_4x4 M;
		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		if(cam->RPos.y>300.0f) {			// for mega zoom, draw only a triangle
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glColor3f(player_color[owner_id*3],player_color[owner_id*3+1],player_color[owner_id*3+2]);
			glBegin(GL_TRIANGLES);
				glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glVertex3f(model->center.x-unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glVertex3f(model->center.x,model->center.y,model->center.z+unit_manager.unit_type[type_id].FootprintZ*4.0f);
			glEnd();
			if(owner_id==players.local_human_id && sel) {
				glColor3f(1.0f,1.0f,0.0f);
				glBegin(GL_LINE_STRIP);
					glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x-unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x,model->center.y,model->center.z+unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glEnd();
				}
			}
		else {
			float scale=unit_manager.unit_type[type_id].Scale;
			glScalef(scale,scale,scale);

			M=RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale);			// Matrice pour le calcul des positions des éléments du modèle de l'unité

			VECTOR *target=NULL,*center=NULL;
			POINTF upos;
			bool c_part=false;
			bool reverse=false;
			float size=0.0f;
			if(port[BUILD_PERCENT_LEFT]==0.0f && mission!=NULL
			&& (mission->mission==MISSION_REPAIR || mission->mission==MISSION_BUILD_2 || mission->mission==MISSION_RECLAIM))
				if(c_time>=0.25f) {
					reverse=(mission->mission==MISSION_RECLAIM);
					c_time=0.0f;
					c_part=true;
					target=&(mission->target);
					upos.x=upos.y=upos.z=0.0f;
					upos=upos+Pos;
					if(mission->p!=NULL && ((UNIT*)mission->p)->flags && ((UNIT*)mission->p)->model!=NULL) {
						size=((UNIT*)mission->p)->model->size2;
						center=&((UNIT*)mission->p)->model->center;
						}
					else if(mission->mission==MISSION_RECLAIM) {		// Récupération d'objets
						D.x=D.y=D.z=0.f;
						center=&D;
						size=32.0f;
						}
					else
						c_part=false;
					}
			if(port[BUILD_PERCENT_LEFT]==0.0f) {
				model->draw(&data,owner_id==players.local_human_id && sel,false,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				if(height_line && h>1.0f && unit_manager.unit_type[type_id].canfly) {		// For flying units, draw a line that shows how high is the unit
					cam->SetView();					// Reset the camera
					glDisable(GL_TEXTURE_2D);
					glDisable(GL_LIGHTING);
					glColor3f(1.0f,1.0f,0.0f);
					glBegin(GL_LINES);
						for(float y=Pos.y;y>Pos.y-h;y-=10.0f) {
							glVertex3f(Pos.x,y,Pos.z);
							glVertex3f(Pos.x,y-5.0f,Pos.z);
							}
					glEnd();
					}
				}
			else if(port[BUILD_PERCENT_LEFT]<=33.0f) {
				float h=4.0f*model->size2;
				double eqn[4]= { 0.0f, 1.0f, 0.0f, model->center.y+h*0.5f-h*(33.0f-port[BUILD_PERCENT_LEFT])*0.033333f};

				glClipPlane(GL_CLIP_PLANE0, eqn);
				glEnable(GL_CLIP_PLANE0);
				model->draw(&data,owner_id==players.local_human_id && sel,true,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);

				eqn[1]=-eqn[1];	eqn[3]=-eqn[3];
				glClipPlane(GL_CLIP_PLANE0, eqn);
				model->draw(&data,owner_id==players.local_human_id && sel,false,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glDisable(GL_CLIP_PLANE0);

				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				}
			else if(port[BUILD_PERCENT_LEFT]<=66.0f) {
				float h=4.0f*model->size2;
				double eqn[4]= { 0.0f, 1.0f, 0.0f, model->center.y+h*0.5f-h*(66.0f-port[BUILD_PERCENT_LEFT])*0.033333f};

				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				glClipPlane(GL_CLIP_PLANE0, eqn);
				glEnable(GL_CLIP_PLANE0);
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

				eqn[1]=-eqn[1];	eqn[3]=-eqn[3];
				glClipPlane(GL_CLIP_PLANE0, eqn);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glDisable(GL_CLIP_PLANE0);
				}
			else {
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				}
			}
	}

	void UNIT::draw_shadow(CAMERA *cam,VECTOR Dir,MAP *map)
	{
		if(!visible) {
			VECTOR S_Pos = map->hit(Pos,Dir);
			int px=((int)(S_Pos.x+map->map_w_d))>>4;
			int py=((int)(S_Pos.z+map->map_h_d))>>4;
			if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	return;	// Shadow out of the map
			if(map->view[py][px]!=1) return;	// Unvisible shadow
			}

		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		float scale=unit_manager.unit_type[type_id].Scale;
		glScalef(scale,scale,scale);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		float scale_dir=model->size2;
		scale_dir+=Pos.y;
		scale_dir*=2.0f;
		model->draw_shadow(((scale_dir*Dir*RotateX(-Angle.x*DEG2RAD))*RotateZ(-Angle.z*DEG2RAD))*RotateY(-Angle.y*DEG2RAD),&data);
	}

	void UNIT::draw_shadow_basic(CAMERA *cam,VECTOR Dir,MAP *map)
	{
		if(!visible) {
			VECTOR S_Pos = map->hit(Pos,Dir);
			int px=((int)(S_Pos.x+map->map_w_d))>>4;
			int py=((int)(S_Pos.z+map->map_h_d))>>4;
			if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	return;	// Shadow out of the map
			if(map->view[py][px]!=1) return;	// Unvisible shadow
			}

		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		float scale=unit_manager.unit_type[type_id].Scale;
		glScalef(scale,scale,scale);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		float scale_dir=model->size2;
		scale_dir+=Pos.y+model->center.y;
		scale_dir*=2.0f;
		model->draw_shadow_basic(((scale_dir*Dir*RotateX(-Angle.x*DEG2RAD))*RotateZ(-Angle.z*DEG2RAD))*RotateY(-Angle.y*DEG2RAD),&data);
	}

	const void UNIT::run_script(const float dt,const int id,MAP *map)			// Interprète les scripts liés à l'unité
	{
		if(flags==0) return;
		if(!script_env[id].running)	return;
		if(script_env[id].wait>0.0f) {
			script_env[id].wait-=dt;
			return;
			}
		if(script==NULL || script_env[id].env==NULL) {
			script_env[id].running=false;
			return;	// S'il n'y a pas de script associé on quitte la fonction
			}
		sint16 script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
		sint16 pos=(script_env[id].env->cur>>8);

		if(script_id<0 || script_id>=script->nb_script)	{
			script_env[id].running=false;
			return;		// Erreur, ce n'est pas un script repertorié
			}

		float divisor=i2pwr16;
		float div=0.5f*divisor;
		POINTF O;
		O.x=O.y=O.z=0.0f;
//		bool done=false;
//		int nb_code=0;

//#define PRINT_CODE			// Affiche le code éxécuté dans la console

//		do
//		{
//			uint32 code=script->script_code[script_id][pos];			// Lit un code
//			pos++;
//			nb_code++;
//			if(nb_code>=1) done=true;			// Pour éviter que le programme ne fige à cause d'un script
//			switch(code)			// Code de l'interpréteur
			switch(script->script_code[script_id][pos++])
			{
			case SCRIPT_MOVE_OBJECT:
#ifdef PRINT_CODE
				printf("MOVE_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis==0)
						v1=-v1;
					data.axe[axis][obj].move_distance=v1*div;
					data.axe[axis][obj].move_distance-=data.axe[axis][obj].pos;
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(data.axe[axis][obj].move_distance<0.0f)
						data.axe[axis][obj].move_speed=-fabs(v2*div*0.5f);
					else
						data.axe[axis][obj].move_speed=fabs(v2*div*0.5f);
				}
				break;
			case SCRIPT_WAIT_FOR_TURN:
#ifdef PRINT_CODE
				printf("WAIT_FOR_TURN\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					float a=data.axe[axis][obj].rot_angle;
					if((data.axe[axis][obj].rot_speed!=0.0f || data.axe[axis][obj].rot_accel!=0.0f) && (a!=0.0f && data.axe[axis][obj].rot_limit))
						pos-=3;
					else if(data.axe[axis][obj].rot_speed!=data.axe[axis][obj].rot_target_speed && data.axe[axis][obj].rot_speed_limit)
						pos-=3;
				}
//				done=true;
				break;
			case SCRIPT_RANDOM_NUMBER:
#ifdef PRINT_CODE
				printf("RANDOM_NUMBER\n");
#endif
				{
					int high=script_env[id].pop();
					int low=script_env[id].pop();
					script_env[id].push((rand_from_table()%(high-low+1))+low);
				}
				break;
			case SCRIPT_GREATER_EQUAL:
#ifdef PRINT_CODE
				printf("GREATER_EQUAL\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1>=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_GREATER:
#ifdef PRINT_CODE
				printf("GREATER\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1>v2 ? 1 : 0);
				}
				break;
			case SCRIPT_LESS:
#ifdef PRINT_CODE
				printf("LESS\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1<v2 ? 1 : 0);
				}
				break;
			case SCRIPT_EXPLODE:
#ifdef PRINT_CODE
				printf("EXPLODE\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int explosion_type=script_env[id].pop();
					if(visible)					// Don't draw things which could tell the player there is something there
						particle_engine.make_fire(O+Pos,1,(int)log((float)(1+explosion_type))<<2,45.0f);
					data.flag[obj]|=FLAG_EXPLODE;
					data.explosion_flag[obj]=explosion_type;
					data.axe[0][obj].move_speed=(25.0f+(rand_from_table()%2501)*0.01f)*(rand_from_table()&1 ? 1.0f : -1.0f);
					data.axe[0][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.axe[1][obj].move_speed=25.0f+(rand_from_table()%2501)*0.01f;
					data.axe[1][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.axe[2][obj].move_speed=(25.0f+(rand_from_table()%2501)*0.01f)*(rand_from_table()&1 ? 1.0f : -1.0f);
					data.axe[2][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.explode=true;
				}
				break;
			case SCRIPT_TURN_OBJECT:
#ifdef PRINT_CODE
				printf("TURN_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis!=2) {
						v1=-v1;
						v2=-v2;
						}
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_angle=-v1*TA2DEG;
					data.axe[axis][obj].rot_accel=0.0f;
					data.axe[axis][obj].rot_angle-=data.axe[axis][obj].angle;
					while(data.axe[axis][obj].rot_angle>180.0f)					// Fait le tour dans le sens le plus rapide
						data.axe[axis][obj].rot_angle-=360.0f;
					while(data.axe[axis][obj].rot_angle<-180.0f)					// Fait le tour dans le sens le plus rapide
						data.axe[axis][obj].rot_angle+=360.0f;
					if(data.axe[axis][obj].rot_angle>0.0f)
						data.axe[axis][obj].rot_speed=fabs(v2*TA2DEG);
					else
						data.axe[axis][obj].rot_speed=-fabs(v2*TA2DEG);
					data.axe[axis][obj].rot_limit=true;
					data.axe[axis][obj].rot_speed_limit=false;
				}
				break;
			case SCRIPT_WAIT_FOR_MOVE:
#ifdef PRINT_CODE
				printf("WAIT_FOR_MOVE\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					float a=data.axe[axis][obj].rot_angle;
					if(data.axe[axis][obj].move_distance!=0.0f)
						pos-=3;
				}
//				done=true;
				break;
			case SCRIPT_CREATE_LOCAL_VARIABLE:
#ifdef PRINT_CODE
				printf("CREATE_LOCAL_VARIABLE\n");
#endif
				break;
			case SCRIPT_SUBTRACT:
#ifdef PRINT_CODE
				printf("SUBSTRACT\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1-v2);
				}
				break;
			case SCRIPT_GET_VALUE_FROM_PORT:
#ifdef PRINT_CODE
				printf("GET_VALUE_FROM_PORT: ");
#endif
				{
					int value=script_env[id].pop();
#ifdef PRINT_CODE
					printf("%d\n",value);
#endif
					switch(value)
					{
					case ATAN:
						{
							int v1=script_env[id].pop();
							int v2=script_env[id].pop();
							port[value]=atan((float)v1/v2);
						}
						break;
					case HYPOT:
						{
							int v1=script_env[id].pop();
							int v2=script_env[id].pop();
							port[value]=sqrt((float)(v1*v1+v2*v2));
						}
						break;
					case BUGGER_OFF:
						port[value]=map->check_rect((((int)(Pos.x+map->map_w_d))>>3)-(unit_manager.unit_type[type_id].FootprintX>>1),(((int)(Pos.z+map->map_h_d))>>3)-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx) ? 0.0f : 1.0f;
						break;
					case YARD_OPEN:
						break;
					case HEALTH:
					case BUILD_PERCENT_LEFT:
						break;
					case STANDINGMOVEORDERS:			// A faire : ajouter le support des ordres de mouvement/feu
					case STANDINGFIREORDERS:
						break;
					default:
						{
							char *op[]={"INCONNU","ACTIVATION","STANDINGMOVEORDERS","STANDINGFIREORDERS","HEALTH","INBUILDSTANCE","BUSY","PIECE_XZ","PIECE_Y",
							"UNIT_XZ","UNIT_Y","UNIT_HEIGHT","XZ_ATAN","XZ_HYPOT","ATAN","HYPOT","GROUND_HEIGHT","BUILD_PERCENT_LEFT","YARD_OPEN",
							"BUGGER_OFF","ARMORED"};
							if(value>20)
								value=0;
							printf("GET_VALUE_FROM_PORT: ");
							printf("opération non gérée : %s\n",op[value]);
						}
						break;
					};
					script_env[id].push((int)port[value]);
				}
				break;
			case SCRIPT_LESS_EQUAL:
#ifdef PRINT_CODE
				printf("LESS_EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1<=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_SPIN_OBJECT:
#ifdef PRINT_CODE
				printf("SPIN_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis!=2) {
						v1=-v1;
						v2=-v2;
						}
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_limit=false;
					data.axe[axis][obj].rot_speed_limit=true;
					data.axe[axis][obj].rot_target_speed=v1*TA2DEG;
					if(v2) {
						if(data.axe[axis][obj].rot_target_speed>data.axe[axis][obj].rot_speed)
							data.axe[axis][obj].rot_accel=fabs(v2*TA2DEG);
						else
							data.axe[axis][obj].rot_accel=-fabs(v2*TA2DEG);
						}
					else {
						data.axe[axis][obj].rot_accel=0;
						data.axe[axis][obj].rot_speed=data.axe[axis][obj].rot_target_speed;
						}
				}
				break;
			case SCRIPT_SLEEP:
#ifdef PRINT_CODE
				printf("SLEEP\n");
#endif
				script_env[id].wait=script_env[id].pop()*0.001f;
//				done=true;
				break;
			case SCRIPT_MULTIPLY:
#ifdef PRINT_CODE
				printf("MULTIPLY\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1*v2);
				}
				break;
			case SCRIPT_CALL_SCRIPT:
#ifdef PRINT_CODE
				printf("CALL_SCRIPT\n");
#endif
				{
				int function_id=script->script_code[script_id][pos];			// Lit un code
				pos++;
				int num_param=script->script_code[script_id][pos];			// Lit un code
				pos++;
				script_env[id].env->cur=script_id+(pos<<8);
				SCRIPT_ENV_STACK *old=script_env[id].env;
				script_env[id].env=new SCRIPT_ENV_STACK;
				script_env[id].env->init();
				script_env[id].env->next=old;
				script_env[id].env->cur=function_id;
				for(int i=0;i<num_param;i++)		// Lit les paramètres
					script_env[id].env->var[i]=script_env[id].pop();
//				done=true;
				pos=0;
				script_id=function_id;
				}
				break;
			case SCRIPT_SHOW_OBJECT:
#ifdef PRINT_CODE
				printf("SHOW_OBJECT\n");
#endif
				data.flag[script->script_code[script_id][pos++]]&=(~FLAG_HIDE);
				break;
			case SCRIPT_EQUAL:
#ifdef PRINT_CODE
				printf("EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1==v2 ? 1 : 0);
				}
				break;
			case SCRIPT_NOT_EQUAL:
#ifdef PRINT_CODE
				printf("NOT_EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1!=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_IF:
#ifdef PRINT_CODE
				printf("IF\n");
#endif
				if(script_env[id].pop()!=0)
					pos++;
				else {
					int target_offset=script->script_code[script_id][pos];			// Lit un code
					pos=target_offset-script->dec_offset[script_id];								// Déplace l'éxecution
					}
				break;
			case SCRIPT_HIDE_OBJECT:
#ifdef PRINT_CODE
				printf("HIDE_OBJECT\n");
#endif
				data.flag[script->script_code[script_id][pos++]]|=FLAG_HIDE;
				break;
			case SCRIPT_SIGNAL:
#ifdef PRINT_CODE
				printf("SIGNAL\n");
#endif
				script_env[id].env->cur=script_id+(pos<<8);			// Sauvegarde la position
				raise_signal(script_env[id].pop());					// Tue tout les processus utilisant ce signal
				return;												// Quitte la fonction d'émulation de scripts
			case SCRIPT_DONT_CACHE:
#ifdef PRINT_CODE
				printf("DONT_CACHE\n");
#endif
				pos++;
				break;
			case SCRIPT_SET_SIGNAL_MASK:
#ifdef PRINT_CODE
				printf("SET_SIGNAL_MASK\n");
#endif
				script_env[id].env->signal_mask=script_env[id].pop();
				break;
			case SCRIPT_NOT:
#ifdef PRINT_CODE
				printf("NOT\n");
#endif
				script_env[id].push(!script_env[id].pop());
				break;
			case SCRIPT_DONT_SHADE:
#ifdef PRINT_CODE
				printf("DONT_SHADE\n");
#endif
				pos++;
				break;
			case SCRIPT_EMIT_SFX:
#ifdef PRINT_CODE
				printf("EMIT_SFX: ");
#endif
				{
					int smoke_type=script_env[id].pop();
					int from_piece=script->script_code[script_id][pos++];
#ifdef PRINT_CODE
					printf("smoke_type %d from %d\n",smoke_type,from_piece);
#endif
					if(visible) {
						compute_model_coord();
						if(data.dir[from_piece].x!=0.0f || data.dir[from_piece].y!=0.0f || data.dir[from_piece].z!=0.0f) {
							VECTOR dir=data.dir[from_piece];
							switch(smoke_type)
							{
							case 0:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,fire,1,10.0f,2.5f,5.0f,true);
								break;
							case 2:
							case 3:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,0,1,10.0f);
								break;
							case 257:			// Fumée
							case 258:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,0,1,10.0f,10.0f,10.0f,true);
								break;
							};
							}
						else
							switch(smoke_type)
							{
							case 0:
								particle_engine.make_smoke(O+Pos+data.pos[from_piece],fire,1,0.0f,0.0f);
								break;
							case 257:
							case 258:
								particle_engine.make_smoke(O+Pos+data.pos[from_piece],0,1,10.0f);
								break;
							};
						}
				}
				break;
			case SCRIPT_PUSH_CONST:
#ifdef PRINT_CODE
				printf("PUSH_CONST\n");
#endif
				script_env[id].push(script->script_code[script_id][pos]);
				pos++;
				break;
			case SCRIPT_PUSH_VAR:
#ifdef PRINT_CODE
				printf("PUSH_VAR\n");
#endif
				script_env[id].push(script_env[id].env->var[script->script_code[script_id][pos]]);
				pos++;
				break;
			case SCRIPT_SET_VAR:
#ifdef PRINT_CODE
				printf("SET_VAR\n");
#endif
				script_env[id].env->var[script->script_code[script_id][pos]]=script_env[id].pop();
				pos++;
				break;
			case SCRIPT_PUSH_STATIC_VAR:
#ifdef PRINT_CODE
				printf("PUSH_STATIC_VAR\n");
#endif
				script_env[id].push(s_var[script->script_code[script_id][pos]]);
				pos++;
				break;
			case SCRIPT_SET_STATIC_VAR:
#ifdef PRINT_CODE
				printf("SET_STATIC_VAR\n");
#endif
				s_var[script->script_code[script_id][pos]]=script_env[id].pop();
				pos++;
				break;
			case SCRIPT_OR:
#ifdef PRINT_CODE
				printf("OR\n");
#endif
				{
					int v1=script_env[id].pop(),v2=script_env[id].pop();
					script_env[id].push(v1|v2);
				}
				break;
			case SCRIPT_START_SCRIPT:				// Transfère l'éxecution à un autre script
#ifdef PRINT_CODE
				printf("START_SCRIPT\n");
#endif
				{
				int function_id=script->script_code[script_id][pos++];			// Lit un code
				int num_param=script->script_code[script_id][pos++];			// Lit un code
				int s_id=launch_script(function_id);
				if(s_id>=0) {
					for(int i=0;i<num_param;i++)		// Lit les paramètres
						script_env[s_id].env->var[i]=script_env[id].pop();
					script_env[s_id].env->signal_mask=script_env[id].env->signal_mask;
					}
				else
					for(int i=0;i<num_param;i++)		// Enlève les paramètres de la pile
						script_env[id].pop();
//				script_env[id].running=false;		// Commenté car empêche le script de l'Atlas de fonctionner correctement
//				done=true;
				}
				break;
			case SCRIPT_RETURN:		// Retourne au script précédent
#ifdef PRINT_CODE
				printf("RETURN\n");
#endif
				{
					script_val[script_id]=script_env[id].env->var[0];
					SCRIPT_ENV_STACK *old=script_env[id].env;
					script_env[id].env=script_env[id].env->next;
					delete old;
					script_env[id].pop();		// Enlève la valeur retournée
				}
				if(script_env[id].env) {
					script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
					pos=(script_env[id].env->cur>>8);
					}
//				done=true;
				break;
			case SCRIPT_JUMP:						// Commande de saut
#ifdef PRINT_CODE
				printf("JUMP\n");
#endif
				{
					int target_offset=script->script_code[script_id][pos];			// Lit un code
					pos=target_offset-script->dec_offset[script_id];								// Déplace l'éxecution
				}
				break;
			case SCRIPT_ADD:
#ifdef PRINT_CODE
				printf("ADD\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1+v2);
				}
				break;	//added
			case SCRIPT_STOP_SPIN:
#ifdef PRINT_CODE
				printf("STOP_SPIN\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v=script_env[id].pop();
					if(axis!=2)
						v=-v;
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_limit=false;
					data.axe[axis][obj].rot_speed_limit=true;
					data.axe[axis][obj].rot_target_speed=0.0f;
					if(v==0) {
						data.axe[axis][obj].rot_speed=0.0f;
						data.axe[axis][obj].rot_accel=0.0f;
						}
					else {
						if(data.axe[axis][obj].rot_speed>0.0f)
							data.axe[axis][obj].rot_accel=-abs(v);
						else
							data.axe[axis][obj].rot_accel=abs(v);
						}
				}
				break;	//added
			case SCRIPT_DIVIDE:
#ifdef PRINT_CODE
				printf("DIVIDE\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v2 / v1);
				}
				break;	//added
			case SCRIPT_MOVE_PIECE_NOW:
#ifdef PRINT_CODE
				printf("MOVE_PIECE_NOW\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(axis==0)
						data.axe[axis][obj].pos=-script_env[id].pop()*div;
					else
						data.axe[axis][obj].pos=script_env[id].pop()*div;
				}
				break;	//added
			case SCRIPT_TURN_PIECE_NOW:
#ifdef PRINT_CODE
				printf("TURN_PIECE_NOW\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v=script_env[id].pop();
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(axis!=2)
						v=-v;
					data.axe[axis][obj].angle=-v*TA2DEG;
				}
				break;	//added
			case SCRIPT_CACHE:
#ifdef PRINT_CODE
				printf("CACHE\n");
#endif
				pos++;
				break;	//added
			case SCRIPT_COMPARE_AND:
#ifdef PRINT_CODE
				printf("COMPARE_AND\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1 && v2);
				}
				break;	//added
			case SCRIPT_COMPARE_OR:
#ifdef PRINT_CODE
				printf("COMPARE_OR\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1 || v2);
				}
				break;	//added
			case SCRIPT_CALL_FUNCTION:
#ifdef PRINT_CODE
				printf("CALL_FUNCTION\n");
#endif
				{
				int function_id=script->script_code[script_id][pos++];			// Lit un code
				int num_param=script->script_code[script_id][pos++];			// Lit un code
				script_env[id].env->cur=script_id+(pos<<8);
				SCRIPT_ENV_STACK *old=script_env[id].env;
				script_env[id].env=new SCRIPT_ENV_STACK;
				script_env[id].env->init();
				script_env[id].env->next=old;
				script_env[id].env->cur=function_id;
				for(int i=0;i<num_param;i++)		// Lit les paramètres
					script_env[id].env->var[i]=script_env[id].pop();
//				done=true;
				pos=0;
				script_id=function_id;
				}
				break;	//added
			case SCRIPT_GET:
				printf("GET *\n");
				{
					int v4=script_env[id].pop();
					int v3=script_env[id].pop();
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					int val=script_env[id].pop();
					switch(val)
					{
					case ACTIVATION:
					case STANDINGMOVEORDERS:
					case STANDINGFIREORDERS:
					case HEALTH:
					case INBUILDSTANCE:
					case BUSY:
					case BUILD_PERCENT_LEFT:
					case YARD_OPEN:
					case BUGGER_OFF:
					case ARMORED:
						script_env[id].push((int)port[val]);
						break;
					case PIECE_XZ:
						compute_model_coord();
						script_env[id].push(((int)data.pos[v1].x<<16) | ((int)data.pos[v1].z));
						break;
					case PIECE_Y:
						compute_model_coord();
						script_env[id].push((int)data.pos[v1].y);
						break;
					case UNIT_XZ:
						if (v1 >= 0 && v1<units.last_index && units.unit[v1].flags)	
							script_env[id].push(((int)units.unit[v1].Pos.x<<16) | ((int)units.unit[v1].Pos.z));
						else
							script_env[id].push(((int)Pos.x<<16) | ((int)Pos.z));
/*						if (v1 == 0)	
							script_env[id].push(((int)Pos.x<<16) | ((int)Pos.z));
						else
							script_env[id].push(0);*/
						break;
					case UNIT_Y:
						if (v1 >= 0 && v1<units.last_index && units.unit[v1].flags)	
							script_env[id].push((int)units.unit[v1].Pos.y);
						else
							script_env[id].push((int)Pos.y);
						break;
					case UNIT_HEIGHT:
						if (v1 >= 0 && v1<units.last_index && units.unit[v1].flags)	
							script_env[id].push((int)units.unit[v1].model->size2);
						else
							script_env[id].push((int)model->size2);
						break;
					case XZ_ATAN:
						{
							int x=v1>>16;
							int z=v1&0xFFFF;
							script_env[id].push((int)atan((float)x/z));
						}
						break;
					case XZ_HYPOT:
						{
							int x=v1>>16;
							int z=v1&0xFFFF;
							script_env[id].push((int)sqrt((float)(x*x+z*z)));
						}
						break;
					case ATAN:
						script_env[id].push((int)atan((float)v1/(v2&0xFFFF)));
						break;
					case HYPOT:
						script_env[id].push((int)sqrt((float)(v1*v1+((v2&0xFFFF)*(v2&0xFFFF)))));
						break;
					case GROUND_HEIGHT:
						script_env[id].push((int)map->get_unit_h((v1>>16)&0xFFFF,v1&0xFFFF));
//						script_env[id].push((int)map->get_unit_h(Pos.x,Pos.z));
						break;
					default:
						printf("GET constante inconnue %d", val);
					};
				}
				break;	//added
			case SCRIPT_SET_VALUE:
#ifdef PRINT_CODE
				printf("SET_VALUE *: ");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
#ifdef PRINT_CODE
					printf("%d %d\n",v1,v2);
#endif
					port[v2]=v1;
					switch(v2)
					{
					case YARD_OPEN:
						if(!map->check_rect((((int)(Pos.x+map->map_w_d))>>3)-(unit_manager.unit_type[type_id].FootprintX>>1),(((int)(Pos.z+map->map_h_d))>>3)-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx))
							port[v2]= (port[v2]==0.0f) ? 1.0f : 0.0f;
						break;
					case BUGGER_OFF:
						if(port[v2])	{
							int px=((int)(Pos.x)+map->map_w_d)>>3;
							int py=((int)(Pos.z)+map->map_h_d)>>3;
							for(int y=py-(unit_manager.unit_type[type_id].FootprintZ>>1);y<=py+(unit_manager.unit_type[type_id].FootprintZ>>1);y++)
								if(y>=0 && y<(map->bloc_h<<1)-1)
									for(int x=px-(unit_manager.unit_type[type_id].FootprintX>>1);x<=px+(unit_manager.unit_type[type_id].FootprintX>>1);x++)
										if(x>=0 && x<(map->bloc_w<<1)-1)
											if(map->map_data[y][x].unit_idx!=-1 && map->map_data[y][x].unit_idx!=idx) {
												int cur_idx=map->map_data[y][x].unit_idx;
												if(units.unit[cur_idx].owner_id==owner_id && units.unit[cur_idx].port[BUILD_PERCENT_LEFT]==0.0f && (units.unit[cur_idx].mission==NULL || units.unit[cur_idx].mission->mission!=MISSION_MOVE)) {
													VECTOR target=units.unit[cur_idx].Pos;
													target.z+=100.0f;
													PATH_NODE *path=NULL;
													if(!unit_manager.unit_type[units.unit[cur_idx].type_id].canfly) {
														float dh_max=unit_manager.unit_type[units.unit[cur_idx].type_id].MaxSlope*H_DIV;
														float h_min=map->sealvl-unit_manager.unit_type[units.unit[cur_idx].type_id].MaxWaterDepth*H_DIV;
														if(unit_manager.unit_type[units.unit[cur_idx].type_id].norestrict)
															h_min=0.0f;
														path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,dh_max,h_min,units.unit[cur_idx].Pos,target,unit_manager.unit_type[units.unit[cur_idx].type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[units.unit[cur_idx].type_id].FootprintX,unit_manager.unit_type[units.unit[cur_idx].type_id].FootprintZ,cur_idx);
														}
													else
														path=direct_path(target);
													sound_manager.play(unit_manager.unit_type[units.unit[cur_idx].type_id].ok1,units.unit[cur_idx].Pos);
													units.unit[cur_idx].set_mission(MISSION_MOVE,&target,false,0,true,NULL,path);
													}
												}
							}
						break;
					};
				}
				break;	//added
			case SCRIPT_ATTACH_UNIT:
				printf("ATTACH_UNIT *\n");
				{
					int v1 = script_env[id].pop();
					int v2 = script_env[id].pop();
					UNIT *target_unit = &(units.unit[v1]);
					target_unit->attached=true;
					link_list[nb_attached]=v2;
					attached_list[nb_attached++]=target_unit->idx;
				}
				pos++;
				break;	//added
			case SCRIPT_DROP_UNIT:
				printf("DROP_UNIT *\n");
				script_env[id].pop();
				pos++;
				break;	//added
			default:
				printf("UNKNOWN %d, STOPPING SCRIPT\n",script->script_code[script_id][--pos]);
				{
					script_val[script_id]=script_env[id].env->var[0];
					SCRIPT_ENV_STACK *old=script_env[id].env;
					script_env[id].env=script_env[id].env->next;
					delete old;
				}
				if(script_env[id].env) {
					script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
					pos=(script_env[id].env->cur>>8);
					}
				else
					script_env[id].running=false;
//				done=true;
			};
//		}while(!done);
		if(script_env[id].env)
			script_env[id].env->cur=script_id+(pos<<8);
	}

	inline float ballistic_angle(float v,float g,float d,float y_s,float y_e)			// Calculs de ballistique pour l'angle de tir
	{
		float v2=v*v;
		float gd=g*d;
		float v2gd = v2/gd;
		float a=v2gd*(4.0f*v2/gd-8.0f*(y_e-y_s)/d)-4.0f;
		if(a<0.0f)				// Pas de solution
			return 360.0f;
		return RAD2DEG*atan(v2gd-0.5f*sqrt(a));
 	}

	const int UNIT::move(const float dt,MAP *map,const int *path_exec,const int key_frame)
	{
		bool was_open = port[YARD_OPEN]!=0.0f ? true : false;
		if(unit_manager.unit_type[type_id].canmove)
			map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
		compute_coord=true;

		if(unit_manager.unit_type[type_id].isfeature) {		// Turn this unit into a feature
			int x=(int)((Pos.x)+map->map_w_d)>>3;
			int y=(int)((Pos.z)+map->map_h_d)>>3;
			if(x>0 && y>0 && x<(map->bloc_w<<1) && y<(map->bloc_h<<1))
				if(map->map_data[y][x].stuff==-1) {
					int type=feature_manager.get_feature_index(unit_manager.unit_type[type_id].Unitname);
					map->map_data[y][x].stuff=features.add_feature(Pos,type);
//					if(type!=-1 && (feature_manager.feature[type].height>5 || feature_manager.feature[type].m3d))
					if(type!=-1 && feature_manager.feature[type].blocking)
						map->rect(x-(feature_manager.feature[type].footprintx>>1),y-(feature_manager.feature[type].footprintz>>1),feature_manager.feature[type].footprintx,feature_manager.feature[type].footprintz,-2-map->map_data[y][x].stuff);
				}
			destroy();
			return -1;
			}

		if(map->ota_data.waterdoesdamage && Pos.y<map->sealvl)
			hp-=dt*map->ota_data.waterdamage;

		bool jump_commands = (((idx+key_frame)&0xF) == 0);		// Saute certaines commandes / Jump some commands so it runs faster with lots of units

		if(hp<=0.0f) {			// L'unité est détruite
			switch(flags&0xFD)
			{
			case 1:				// Début de la mort de l'unité	(Lance le script)
				flags=4;
				{
					int param[]={ severity };
					launch_script(get_script_index(SCRIPT_killed),1,param);
					int w_id=weapons.add_weapon(weapon_manager.get_weapon_index(unit_manager.unit_type[type_id].ExplodeAs),idx);
					b_delay[0]=1.0f;
					if(w_id>=0) {
						weapons.weapon[w_id].Pos=Pos;
						weapons.weapon[w_id].target_pos=Pos;
						weapons.weapon[w_id].target=-1;
						}
				}
				break;
			case 4:				// Vérifie si le script est terminé
				if(!is_running(get_script_index(SCRIPT_killed))) {
					if(b_delay[0]<=0.0f)
						return -1;
					b_delay[0]-=dt;
					for(int i=0;i<data.nb_piece;i++)
						if((data.flag[i]&FLAG_EXPLODE)!=FLAG_EXPLODE || (data.flag[i]&FLAG_EXPLODE && (data.explosion_flag[i]&EXPLODE_BITMAPONLY)))
							data.flag[i]|=FLAG_HIDE;
					}
				break;
			default:		// It doesn't explode (it has been reclaimed for example)
				flags=4;
				{
					int param[]={ severity };
					launch_script(get_script_index(SCRIPT_killed),1,param);
					b_delay[0]=0.0f;
				}
			};
			if(data.nb_piece>0 && port[BUILD_PERCENT_LEFT]==0.0f) {
				data.move(dt,map->ota_data.gravity);
				if(c_time>=0.1f) {
					c_time=0.0f;
					POINTF O;
					O.x=O.y=O.z=0.0f;
					for(int i=0;i<data.nb_piece;i++)
						if(data.flag[i]&FLAG_EXPLODE && (data.explosion_flag[i]&EXPLODE_BITMAPONLY)!=EXPLODE_BITMAPONLY) {
							if(data.explosion_flag[i]&EXPLODE_FIRE) {
								compute_model_coord();
								particle_engine.make_smoke(O+Pos+data.pos[i],fire,1,0.0f,0.0f);
								}
							if(data.explosion_flag[i]&EXPLODE_SMOKE) {
								compute_model_coord();
								particle_engine.make_smoke(O+Pos+data.pos[i],0,1,0.0f,0.0f);
								}
							}
					}
				}
			goto script_exec;
			}
		else if(!jump_commands && do_nothing())
			if(Pos.x<-map->map_w_d || Pos.x>map->map_w_d || Pos.z<-map->map_h_d || Pos.z>map->map_h_d) {
				VECTOR target;
				target=Pos;
				if(target.x<-map->map_w_d+256)
					target.x=-map->map_w_d+256;
				else if(target.x>map->map_w_d-256)
					target.x=map->map_w_d-256;
				if(target.z<-map->map_h_d+256)
					target.z=-map->map_h_d+256;
				else if(target.z>map->map_h_d-256)
					target.z=map->map_h_d-256;
				add_mission(MISSION_MOVE_CALC_PATH,&target,true,1,NULL,NULL);		// On reste sur l'écran
				}

		flags&=0xEF;		// To fix a bug

		if(port[BUILD_PERCENT_LEFT]>0.0f)	{		// L'unité n'est pas finie
			if(!built)
				hp-=dt*0.01f*unit_manager.unit_type[type_id].MaxDamage;
			goto script_exec;
			}

		if(data.nb_piece>0)
			data.move(dt,units.g_dt);

		if(attached)
			goto script_exec;

		if(unit_manager.unit_type[type_id].canload && nb_attached>0) {
			int e=0;
			compute_model_coord();
			for(int i=0;i+e<nb_attached;i++) {
				if(units.unit[attached_list[i]].flags) {
					units.unit[attached_list[i]].Pos=Pos+data.pos[link_list[i]];
					units.unit[attached_list[i]].Pos.y-=units.unit[attached_list[i]].model->size2;
					units.unit[attached_list[i]].Angle=Angle;
					}
				else {
					e++;
					i--;
					continue;
					}
				attached_list[i]=attached_list[i+e];
				}
			nb_attached-=e;
			}

		if(planned_weapons>0.0f) {			// Construit des armes / build weapons
			float old=planned_weapons-(int)planned_weapons;
			int idx=-1;
			if(unit_manager.unit_type[type_id].weapon[0] && unit_manager.unit_type[type_id].weapon[0]->stockpile)	idx=0;
			else if(unit_manager.unit_type[type_id].weapon[1] && unit_manager.unit_type[type_id].weapon[1]->stockpile)	idx=0;
			else if(unit_manager.unit_type[type_id].weapon[2] && unit_manager.unit_type[type_id].weapon[2]->stockpile)	idx=0;
			if(idx!=-1 && unit_manager.unit_type[type_id].weapon[idx]->reloadtime!=0.0f)	{
				float dn=dt/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				float conso_metal=((float)unit_manager.unit_type[type_id].weapon[idx]->metalpershot)/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				float conso_energy=((float)unit_manager.unit_type[type_id].weapon[idx]->energypershot)/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
					players.metal_u[owner_id]+=conso_metal;
					players.energy_u[owner_id]+=conso_energy;
					planned_weapons-=dn;
					float last=planned_weapons-(int)planned_weapons;
					if((last==0.0f && last!=old) || (last>old && old>0.0f) || planned_weapons<=0.0f)		// On en a fini une / one is made
						stock[idx]++;
					if(planned_weapons<0.0f)
						planned_weapons=0.0f;
					}
				}
			}

		V_Angle.x=V_Angle.y=V_Angle.z=0.0f;
		c_time+=dt;
		if(mission) {
				mission->time+=dt;
				switch(mission->mission)
				{
				case MISSION_ATTACK:
				case MISSION_AIM:
				case MISSION_SHOOT:
					b_delay[0]+=dt;
					b_delay[1]+=dt;
					b_delay[2]+=dt;
					break;
				};
				switch(mission->mission)						// Commandes générales
				{
				case MISSION_LOAD:
					if(mission->p!=NULL) {
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags==0) {
							next_mission();
							break;
							}
						VECTOR Dir=target_unit->Pos-Pos;
						Dir.y=0.0f;
						mission->target=target_unit->Pos;
						float dist=Dir.Sq();
						int maxdist=0;
						if(unit_manager.unit_type[type_id].TransMaxUnits==1)		// Code for units like the arm atlas
							maxdist=3;
						else
							maxdist=unit_manager.unit_type[type_id].SightDistance;
						if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
							VECTOR target;
							dist=sqrt(dist);
							target=mission->target;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
							}
						else {
							if(mission->last_d>=0.0f) {
								if(unit_manager.unit_type[type_id].TransMaxUnits==1) {		// Code for units like the arm atlas
									if(nb_attached==0) {
										int param[]={ (int)(target_unit->Pos.y+target_unit->model->size2) };
										launch_script(get_script_index(SCRIPT_BeginTransport),1,param);
										target_unit->attached=true;
										link_list[nb_attached]=0;
										attached_list[nb_attached++]=target_unit->idx;
										}
									next_mission();
									}
								else {
									if(nb_attached>=unit_manager.unit_type[type_id].TransMaxUnits) {
										next_mission();
										break;
										}
									int param[]= { target_unit->idx };
									launch_script(get_script_index(SCRIPT_TransportPickup),1,param);
									}
								mission->last_d=-1.0f;
								}
							else {
								if(!is_running(get_script_index(SCRIPT_TransportPickup)))
									next_mission();
								}
							}
						}
					break;
				case MISSION_RECLAIM:
					if(mission->p!=NULL)	{		// Récupère une unité
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags && target_unit->hp<=0.0f) {		// Travail terminé
							launch_script(get_script_index(SCRIPT_stopbuilding));
							launch_script(get_script_index(SCRIPT_stop));
							target_unit->flags|=0x10;			// This unit is being reclaimed it doesn't explode!
							next_mission();
							}
						else if(target_unit->flags) {
							VECTOR Dir=target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(mission->last_d>=0.0f) {
									launch_script(get_script_index(SCRIPT_startbuilding));
									launch_script(get_script_index(SCRIPT_go));
									mission->last_d=-1.0f;
									}
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else {			// Récupère l'unité
									float recup=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].HealTime;
									if(recup>target_unit->hp)	recup=target_unit->hp;
									target_unit->hp-=recup;
									target_unit->flags|=0x10;			// This unit is being reclaimed it doesn't explode!
									if(dt>0.0f)
										players.metal_t[owner_id]+=recup/dt*unit_manager.unit_type[target_unit->type_id].BuildCostMetal/unit_manager.unit_type[target_unit->type_id].MaxDamage;
									}
								}
							}
						}
					else if(mission->data>=0 && mission->data<features.nb_features && features.feature[mission->data].type>=0)	{	// Récupère un élément du décors/une carcasse
						if(features.feature[mission->data].hp<=0.0f) {		// Travail terminé
							int x=((int)(features.feature[mission->data].Pos.x)+map->map_w_d-4)>>3;		// Efface l'objet
							int y=((int)(features.feature[mission->data].Pos.z)+map->map_h_d-4)>>3;
							map->rect(x-(feature_manager.feature[features.feature[mission->data].type].footprintx>>1),y-(feature_manager.feature[features.feature[mission->data].type].footprintz>>1),feature_manager.feature[features.feature[mission->data].type].footprintx,feature_manager.feature[features.feature[mission->data].type].footprintz,-1);
							features.delete_feature(mission->data);			// Supprime l'objet
							launch_script(get_script_index(SCRIPT_stopbuilding));
							launch_script(get_script_index(SCRIPT_stop));
							next_mission();
							}
						else {
							VECTOR Dir=features.feature[mission->data].Pos-Pos;
							Dir.y=0.0f;
							mission->target=features.feature[mission->data].Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(mission->last_d>=0.0f) {
									launch_script(get_script_index(SCRIPT_startbuilding));
									launch_script(get_script_index(SCRIPT_go));
									mission->last_d=-1.0f;
									}
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else {			// Récupère l'objet
									float recup=dt*unit_manager.unit_type[type_id].WorkerTime;
									if(recup>features.feature[mission->data].hp)	recup=features.feature[mission->data].hp;
									features.feature[mission->data].hp-=recup;
									if(dt>0.0f) {
										players.metal_t[owner_id]+=recup/dt*feature_manager.feature[features.feature[mission->data].type].metal/feature_manager.feature[features.feature[mission->data].type].damage;
										players.energy_t[owner_id]+=recup/dt*feature_manager.feature[features.feature[mission->data].type].energy/feature_manager.feature[features.feature[mission->data].type].damage;
										}
									}
								}
							}
						}
					else
						next_mission();
					break;
				case MISSION_GUARD:
					if(jump_commands)	break;
					if(mission->p!=NULL && ((UNIT*)mission->p)->flags && ((UNIT*)mission->p)->owner_id==owner_id) {		// On ne défend pas n'importe quoi
						if(unit_manager.unit_type[type_id].Builder) {
							if(((UNIT*)mission->p)->port[BUILD_PERCENT_LEFT]>0.0f || ((UNIT*)mission->p)->hp<unit_manager.unit_type[((UNIT*)mission->p)->type_id].MaxDamage) {		// Répare l'unité
								add_mission(MISSION_REPAIR,&((UNIT*)mission->p)->Pos,true,0,((UNIT*)mission->p),NULL);
								break;
								}
							else if(((UNIT*)mission->p)->mission!=NULL && (((UNIT*)mission->p)->mission->mission==MISSION_BUILD_2 || ((UNIT*)mission->p)->mission->mission==MISSION_REPAIR)) {		// L'aide à construire
								add_mission(MISSION_REPAIR,&((UNIT*)mission->p)->mission->target,true,0,((UNIT*)mission->p)->mission->p,NULL);
								break;
								}
							}
						if(unit_manager.unit_type[type_id].canattack) {
							if(((UNIT*)mission->p)->mission!=NULL && (((UNIT*)mission->p)->mission->mission==MISSION_SHOOT || ((UNIT*)mission->p)->mission->mission==MISSION_AIM || ((UNIT*)mission->p)->mission->mission==MISSION_ATTACK)) {		// L'aide à construire
								add_mission(MISSION_ATTACK,&((UNIT*)mission->p)->mission->target,true,0,((UNIT*)mission->p)->mission->p,NULL);
								break;
								}
							}
						if(((VECTOR)(Pos-((UNIT*)mission->p)->Pos)).Sq()>=25600.0f) {			// On reste assez près
							VECTOR target=((UNIT*)mission->p)->Pos;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,10,NULL,NULL);
							break;
							}
						}
					else
						next_mission();
					break;
				case MISSION_MOVE_CALC_PATH:
					if(jump_commands)	break;
					if(*path_exec<MAX_PATH_EXEC)
					{
						*path_exec++;
						if(unit_manager.unit_type[type_id].canfly) {
							if(mission->data<=0)
								mission->path=direct_path(mission->target);
							else {
								VECTOR Dir=mission->target-Pos;
								Dir.Unit();
								mission->path=direct_path(mission->target-(mission->data<<3)*Dir);
								}
							}
						else {
							float dh_max=unit_manager.unit_type[type_id].MaxSlope*H_DIV;
							float h_min=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
							if(unit_manager.unit_type[type_id].norestrict)
								h_min=0.0f;
							if(mission->data<=0)
								mission->path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
														dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx);
							else
								mission->path=compute_path_min_dist(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
														dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,mission->data);
							}
						c_time=0.0f;
						mission->mission=MISSION_MOVE;
					}
					break;
				case MISSION_PATROL:					// Mode patrouille
					{
						add_mission(MISSION_MOVE_CALC_PATH,&Pos,false,0,NULL,NULL,1);			// Retour à la case départ après l'éxécution de tous les ordres / back to beginning
						add_mission(MISSION_PATROL,&mission->target,false,0,NULL,NULL,1);		// Retour à la case départ après l'éxécution de tous les ordres / restart patrol order

						if(unit_manager.unit_type[type_id].canfly)
							mission->path=direct_path(mission->target);
						else {
							float dh_max=unit_manager.unit_type[type_id].MaxSlope*H_DIV;
							float h_min=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
							if(unit_manager.unit_type[type_id].norestrict)
								h_min=0.0f;
								mission->path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
															dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx);
							}
						c_time=0.0f;
						mission->mission=MISSION_MOVE;
						mission->flags=1;
					}
					break;
				case MISSION_STANDBY:
				case MISSION_VTOL_STANDBY:
					if(jump_commands)	break;
					if(mission->data>5) {
						if(mission->next)		// If there is a mission after this one
							next_mission();
						}
					else
						mission->data++;
					break;
				case MISSION_ATTACK:										// Attaque une unité / attack a unit
					if(jump_commands)	break;
					{
						UNIT *target_unit=(mission->flags&4)==4 ? NULL : (UNIT*) mission->p;
						WEAPON *target_weapon=(mission->flags&4)==4 ? (WEAPON*) mission->p : NULL;
						if((target_unit!=NULL && target_unit->flags) || (target_weapon!=NULL && target_weapon->weapon_id!=-1)) {
							VECTOR Dir=target_unit==NULL ? target_weapon->Pos-Pos : target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit==NULL ? target_weapon->Pos : target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=0;
							int mindist=0;
							if(mission->data<0)	mission->data=0;
							else mission->data%=3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;

							if(unit_manager.unit_type[type_id].weapon[mission->data]) {
								if(Dir%V<0.0f && !unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight
								&& !(unit_manager.unit_type[type_id].weapon[mission->data]->rendertype==RENDER_TYPE_LASER || unit_manager.unit_type[type_id].weapon[mission->data]->rendertype==RENDER_TYPE_LIGHTNING)
								&& unit_manager.unit_type[type_id].canfly) {
									mission->data++;
									break;		// Ne tire pas si on s'éloigne de la cible avec une arme non "laser" / dont fire if moving away from the target with a non laser weapon
									}
								if(unit_manager.unit_type[type_id].attackrunlength>0) {
									float t=sqrt(2.0f/map->ota_data.gravity*fabs(Pos.y-mission->target.y));
									mindist=(int)(t*V.Norm())-(unit_manager.unit_type[type_id].attackrunlength+2>>2);
									maxdist=mindist+(unit_manager.unit_type[type_id].attackrunlength+1>>1);
									}
								else
									maxdist=unit_manager.unit_type[type_id].weapon[mission->data]->range>>1;
								if(burst[mission->data]==0 && b_delay[mission->data]<unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime
								&& !unit_manager.unit_type[type_id].weapon[mission->data]->stockpile)	 {
									mission->data++;
									break;
									}
								}
							else {
								next_mission();
								break;
								}

							if(unit_manager.unit_type[type_id].weapon[mission->data]->commandfire && (mission->flags&8)==8) {		// Don't let it fire with a weapon if it is not allowed to
								mission->data++;
								break;
								}

							if(unit_manager.unit_type[type_id].weapon[mission->data]->stockpile && stock[mission->data]<=0) {	// We have nothing to fire / rien à tirer
								mission->data=++mission->data%3;
								break;
								}

							if(dist>maxdist*maxdist || dist<mindist*mindist) {	// Si l'unité est trop loin de sa cible / if unit isn't where it should be
								if(!unit_manager.unit_type[type_id].canmove) {		// Bah là si on peut pas bouger faut changer de cible!! / need to change target
									next_mission();
									break;
									}
								else if(!unit_manager.unit_type[type_id].canfly) {
									VECTOR target;
									target=mission->target;
									c_time=0.0f;
									add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher / tell it to go closer
									}
								}
							else {
								int param[] = { (int)(unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime*1000.0f) };
								launch_script(get_script_index(SCRIPT_SetMaxReloadTime),1,param);
								switch(mission->data)
								{
								case 0:
									launch_script(get_script_index(SCRIPT_QueryPrimary));		break;
								case 1:
									launch_script(get_script_index(SCRIPT_QuerySecondary));		break;
								case 2:
									launch_script(get_script_index(SCRIPT_QueryTertiary));		break;
								};
								mission->mission=MISSION_AIM;							// sinon on lui demande de viser / or make it aim
								}
							}
						else {
							launch_script(get_script_index(SCRIPT_TargetCleared));
							next_mission();
							}
					}
					break;
				case MISSION_AIM:											// Vise une unité / aiming code
					if(jump_commands)	break;
					if(((mission->flags&4)==4 && ((WEAPON*)(mission->p))->weapon_id!=-1) || ((mission->flags&4)!=4 && ((UNIT*)(mission->p))->flags)) {
						int query_id=-1;
						switch(mission->data)
						{
						case 0:
							query_id=get_script_index(SCRIPT_QueryPrimary);		break;
						case 1:
							query_id=get_script_index(SCRIPT_QuerySecondary);	break;
						case 2:
							query_id=get_script_index(SCRIPT_QueryTertiary);	break;
						};
						if(!is_running(query_id) && !is_running(get_script_index(SCRIPT_RequestState))) {
							if(b_delay[mission->data]>=unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime || unit_manager.unit_type[type_id].weapon[mission->data]->stockpile) {
								if(unit_manager.unit_type[type_id].weapon[mission->data]->turret) {			// Si l'unité doit viser, on la fait viser / if it must aim, we make it aim
									UNIT *target_unit=(mission->flags&4)==4 ? NULL : (UNIT*) mission->p;
									WEAPON *target_weapon=(mission->flags&4)==4 ? (WEAPON*) mission->p : NULL;
									int start_piece=script_val[query_id];
									if(start_piece<0 || start_piece>=data.nb_piece)
										start_piece=0;
									VECTOR target=target_unit==NULL ? target_weapon->Pos-(Pos+data.pos[start_piece]) : target_unit->Pos+target_unit->model->center-(Pos+data.pos[start_piece]);
									float dist=target.Norm();
									target=(1.0f/dist)*target;
									VECTOR I,J,IJ,RT;
									I.x=0.0f;	I.z=1.0f;	I.y=0.0f;
									J.x=1.0f;	J.z=0.0f;	J.y=0.0f;
									IJ.x=0.0f;	IJ.z=0.0f;	IJ.y=1.0f;
									RT=target;
									RT.y=0.0f;
									RT.Unit();
									float angle=acos(I%RT)*RAD2DEG;
									if(J%RT<0.0f) angle=-angle;
									angle-=Angle.y;
									if(angle<-180.0f)	angle+=360.0f;
									else if(angle>180.0f)	angle-=360.0f;
									int aiming[]={ (int)(angle*DEG2TA), -4096 };
									if(unit_manager.unit_type[type_id].weapon[mission->data]->ballistic) {		// Calculs de ballistique / ballistic calculations
										VECTOR D=target_unit==NULL ? (Pos+data.pos[start_piece]-target_weapon->Pos) : (Pos+data.pos[start_piece]-target_unit->Pos-target_unit->model->center);
										D.y=0.0f;
										float v;
										if(unit_manager.unit_type[type_id].weapon[mission->data]->startvelocity==0.0f)
											v=unit_manager.unit_type[type_id].weapon[mission->data]->weaponvelocity;
										else
											v=unit_manager.unit_type[type_id].weapon[mission->data]->startvelocity;
										if(target_unit==NULL)
											aiming[1]=(int)(ballistic_angle(v,map->ota_data.gravity,D.Norm(),(Pos+data.pos[start_piece]).y,target_weapon->Pos.y)*DEG2TA);
										else
											aiming[1]=(int)(ballistic_angle(v,map->ota_data.gravity,D.Norm(),(Pos+data.pos[start_piece]).y,target_unit->Pos.y+target_unit->model->center.y)*DEG2TA);
										}
									else {
										VECTOR K=target;
										K.y=0.0f;
										K.Unit();
										angle=acos(K%target)*RAD2DEG;
										if(target.y>0.0f)
											angle=-angle;
										if(angle>180.0f)	angle-=360.0f;
										if(angle<-180.0f)	angle+=360.0f;
										if(fabs(angle)>180.0f) {
											next_mission();
											break;
											}
										aiming[1]=(int)(angle*DEG2TA);
										}
									if(unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight) {
										if(target_unit==NULL)
											aim_dir=((WEAPON*)(mission->p))->Pos-(Pos+data.pos[start_piece]);
										else
											aim_dir=((UNIT*)(mission->p))->Pos+((UNIT*)(mission->p))->model->center-(Pos+data.pos[start_piece]);
										aim_dir.Unit();
										}
									else
										aim_dir=cos(aiming[1]*TA2RAD)*(cos(aiming[0]*TA2RAD+Angle.y*DEG2RAD)*I+sin(aiming[0]*TA2RAD+Angle.y*DEG2RAD)*J)+sin(aiming[1]*TA2RAD)*IJ;
									switch(mission->data)
									{
									case 0:
										launch_script(get_script_index(SCRIPT_AimPrimary),2,aiming);	break;
									case 1:
										launch_script(get_script_index(SCRIPT_AimSecondary),2,aiming);	break;
									case 2:
										launch_script(get_script_index(SCRIPT_AimTertiary),2,aiming);	break;
									};
									}
								else
									switch(mission->data)
									{
									case 0:
										launch_script(get_script_index(SCRIPT_AimPrimary));	break;
									case 1:
										launch_script(get_script_index(SCRIPT_AimSecondary));	break;
									case 2:
										launch_script(get_script_index(SCRIPT_AimTertiary));	break;
									};
								mission->time=0.0f;
								mission->mission=MISSION_SHOOT;									// (puis) on lui demande de tirer / tell it to fire
								}
							}
						}
					else {
						launch_script(get_script_index(SCRIPT_TargetCleared));
						next_mission();
						}
					break;
				case MISSION_SHOOT:											// Tire sur une unité / fire!
					if(((mission->flags&4)==4 && ((WEAPON*)(mission->p))->weapon_id!=-1) || ((mission->flags&4)!=4 && ((UNIT*)(mission->p))->flags)) {
						if(burst[mission->data]>0 && b_delay[mission->data]<unit_manager.unit_type[type_id].weapon[mission->data]->burstrate)		break;
						int query_id=-1;
//						char *Aim_script=NULL;
//						char *Fire_script=NULL;
						int Aim_script = -1;
						int Fire_script = -1;
						switch(mission->data)
						{
						case 0:
							query_id=get_script_index(SCRIPT_QueryPrimary);
							Aim_script=SCRIPT_AimPrimary;
							Fire_script=SCRIPT_FirePrimary;
							break;
						case 1:
							query_id=get_script_index(SCRIPT_QuerySecondary);
							Aim_script=SCRIPT_AimSecondary;
							Fire_script=SCRIPT_FireSecondary;
							break;
						case 2:
							query_id=get_script_index(SCRIPT_QueryTertiary);
							Aim_script=SCRIPT_AimTertiary;
							Fire_script=SCRIPT_FireTertiary;
							break;
						};
						if(!is_running(get_script_index(Aim_script))) {
							if((players.metal[owner_id]<unit_manager.unit_type[type_id].weapon[mission->data]->metalpershot
							|| players.energy[owner_id]<unit_manager.unit_type[type_id].weapon[mission->data]->energypershot)
							&& !unit_manager.unit_type[type_id].weapon[mission->data]->stockpile) {
								mission->mission=MISSION_ATTACK;		// Pas assez d'énergie pour tirer / not enough energy to fire
								break;
								}
							if(unit_manager.unit_type[type_id].weapon[mission->data]->stockpile && stock[mission->data]<=0) {
								mission->mission=MISSION_ATTACK;		// Plus rien pour tirer / nothing to fire
								break;
								}
							else
								stock[mission->data]--;
							players.metal[owner_id]-=unit_manager.unit_type[type_id].weapon[mission->data]->metalpershot;
							players.energy[owner_id]-=unit_manager.unit_type[type_id].weapon[mission->data]->energypershot;
							launch_script(get_script_index(Fire_script));
							sound_manager.play(unit_manager.unit_type[type_id].weapon[mission->data]->soundstart,Pos);
							int start_piece=script_val[query_id];
							if(start_piece>=0 && start_piece<data.nb_piece) {
								compute_model_coord();
								VECTOR Dir=data.dir[start_piece];
								if(Dir.x==0.0f && Dir.y==0.0f && Dir.z==0.0f) {
									if(unit_manager.unit_type[type_id].weapon[mission->data]->vlaunch) {
										Dir.x=0.0f;
										Dir.y=1.0f;
										Dir.z=0.0f;
										}
									else {
										Dir=aim_dir;
										if(unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight && !unit_manager.unit_type[type_id].weapon[mission->data]->turret) {
											if(mission->flags&4)
												Dir=((WEAPON*)(mission->p))->Pos-(Pos+data.pos[start_piece]);
											else
												Dir=((UNIT*)(mission->p))->Pos+((UNIT*)(mission->p))->model->center-(Pos+data.pos[start_piece]);
											Dir.Unit();
											}
										}
									}
								if(mission->flags&4)
									shoot(((WEAPON*)(mission->p))->idx,Pos+data.pos[start_piece],Dir,mission->data);
								else
									shoot(((UNIT*)(mission->p))->idx,Pos+data.pos[start_piece],Dir,mission->data);
								burst[mission->data]++;
								if(burst[mission->data]>=unit_manager.unit_type[type_id].weapon[mission->data]->burst)
									burst[mission->data]=0;
								b_delay[mission->data]=0.0f;
								}
							if((mission->flags&4)!=4 && ((UNIT*)(mission->p))->hp>0) {				// La cible est-elle détruite ?? / is target destroyed ??
								if(burst[mission->data]==0) {
									mission->mission=MISSION_ATTACK;
									mission->data++;
									mission->time=0.0f;
									}
								}
							else {
								launch_script(get_script_index(SCRIPT_TargetCleared));
								next_mission();
								}
							}
						else if(mission->time>0.25f && unit_manager.unit_type[type_id].weapon[mission->data]->turret) {
							mission->mission=MISSION_AIM;
//							kill_script(get_script_index(Aim_script));
							}
						}
					else {
						launch_script(get_script_index(SCRIPT_TargetCleared));
						next_mission();
						}
					break;
				case MISSION_STOP:											// Arrête tout ce qui était en cours / stop everything running
					if(jump_commands)	break;
					if(mission->data>5) {
						if(mission->next) {
							next_mission();
							if(mission!=NULL && mission->mission==MISSION_STOP)		// Mode attente / wait mode
								mission->data=6;
							}
						}
					else {
						if(mission->data==0) {
							launch_script(get_script_index(SCRIPT_StopMoving));		// Arrête tout / stop everything
							launch_script(get_script_index(SCRIPT_stopbuilding));
							}
						mission->data++;
						}
					break;
				case MISSION_REPAIR:
					{
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit!=NULL && target_unit->flags && target_unit->port[BUILD_PERCENT_LEFT]==0.0f) {
							if(target_unit->hp>=unit_manager.unit_type[target_unit->type_id].MaxDamage || unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {
								if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
									target_unit->hp=unit_manager.unit_type[target_unit->type_id].MaxDamage;
								next_mission();
								}
							else {
								VECTOR Dir=target_unit->Pos-Pos;
								Dir.y=0.0f;
								mission->target=target_unit->Pos;
								float dist=Dir.Sq();
								int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
								if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
									VECTOR target;
//									dist=sqrt(dist);
									target=mission->target;
									add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
									mission->data=0;
									c_time=0.0f;
									}
								else {
									if(mission->data==0) {
										mission->data=1;
										launch_script(get_script_index(SCRIPT_startbuilding));
										launch_script(get_script_index(SCRIPT_go));
										}
									if(dist!=0.0f)
										Dir=1.0f/sqrt(dist)*Dir;
									float dest_angle=acos(Dir.z)*RAD2DEG;
									if(Dir.x<0.0f) dest_angle=-dest_angle;
									while(fabs(dest_angle-Angle.y)>180.0f) {
										if(dest_angle<Angle.y)
											Angle.y-=360.0f;
										else
											Angle.y+=360.0f;
										}
									if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle) {
										float aspeed=unit_manager.unit_type[type_id].TurnRate;
										if(dest_angle<Angle.y)
											aspeed=-aspeed;
										float a=dest_angle-Angle.y;
										V_Angle.y=aspeed*0.1f;
										float b=dest_angle-(Angle.y+dt*V_Angle.y);
										if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
											V_Angle.y=0.0f;
											Angle.y=dest_angle;
											}
										}
									else {
										float conso_metal=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostMetal))/unit_manager.unit_type[target_unit->type_id].HealTime;
										float conso_energy=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostEnergy))/unit_manager.unit_type[target_unit->type_id].HealTime;
										if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
											players.metal_u[owner_id]+=conso_metal;
											players.energy_u[owner_id]+=conso_energy;
											target_unit->hp+=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].HealTime;
											}
										target_unit->built=true;
										}
									}
								}
							}
						else if(target_unit!=NULL && target_unit->flags) {
							VECTOR Dir=target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else
									mission->mission=MISSION_BUILD_2;		// Change de type de mission
								}
							}
					}
					break;
				case MISSION_BUILD_2:
					{
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags) {
							if(target_unit->port[BUILD_PERCENT_LEFT]<=0.0f) {
								target_unit->port[BUILD_PERCENT_LEFT]=0.0f;
								target_unit->launch_script(target_unit->get_script_index(SCRIPT_create));		// Script post création
								if(unit_manager.unit_type[target_unit->type_id].ActivateWhenBuilt)		// Activation automatique
									target_unit->activate();
								if(unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {		// Ordre de se déplacer
									VECTOR target=Pos;
									target.z+=128.0f;
									target_unit->set_mission(MISSION_MOVE_CALC_PATH,&target,false,5,true,NULL,NULL);		// Fait sortir l'unité du bâtiment
									}
								mission->p=NULL;
								next_mission();
								deactivate();
								}
							else {
								float conso_metal=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostMetal))/unit_manager.unit_type[target_unit->type_id].BuildTime;
								float conso_energy=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostEnergy))/unit_manager.unit_type[target_unit->type_id].BuildTime;
								if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
									players.metal_u[owner_id]+=conso_metal;
									players.energy_u[owner_id]+=conso_energy;
									target_unit->port[BUILD_PERCENT_LEFT]-=dt*unit_manager.unit_type[type_id].WorkerTime*100.0f/unit_manager.unit_type[target_unit->type_id].BuildTime;
									target_unit->hp+=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].BuildTime;
									}
								if(unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {
									int script_id_buildinfo = get_script_index(SCRIPT_QueryBuildInfo);
									if(script_id_buildinfo>=0) {
										compute_model_coord();
										map->rect(target_unit->cur_px-(unit_manager.unit_type[target_unit->type_id].FootprintX>>1),target_unit->cur_py-(unit_manager.unit_type[target_unit->type_id].FootprintZ>>1),unit_manager.unit_type[target_unit->type_id].FootprintX,unit_manager.unit_type[target_unit->type_id].FootprintZ,-1,unit_manager.unit_type[target_unit->type_id].yardmap,target_unit->port[YARD_OPEN]!=0.0f);
										target_unit->Pos=Pos+data.pos[script_val[script_id_buildinfo]];
										target_unit->Angle=Angle;
										target_unit->Angle.y+=data.axe[1][script_val[script_id_buildinfo]].angle;
										}
									}
								target_unit->built=true;
								}
							}
						else
							next_mission();
						}
					break;
				case MISSION_BUILD:
					if(mission->p) {
						launch_script(get_script_index(SCRIPT_startbuilding));
						mission->mission=MISSION_BUILD_2;		// Change les objectifs
						((UNIT*)(mission->p))->built=true;
						}
					else {
						VECTOR Dir=mission->target-Pos;
						Dir.y=0.0f;
						float dist=Dir.Sq();
						int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
						if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
							VECTOR target;
							target=mission->target;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
							}
						else {
							if(dist!=0.0f)
								Dir=(1.0f/sqrt(dist))*Dir;
							float dest_angle=acos(Dir.z)*RAD2DEG;
							if(Dir.x<0.0f) dest_angle=-dest_angle;
							while(fabs(dest_angle-Angle.y)>180.0f) {
								if(dest_angle<Angle.y)
									Angle.y-=360.0f;
								else
									Angle.y+=360.0f;
								}
							if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
								float aspeed=unit_manager.unit_type[type_id].TurnRate;
								if(dest_angle<Angle.y)
									aspeed=-aspeed;
								float a=dest_angle-Angle.y;
								V_Angle.y=aspeed*0.1f;
								float b=dest_angle-(Angle.y+dt*V_Angle.y);
								if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
									V_Angle.y=0.0f;
									Angle.y=dest_angle;
									}
								}
							else {
								if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT || (unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT && port[YARD_OPEN])) {
									if(map->check_rect((((int)(mission->target.x)+map->map_w_d)>>3)-(unit_manager.unit_type[mission->data].FootprintX>>1),(((int)(mission->target.z)+map->map_h_d)>>3)-(unit_manager.unit_type[mission->data].FootprintZ>>1),unit_manager.unit_type[mission->data].FootprintX,unit_manager.unit_type[mission->data].FootprintZ,-1)) {				// Vérifie s'il y a la place de construire l'unité
										mission->p=create_unit(mission->data,owner_id,mission->target,map);
										if(mission->p) {
											((UNIT*)(mission->p))->hp=0.000001f;
											((UNIT*)(mission->p))->built=true;
											}
										}
									else if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
										next_mission();
									}
								else {
									launch_script(get_script_index(SCRIPT_go));
									launch_script(get_script_index(SCRIPT_QueryBuildInfo));
									}
								}
							}
						}
					break;
				};

			switch(unit_manager.unit_type[type_id].TEDclass)			// Commandes particulières
			{
			case CLASS_SHIP:
				switch(mission->mission)
				{
				case MISSION_STANDBY:
					if(mission->next)
						next_mission();
					V=units.exp_dt_1*V;			// Frottements
					break;
				case MISSION_MOVE:
					{
						VECTOR J,I;
						VECTOR Target=mission->target;
						if(mission->path)
							Target=mission->path->Pos;
						J=Target-Pos;
						J.y=0.0f;
						float dist=J.Norm();
						if((dist>mission->last_d && dist<15.0f) || mission->path==NULL) {		// Là on est arrivé
							bool vide=(mission->path==NULL);
							if(mission->path)
								mission->path=next_node(mission->path);
							mission->last_d=9999999.0f;
							if(mission->path==NULL) {
								if(owner_id==players.local_human_id && !vide)
									sound_manager.play(unit_manager.unit_type[type_id].arrived1,Pos);
								launch_script(get_script_index(SCRIPT_StopMoving));
								next_mission();
								break;
								}
							}
						else
							mission->last_d=dist;
						if(dist>0.0f)
							J=1.0f/dist*J;
						float dest_angle=acos(J.z)*RAD2DEG;
						if(J.x<0.0f) dest_angle=-dest_angle;
						while(fabs(dest_angle-Angle.y)>180.0f) {
							if(dest_angle<Angle.y)
								Angle.y-=360.0f;
							else
								Angle.y+=360.0f;
							}
						if(dist>0.0f && fabs(dest_angle-Angle.y)>=1.0f) {
							float aspeed=unit_manager.unit_type[type_id].TurnRate;
							if(dest_angle<Angle.y)
								aspeed=-aspeed;
							float a=dest_angle-Angle.y;
							V_Angle.y=aspeed*0.1f;
							float b=dest_angle-(Angle.y+dt*V_Angle.y);
							if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
								V_Angle.y=0.0f;
								Angle.y=dest_angle;
								}
							}

						J.z=cos(Angle.y*DEG2RAD);
						J.x=sin(Angle.y*DEG2RAD);
						J.y=0.0f;
						I.z=-J.x;
						I.x=J.z;
						I.y=0.0f;
						V=(V%J)*J+units.exp_dt_1*(V%I)*I;
						if(!(dist<128.0f && fabs(Angle.y-dest_angle)>=1)) {
							float speed=V.Norm();
							float time_to_stop=speed/unit_manager.unit_type[type_id].BrakeRate;
							float min_dist=time_to_stop*(speed-unit_manager.unit_type[type_id].BrakeRate*0.5f*time_to_stop);
							if(min_dist>=dist)			// Freine si nécessaire
								V=V-(unit_manager.unit_type[type_id].BrakeRate*dt)*J;
							else if(speed<unit_manager.unit_type[type_id].MaxVelocity)
								V=V+(unit_manager.unit_type[type_id].Acceleration*dt)*J;
							}
						else
							V=units.exp_dt_1*V;			// Frottements
					}
					break;
				default:
					V=units.exp_dt_1*V;			// Frottements
				};
				break;
			case CLASS_PLANT:
				switch(mission->mission)
				{
				case MISSION_STANDBY:
				case MISSION_BUILD:
				case MISSION_BUILD_2:
				case MISSION_REPAIR:
					break;
				default:
					next_mission();
				};
				break;
			case CLASS_WATER:
			case CLASS_VTOL:
			case CLASS_KBOT:
			case CLASS_COMMANDER:
			case CLASS_TANK:
			case CLASS_CNSTR:
				{
				VECTOR K,PV;
				K.x=0.0f;
				K.y=1.0f;
				K.z=0.0f;
				if(mission->mission!=MISSION_MOVE && ((mission->mission!=MISSION_ATTACK && mission->mission!=MISSION_AIM && mission->mission!=MISSION_SHOOT && unit_manager.unit_type[type_id].canfly) || !unit_manager.unit_type[type_id].canfly)) {
					PV=(V%K)*K;
					V=units.exp_dt_2*(V-PV)+PV;		// N'altère pas la déscente des unités
					}
				switch(mission->mission)
				{
				case MISSION_ATTACK:
				case MISSION_AIM:
				case MISSION_SHOOT:
				case MISSION_PATROL:
				case MISSION_REPAIR:
				case MISSION_BUILD:
				case MISSION_BUILD_2:
					if(unit_manager.unit_type[type_id].canfly)
						activate();
					break;
				case MISSION_MOVE:
					{
						if(unit_manager.unit_type[type_id].canfly)
							activate();
						VECTOR J,I;
						VECTOR Target=mission->target;
						if(mission->path)
							Target=mission->path->Pos;
						J=Target-Pos;
						J.y=0.0f;
						float dist=J.Norm();
						if((dist>mission->last_d && dist<15.0f) || mission->path==NULL) {
							if(mission->path)
								mission->path=next_node(mission->path);
							mission->last_d=9999999.0f;
							if(mission->path==NULL) {		// Si l'on est arrivé au bout du chemin
								sound_manager.play(unit_manager.unit_type[type_id].arrived1,Pos);
								launch_script(get_script_index(SCRIPT_StopMoving));
								next_mission();
								break;
								}
							}
						else
							mission->last_d=dist;
						if(dist>0.0f)
							J=1.0f/dist*J;
						float dest_angle=acos(J.z)*RAD2DEG;
						if(J.x<0.0f) dest_angle=-dest_angle;
						while(fabs(dest_angle-Angle.y)>180.0f) {
							if(dest_angle<Angle.y)
								Angle.y-=360.0f;
							else
								Angle.y+=360.0f;
							}
						if(dist>0.0f && fabs(dest_angle-Angle.y)>=1.0f) {
							float aspeed=unit_manager.unit_type[type_id].TurnRate;
							if(dest_angle<Angle.y)
								aspeed=-aspeed;
							float a=dest_angle-Angle.y;
							V_Angle.y=aspeed*0.1f;
							float b=dest_angle-(Angle.y+dt*V_Angle.y);
							if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
								V_Angle.y=0.0f;
								Angle.y=dest_angle;
								}
							}

						J.z=cos(Angle.y*DEG2RAD);
						J.x=sin(Angle.y*DEG2RAD);
						J.y=0.0f;
						I.z=-J.x;
						I.x=J.z;
						I.y=0.0f;
						V=(V%K)*K+(V%J)*J+units.exp_dt_4*(V%I)*I;
						if(!(dist<128.0f && fabs(Angle.y-dest_angle)>=1)) {
							float speed=V.Norm();
							float time_to_stop=speed/unit_manager.unit_type[type_id].BrakeRate;
							float min_dist=time_to_stop*(speed-unit_manager.unit_type[type_id].BrakeRate*0.5f*time_to_stop);
							if(min_dist>=dist)			// Freine si nécessaire
								V=V-unit_manager.unit_type[type_id].BrakeRate*dt*J;
							else if(speed<unit_manager.unit_type[type_id].MaxVelocity)
								V=V+unit_manager.unit_type[type_id].Acceleration*dt*J;
							}
					}
					break;
				default:
					if(unit_manager.unit_type[type_id].canfly)
						deactivate();
					};
				}
				break;
			case CLASS_UNDEF:
			case CLASS_METAL:
			case CLASS_ENERGY:
			case CLASS_SPECIAL:
			case CLASS_FORT:
				break;
			default:
				printf("type inconnu %d\n",unit_manager.unit_type[type_id].TEDclass);
			};
			switch(mission->mission)		// Pour le code post déplacement
			{
			case MISSION_BUILD:
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
				break;
			case MISSION_GUARD:
				break;
			case MISSION_STOP:
				break;
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
				if(unit_manager.unit_type[type_id].canfly) {			// Un avion??
					activate();
					VECTOR J,I,K;
					K.x=K.z=0.0f;
					K.y=1.0f;
					VECTOR Target=mission->target;
					J=Target-Pos;
					J.y=0.0f;
					float dist=J.Norm();
					mission->last_d=dist;
					if(dist>0.0f)
						J=1.0f/dist*J;
					float dest_angle=acos(J.z)*RAD2DEG;
					if(J.x<0.0f) dest_angle=-dest_angle;
					while(fabs(dest_angle-Angle.y)>180.0f) {
						if(dest_angle<Angle.y)
							Angle.y-=360.0f;
						else
							Angle.y+=360.0f;
						}
//					if(dist>2.0f*unit_manager.unit_type[type_id].SightDistance && fabs(dest_angle-Angle.y)>=1.0f) {
					if(dist>(unit_manager.unit_type[type_id].ManeuverLeashLength>>1) && fabs(dest_angle-Angle.y)>=1.0f) {
						float aspeed=unit_manager.unit_type[type_id].TurnRate;
						if(dest_angle<Angle.y)
							aspeed=-aspeed;
						float a=dest_angle-Angle.y;
						V_Angle.y=aspeed*0.1f;
						float b=dest_angle-(Angle.y+dt*V_Angle.y);
						if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
							V_Angle.y=0.0f;
							Angle.y=dest_angle;
							}
						}

					J.z=cos(Angle.y*DEG2RAD);
					J.x=sin(Angle.y*DEG2RAD);
					J.y=0.0f;
					I.z=-J.x;
					I.x=J.z;
					I.y=0.0f;
					V=(V%K)*K+(V%J)*J+units.exp_dt_4*(V%I)*I;
					float speed=V.Norm();
					if(speed<unit_manager.unit_type[type_id].MaxVelocity)
						V=V+unit_manager.unit_type[type_id].Acceleration*dt*J;
					}
				break;
			case MISSION_MOVE:			// Oriente l'unité si elle est au sol
				if(jump_commands)	break;
				if(!unit_manager.unit_type[type_id].canfly && !unit_manager.unit_type[type_id].Upright && unit_manager.unit_type[type_id].TEDclass!=CLASS_SHIP && unit_manager.unit_type[type_id].TEDclass!=CLASS_WATER) {
					VECTOR I,J,K,A,B,C;
					MATRIX_4x4 M=RotateY((Angle.y+90.0f)*DEG2RAD);
					I.x=I.y=I.z=0.0f;
					J=K=I;
					I.x=1.0f;
					J.z=1.0f;
					K.y=1.0f;
					A=Pos-unit_manager.unit_type[type_id].FootprintZ*4.0f*I*M;
					B=Pos+(unit_manager.unit_type[type_id].FootprintX*4.0f*I-unit_manager.unit_type[type_id].FootprintZ*4.0f*J)*M;
					C=Pos+(unit_manager.unit_type[type_id].FootprintX*4.0f*I+unit_manager.unit_type[type_id].FootprintZ*4.0f*J)*M;
					A.y=map->get_unit_h(A.x,A.z);	// Projete le triangle
					B.y=map->get_unit_h(B.x,B.z);
					C.y=map->get_unit_h(C.x,C.z);
					VECTOR D=(B-A)*(B-C);
					if(D.y>=0.0f) {					// On ne met pas une unité à l'envers!!
						D.Unit();
						float angle_1=(D.z!=0.0f || D.y!=0.0f) ? acos(D.y/sqrt(D.y*D.y+D.z*D.z))*RAD2DEG : 0.0f;
						if(D.z<0.0f)	angle_1=-angle_1;
						D=D*RotateX(-angle_1*DEG2RAD);
						float angle_2=VAngle(D,K)*RAD2DEG;
						if(D.x>0.0f)	angle_2=-angle_2;
						if(fabs(angle_1-Angle.x)<=10.0f && fabs(angle_2-Angle.z)<=10.0f) {
							Angle.x=angle_1;
							Angle.z=angle_2;
							}
						}
					}
				if((mission->flags&1)==0)		// Si l'on n'est pas en mode patrouille
					break;
			default:								// Si l'unité peut attaquer d'elle même les unités enemies proches, elle le fait / Attack nearby enemies
				if(jump_commands)	break;
				if(unit_manager.unit_type[type_id].AutoFire && unit_manager.unit_type[type_id].canattack
				&& (unit_manager.unit_type[type_id].weapon[0]==NULL || (unit_manager.unit_type[type_id].weapon[0]!=NULL && !unit_manager.unit_type[type_id].weapon[0]->commandfire))
				&& (unit_manager.unit_type[type_id].weapon[1]==NULL || (unit_manager.unit_type[type_id].weapon[1]!=NULL && !unit_manager.unit_type[type_id].weapon[1]->commandfire))
				&& (unit_manager.unit_type[type_id].weapon[2]==NULL || (unit_manager.unit_type[type_id].weapon[2]!=NULL && !unit_manager.unit_type[type_id].weapon[2]->commandfire))) {
					int dx=unit_manager.unit_type[type_id].SightDistance+(int)(h+0.5f)>>3;
					int enemy_idx=-1;
					if(unit_manager.unit_type[type_id].weapon[0]!=NULL && (unit_manager.unit_type[type_id].weapon[0]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[0]->range>>4;
					if(unit_manager.unit_type[type_id].weapon[1]!=NULL && (unit_manager.unit_type[type_id].weapon[1]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[1]->range>>4;
					if(unit_manager.unit_type[type_id].weapon[2]!=NULL && (unit_manager.unit_type[type_id].weapon[2]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[2]->range>>4;
					int sx=rand_from_table()&0xF;
					int sy=rand_from_table()&0xF;
					int mask=1<<owner_id;
					for(int y=cur_py-dx+sy;y<=cur_py+dx;y+=0x10) {
						if(y>=0 && y<map->bloc_h_db-1)
							for(int x=cur_px-dx+sx;x<=cur_px+dx;x+=0x10)
								if(x>=0 && x<map->bloc_w_db-1 && (map->map_data[y][x].view&mask)==mask && map->map_data[y][x].unit_idx>-1 && map->map_data[y][x].unit_idx!=idx && units.unit[map->map_data[y][x].unit_idx].owner_id!=owner_id) {
									int cur_idx=map->map_data[y][x].unit_idx;
									if(unit_manager.unit_type[units.unit[cur_idx].type_id].ShootMe) {
										enemy_idx=map->map_data[y][x].unit_idx;
										break;
										}
									}
						if(enemy_idx>=0)	break;
						}
					if(enemy_idx>=0)			// Si on a trouvé une unité, on l'attaque
						set_mission(MISSION_ATTACK,&(units.unit[enemy_idx].Pos),false,0,true,&(units.unit[enemy_idx]),NULL,8);	// 8 = automatic fire
					}
				if(unit_manager.unit_type[type_id].antiweapons && unit_manager.unit_type[type_id].weapon[0]) {
					float coverage=unit_manager.unit_type[type_id].weapon[0]->coverage*unit_manager.unit_type[type_id].weapon[0]->coverage;
					float range=unit_manager.unit_type[type_id].weapon[0]->range*unit_manager.unit_type[type_id].weapon[0]->range>>2;
					int enemy_idx=-1;
					byte e=0;
					for(byte i=0;i+e<mem_size;i++) {
						if(memory[i+e]<0 || memory[i+e]>=weapons.nb_weapon || weapons.weapon[memory[i+e]].weapon_id==-1) {
							e++;
							i--;
							continue;
							}
						memory[i]=memory[i+e];
						}
					mem_size-=e;
					for(int i=0;i<weapons.nb_weapon;i+=(rand_from_table()%10)+1)
						if(weapons.weapon[i].weapon_id!=-1 && units.unit[weapons.weapon[i].shooter_idx].owner_id!=owner_id
						&& weapon_manager.weapon[weapons.weapon[i].weapon_id].targetable)
						if(((VECTOR)(weapons.weapon[i].target_pos-Pos)).Sq()<=coverage
						&& ((VECTOR)(weapons.weapon[i].Pos-Pos)).Sq()<=range) {
							int idx=-1;
							for(e=0;e<mem_size;e++)
								if(memory[e]==i) {
									idx=i;
									break;
									}
							if(idx==-1) {
								enemy_idx=i;
								if(mem_size<10) {
									memory[mem_size]=i;
									mem_size++;
									}
								break;
								}
							}
					if(enemy_idx>=0)			// Si on a trouvé une unité, on l'attaque
						add_mission(MISSION_ATTACK,&(weapons.weapon[enemy_idx].Pos),false,0,&(weapons.weapon[enemy_idx]),NULL,12);	// 12 = 4 | 8, targets a weapon and automatic fire
					}
				break;
			};
			}

		if(port[BUILD_PERCENT_LEFT]==0.0f) {
			Angle=Angle+dt*V_Angle;
			VECTOR OPos=Pos;
			sint32	o_px=cur_px;
			sint32	o_py=cur_py;
			V.y-=dt*map->ota_data.gravity;			// L'unité subit la force de gravitation
			Pos=Pos+dt*V;			// Déplace l'unité
			cur_px = ((int)(Pos.x)+map->map_w_d)>>3;
			cur_py = ((int)(Pos.z)+map->map_h_d)>>3;
			if(mission!=NULL && (mission->flags&2)!=0)	goto script_exec;
			if(unit_manager.unit_type[type_id].canmove && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
			if(!map->check_rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx)) {
				if(!unit_manager.unit_type[type_id].canfly) {
					bool back=false;
					if(map->check_rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx)) {
						Pos=OPos;
						cur_px = o_px;
						cur_py = o_py;
						back=true;
						}
					if((mission==NULL || (mission->flags&2)==0) && (mission->mission!=MISSION_STOP || mission->next==NULL)) {
						if(mission && mission->mission==MISSION_MOVE && back) {		// Recherche un autre chemin
							mission->mission=MISSION_MOVE_CALC_PATH;
							destroy_path(mission->path);
							mission->path=NULL;
							}
						VECTOR target=Pos;
						target.x+=(rand_from_table()&0x1F)-16;		// cherche ailleurs de la place
						target.z+=(rand_from_table()&0x1F)-16;
						add_mission(MISSION_MOVE_CALC_PATH,&target,true,0,NULL,NULL,2);		// On lui demande de s'éloigner
						}
					}
				else if(mission->mission!=MISSION_MOVE && mission->mission!=MISSION_BUILD && mission->mission!=MISSION_BUILD_2 && mission->mission!=MISSION_REPAIR && mission->mission!=MISSION_ATTACK && mission->mission!=MISSION_AIM && mission->mission!=MISSION_SHOOT) {
					if(map->check_rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx))
						Pos=OPos;
					if((mission==NULL || (mission->flags&2)==0) && (mission->mission!=MISSION_STOP || mission->next==NULL)) {
						VECTOR target=Pos;
						target.x+=(rand_from_table()&0x1F)-16;		// cherche ailleurs de la place pour attérrir
						target.z+=(rand_from_table()&0x1F)-16;
						add_mission(MISSION_MOVE_CALC_PATH,&target,true,0,NULL,NULL,2);		// On lui demande de s'éloigner
						}
					}
				}
			}
		script_exec:
		if(map && !jump_commands && !attached) {
			float min_h=map->get_unit_h(Pos.x,Pos.z);
			h=Pos.y-min_h;
			if(!unit_manager.unit_type[type_id].norestrict && Pos.y<map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV) {
				Pos.y=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if((unit_manager.unit_type[type_id].Floater || (unit_manager.unit_type[type_id].TEDclass&CLASS_SHIP))
				 && Pos.y<map->sealvl+(unit_manager.unit_type[type_id].AltFromSeaLevel)*H_DIV) {
				Pos.y=map->sealvl+unit_manager.unit_type[type_id].AltFromSeaLevel*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			else if(unit_manager.unit_type[type_id].TEDclass&CLASS_WATER)	{
				Pos.y=map->sealvl-unit_manager.unit_type[type_id].WaterLine*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if(min_h>Pos.y) {
				Pos.y=min_h;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if(unit_manager.unit_type[type_id].canfly && port[BUILD_PERCENT_LEFT]==0.0f) {
				if(mission && (mission->mission==MISSION_MOVE || mission->mission==MISSION_BUILD || mission->mission==MISSION_BUILD_2 || mission->mission==MISSION_REPAIR
				|| mission->mission==MISSION_ATTACK || mission->mission==MISSION_AIM || mission->mission==MISSION_SHOOT || nb_attached>0)) {
					float ideal_h=max(min_h,map->sealvl)+unit_manager.unit_type[type_id].CruiseAlt*H_DIV;
					V.y=(ideal_h-Pos.y)*0.5f;
					}
				else {
					float ideal_h=max(min_h,map->sealvl);
					V.y=(ideal_h-Pos.y)*0.1f;
					}
				}
			port[GROUND_HEIGHT]=Pos.y-min_h;
			}
		port[HEALTH]=hp*100/unit_manager.unit_type[type_id].MaxDamage;
		if(nb_running>0) {
			for(int i=0;i<nb_running;i++)
				run_script(dt,i,map);
			int e=0;
			for(int i=0;i+e<nb_running;) {				// Efface les scripts qui se sont arrêtés
				if(script_env[i+e].running) {
					script_env[i]=script_env[i+e];
					i++;
					}
				else {
					script_env[i+e].destroy();
					e++;
					}
				}
			nb_running-=e;
			}
		if(unit_manager.unit_type[type_id].canmove || !jump_commands) {
			map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			if(!attached)
				map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			}
		else if(port[YARD_OPEN]!=0.0f ? !was_open : was_open) {
			map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			if(!attached)
				map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			}
		built=false;
		attacked=false;
		return 0;
	}

	bool UNIT::hit(VECTOR P,VECTOR Dir,VECTOR *hit_vec)
	{
		if(flags==0)	return false;
		if(model) {
			VECTOR c_dir=model->center+Pos-P;
			if(((VECTOR)(c_dir*Dir)).Sq()<(model->size*0.5f)) {
				float scale=unit_manager.unit_type[type_id].Scale;
//				MATRIX_4x4 M=RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale);
				MATRIX_4x4 M=RotateX(-Angle.x*DEG2RAD)*RotateZ(-Angle.z*DEG2RAD)*RotateY(-Angle.y*DEG2RAD)*Scale(1.0f/scale);
				VECTOR RP=(1.0f/scale)*(P-Pos);
				bool is_hit=model->hit(RP,Dir,&data,hit_vec,M);
				if(is_hit) {
					*hit_vec=(*hit_vec)*(RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale))+Pos;
					*hit_vec=((*hit_vec-P)%Dir)*Dir+P;
					}
				return is_hit;
				}
			}
		return false;
	}

	void UNIT::show_orders(bool only_build_commands)				// Dessine les ordres reçus
	{
		MISSION *cur=mission;
		glEnable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		VECTOR p_target=Pos;
		float rab=(Atimer%1000000)*0.000001f;
		while(cur) {
			if(cur->step)				// S'il s'agit d'une étape on ne la montre pas
				cur=cur->next;
			if(!only_build_commands)
			{
			int curseur=anim_cursor(CURSOR_CROSS_LINK);
			float dx=0.5f*cursor.anm[CURSOR_CROSS_LINK].ofs_x[curseur];
			float dz=0.5f*cursor.anm[CURSOR_CROSS_LINK].ofs_y[curseur];
			float sx=0.5f*(cursor.anm[CURSOR_CROSS_LINK].bmp[curseur]->w-1);
			float sy=0.5f*(cursor.anm[CURSOR_CROSS_LINK].bmp[curseur]->h-1);
			float x,y,z;
			float dist=((VECTOR)(cur->target-p_target)).Norm();
			int rec=(int)(dist/30.0f);
			switch(cur->mission)
			{
			case MISSION_LOAD:
			case MISSION_UNLOAD:
			case MISSION_GUARD:
			case MISSION_PATROL:
			case MISSION_MOVE_CALC_PATH:
			case MISSION_MOVE:
			case MISSION_BUILD:
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
			case MISSION_RECLAIM:
				if(rec>0) {
					glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_CROSS_LINK].glbmp[curseur]);
					glBegin(GL_QUADS);
						for(int i=0;i<rec;i++) {
							x=p_target.x+(cur->target.x-p_target.x)*(i+rab)/rec;
							y=p_target.y+(cur->target.y-p_target.y)*(i+rab)/rec;
							z=p_target.z+(cur->target.z-p_target.z)*(i+rab)/rec;
							x-=dx;
							y+=0.75f;
							z-=dz;
							glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
							glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
							glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
							glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
							}
					glEnd();
					}
				p_target=cur->target;
			};
			}
			glDisable(GL_DEPTH_TEST);
			switch(cur->mission)
			{
			case MISSION_LOAD:
				if(!only_build_commands)
				{
				int curseur=anim_cursor(CURSOR_LOAD);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_LOAD].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_LOAD].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_LOAD].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_LOAD].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_LOAD].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_MOVE_CALC_PATH:
			case MISSION_MOVE:
				if(!only_build_commands)
				{
				int curseur=anim_cursor(CURSOR_MOVE);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_MOVE].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_MOVE].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_MOVE].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_MOVE].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_MOVE].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_BUILD:
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				if(cur->data>=0 && cur->data<unit_manager.nb_unit) {
					float DX = (unit_manager.unit_type[cur->data].FootprintX<<2);
					float DZ = (unit_manager.unit_type[cur->data].FootprintZ<<2);
					float blue = 0.0f, green = 1.0f;
					if(only_build_commands) {
						blue = 1.0f;
						green = 0.0f;
						}
					glPushMatrix();
					glTranslatef(cur->target.x,cur->target.y,cur->target.z);
					glDisable(GL_CULL_FACE);
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glBegin(GL_QUADS);
						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,-DZ);			// First quad
						glVertex3f(DX,0.0f,-DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);
						glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,-DZ);			// Second quad
						glVertex3f(-DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
						glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(DX,0.0f,-DZ);			// Third quad
						glVertex3f(DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
						glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);
					glEnd();
					glDisable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glPopMatrix();
					if(unit_manager.unit_type[cur->data].model!=NULL) {
						glEnable(GL_LIGHTING);
						glEnable(GL_CULL_FACE);
						glPushMatrix();
						glTranslatef(cur->target.x,cur->target.y,cur->target.z);
						glColor4f(0.0f,green,blue,0.5f);
						unit_manager.unit_type[cur->data].model->obj.draw(NULL,false,false,false);
						glPopMatrix();
						glEnable(GL_BLEND);
						glEnable(GL_TEXTURE_2D);
						glDisable(GL_LIGHTING);
						glDisable(GL_CULL_FACE);
						glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
						}
					glPushMatrix();
					glTranslatef(cur->target.x,cur->target.y,cur->target.z);
					glDisable(GL_CULL_FACE);
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glBegin(GL_QUADS);
						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,DZ);			// Fourth quad
						glVertex3f(DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
						glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
					glEnd();
					glPopMatrix();
					glDisable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glColor4f(1.0f,1.0f,1.0f,1.0f);
					}
				break;
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_REPAIR);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_REPAIR].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_REPAIR].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_REPAIR].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_RECLAIM:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_RECLAIM);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_RECLAIM].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_RECLAIM].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_RECLAIM].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_PATROL:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_PATROL);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_PATROL].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_PATROL].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_PATROL].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_PATROL].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_PATROL].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_GUARD:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_GUARD);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_GUARD].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_GUARD].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_GUARD].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_GUARD].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_GUARD].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_ATTACK);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_ATTACK].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_ATTACK].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_ATTACK].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_ATTACK].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_ATTACK].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			};
			glEnable(GL_DEPTH_TEST);
			cur=cur->next;
			}
		glDisable(GL_BLEND);
	}

	bool INGAME_UNITS::select(CAMERA *cam,int sel_x[],int sel_y[])
	{
		bool selected=false;
		cam->SetView();
		MATRIX_4x4 ModelView,Project,T;
		int	viewportCoords[4] = {0, 0, 0, 0};
		glGetIntegerv(GL_VIEWPORT, viewportCoords);
		glGetFloatv(GL_MODELVIEW_MATRIX, (float*)ModelView.E);
		glGetFloatv(GL_PROJECTION_MATRIX, (float*)Project.E);
		ModelView=Transpose(ModelView);
		Project=Transpose(Project);
		float VW=(viewportCoords[2]-viewportCoords[0])*0.5f;
		float VH=-(viewportCoords[3]-viewportCoords[1])*0.5f;
		T=ModelView*Project;		// Matrice de transformation

		POINTF UPos,O;
		O.x=O.y=O.z=0.0f;
		int X1,Y1,X2,Y2;
		X1=min(sel_x[0],sel_x[1]);
		Y1=min(sel_y[0],sel_y[1]);
		X2=max(sel_x[0],sel_x[1]);
		Y2=max(sel_y[0],sel_y[1]);

		for(uint16 e=0;e<index_list_size;e++) {
			uint16 i = idx_list[e];
			if(unit[i].flags!=0 && unit[i].owner_id==players.local_human_id && unit[i].port[BUILD_PERCENT_LEFT]==0.0f && unit[i].visible) {		// Ne sélectionne que les unités achevées
				if(key[KEY_LSHIFT] && unit[i].sel) {
					selected=true;
					continue;
					}
				if(!key[KEY_LSHIFT])
					unit[i].sel=false;

				VECTOR Vec=unit[i].Pos-cam->Pos;
				float d=Vec.Sq();
				if(d>16384.0f && (Vec%cam->Dir)<=0.0f) continue;

				UPos=glNMult((O+unit[i].Pos),T);		// Transforme la position de l'unité
				UPos.x=UPos.x*VW+VW;
				UPos.y=UPos.y*VH-VH;

				if(X1<=UPos.x && X2>=UPos.x && Y1<=UPos.y && Y2>=UPos.y)
					selected=unit[i].sel=true;
				}
			}
		return selected;
	}

	int INGAME_UNITS::pick(CAMERA *cam,int sensibility)
	{
		int index=-1;

		if(nb_unit<=0)
			return index;

		VECTOR Dir;
		Dir=cam->Dir+2.0f*(mouse_x-SCREEN_W_HALF)*SCREEN_W_INV*cam->Side-1.5f*(mouse_y-SCREEN_H_HALF)*SCREEN_H_INV*cam->Up;
		Dir.Unit();		// Direction pointée par le curseur

		bool detectable=false;
		int i;

		for(uint16 e=0;e<index_list_size;e++) {
			i = idx_list[e];
			if(unit[i].flags==0 || !unit[i].visible) continue;		// Si l'unité n'existe pas on la zappe
			unit[i].flags&=0xFD;	// Enlève l'indicateur de possibilité d'intersection
			VECTOR center=unit[i].model->center+unit[i].Pos-cam->Pos;
			float size=unit[i].model->size*unit_manager.unit_type[unit[i].type_id].Scale*unit_manager.unit_type[unit[i].type_id].Scale;
			center=Dir*center;
			float dist=center.Sq();
			if(dist<size) {
				detectable=true;
				unit[i].flags|=0x2;		// Unité détectable
				}
			}

		if(!detectable)			// Si aucune unité n'est à proximité du curseur, on zappe la méthode OpenGl
			return index;

		glDisable(GL_LIGHTING);
		glDisable(GL_FOG);

		int BUFSIZE=max_unit<<2;

		GLuint *selectBuf = new GLuint[BUFSIZE];
		GLint hits;

		glSelectBuffer (BUFSIZE, selectBuf);
		glRenderMode (GL_SELECT);

		glInitNames();
		glPushName(0);

		int	viewport[4] = {0, 0, 0, 0};
		glGetIntegerv(GL_VIEWPORT, viewport);

		if(cam->Ref!=NULL)
			cam->Pos=cam->RPos+(*(cam->Ref));
		else
			cam->Pos=cam->RPos;

		VECTOR	FP;
		FP=cam->Pos+cam->Dir;

		cam->zfar2=cam->zfar*cam->zfar;
		glMatrixMode (GL_PROJECTION);			// Matrice de projection
		glLoadIdentity ();

		gluPickMatrix(mouse_x, SCREEN_H-mouse_y, sensibility, sensibility, viewport);

		glFrustum (-1.0f*cam->znear, 1.0f*cam->znear, -0.75f*cam->znear, 0.75f*cam->znear, cam->znear, cam->zfar);

		glMatrixMode (GL_MODELVIEW);

		glLoadIdentity();			// Charge la matrice identité

		gluLookAt(cam->Pos.x,cam->Pos.y,cam->Pos.z,FP.x,FP.y,FP.z,cam->Up.x,cam->Up.y,cam->Up.z);

		for(uint16 e=0;e<index_list_size;e++) {
			i = idx_list[e];
			if((unit[i].flags&0x2)==0x2) {			// Si l'unité existe et est sélectionnable
				unit[i].flags&=0xFD;
				if(unit[i].model) {
					glLoadName(i);
					VECTOR D;
					D=unit[i].Pos-cam->Pos;		// Vecteur "viseur unité" partant de la caméra vers l'unité

					float dist=D.Sq();
					if(!((dist>=16384.0f && (D%cam->Dir)<=0.0f)||(D%cam->Dir)>cam->zfar2)) {
						glPushMatrix();

						glTranslatef(unit[i].Pos.x,unit[i].Pos.y,unit[i].Pos.z);
						glRotatef(unit[i].Angle.x,1.0f,0.0f,0.0f);
						glRotatef(unit[i].Angle.y,0.0f,1.0f,0.0f);
						glRotatef(unit[i].Angle.z,0.0f,0.0f,1.0f);
						float scale=unit_manager.unit_type[unit[i].type_id].Scale;
						glScalef(scale,scale,scale);
						unit[i].model->draw(&(unit[i].data),false,true);

						glPopMatrix();
						}
					}
				}
			}

		glFlush ();

		hits = glRenderMode (GL_RENDER);

		glEnable(GL_LIGHTING);
		glEnable(GL_FOG);

		int j;
		GLuint names, *ptr;

		float nearest=2.0f;
		ptr = (GLuint *) selectBuf;
		for (i = 0; i < hits; i++) { /*  for each hit  */
			names = *ptr;
			ptr++;
			float z1=(float) *ptr/0x7fffffff; ptr++;
			float z2=(float) *ptr/0x7fffffff; ptr++;
			int v=*ptr; ptr++;

			if(z1<nearest) {
				nearest=z1;
				index=v;
				}
			}

		delete[] selectBuf;
		return index;
	}

	void UNIT::shoot(int target,VECTOR startpos,VECTOR Dir,int w_id)
	{
		POINTF O;
		O.x=O.y=O.z=0.0f;
		if(unit_manager.unit_type[type_id].weapon[w_id]->startsmoke && visible)
			particle_engine.make_smoke(O+startpos,0,1,0.0f);
		int w_idx=weapons.add_weapon(unit_manager.unit_type[type_id].weapon[w_id]->nb_id,idx);
		weapons.weapon[w_idx].Pos=startpos;
		if(unit_manager.unit_type[type_id].weapon[w_id]->startvelocity==0.0f && !unit_manager.unit_type[type_id].weapon[w_id]->selfprop)
			weapons.weapon[w_idx].V=unit_manager.unit_type[type_id].weapon[w_id]->weaponvelocity*Dir;
		else
			weapons.weapon[w_idx].V=unit_manager.unit_type[type_id].weapon[w_id]->startvelocity*Dir;
		if(unit_manager.unit_type[type_id].weapon[w_id]->dropped)
			weapons.weapon[w_idx].V=weapons.weapon[w_idx].V+V;
		weapons.weapon[w_idx].owner=owner_id;
		weapons.weapon[w_idx].target=target;
		if(unit_manager.unit_type[type_id].weapon[w_id]->interceptor)
			weapons.weapon[w_idx].target_pos=weapons.weapon[target].Pos;
		else
			weapons.weapon[w_idx].target_pos=units.unit[target].Pos+units.unit[target].model->center;
		weapons.weapon[w_idx].stime=0.0f;
		weapons.weapon[w_idx].visible=visible;
	}

void *create_unit(int type_id,int owner,VECTOR pos,MAP *map)
{
	int id=units.create(type_id,owner);
	if(id>=0) {
		units.unit[id].Pos=pos;
		units.unit[id].port[BUILD_PERCENT_LEFT]=100.0f;
		units.unit[id].cur_px = ((int)(units.unit[id].Pos.x)+map->map_w_d)>>3;
		units.unit[id].cur_py = ((int)(units.unit[id].Pos.z)+map->map_h_d)>>3;
		return &(units.unit[id]);
		}
	return NULL;
}

const bool can_be_built(const VECTOR Pos,MAP *map,const int unit_type_id)
{
	if(unit_type_id<0 || unit_type_id>=unit_manager.nb_unit)	return false;	// test the ID

	if(map==NULL)	return false;		// to avoid an error

	int w = unit_manager.unit_type[unit_type_id].FootprintX;
	int h = unit_manager.unit_type[unit_type_id].FootprintZ;
	int x = ((int)(Pos.x+map->map_w_d)>>3)-(w>>1);
	int y = ((int)(Pos.z+map->map_h_d)>>3)-(h>>1);
	if(x<0 || y<0 || x+w>=(map->bloc_w<<1) || y+h>=(map->bloc_h<<1))	return false;	// check if it is inside the map

	if(!map->check_rect(x,y,w,h,-1))	return false;		// There already something
	float dh = fabs(map->check_rect_dh(x,y,w,h));
	float depth = map->check_depth(x,y,w,h);

	if(dh>unit_manager.unit_type[unit_type_id].MaxSlope*H_DIV)	return false;	// Check the slope

			// Check if unit can be there
	if(depth<unit_manager.unit_type[unit_type_id].MinWaterDepth*H_DIV || depth>unit_manager.unit_type[unit_type_id].MaxWaterDepth*H_DIV)	return false;
//	if(depth>0 && (unit_manager.unit_type[unit_type_id].Category&NOTSUB))	return false;

	if(!map->check_vents(x,y,w,h,unit_manager.unit_type[unit_type_id].yardmap))	return false;

	if(map->check_lava(x+1>>1,y+1>>1,w+1>>1,h+1>>1))	return false;

	return true;
}
