/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2006  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include "stdafx.h"
#include "TA3D_NameSpace.h"

namespace TA3D
{
	String Lowercase( const String &sstring )
	{
		String Return;

		Return.resize (sstring.length());
		for( uint32 i = 0; i < sstring.length(); i++)
			Return[i] = tolower (sstring[i]);

		return (Return);
	}

	String Uppercase( const String &sstring )
	{
		String Return;

		Return.resize (sstring.length());
		for( uint32 i = 0; i < sstring.length(); i++)
			Return[i] = toupper (sstring[i]);

		return (Return);
	}

	bool StartsWith( const String &a, const String &b )
	{
		String y = Lowercase( a );
		String z = Lowercase( b );
		uint16 ai, bi;

		ai = (uint16)y.length();
		bi = (uint16)z.length();

		if( ai > bi )
			return  ( (y.compare( 0, bi, z ) == 0 ) ? true : false );
		else
			return ( (z.compare( 0, ai, y ) == 0 ) ? true : false );
	}

	String format(const char *fmt, ...)
	{
		if( !fmt )
			return "";

		int   result = -1, length = 256;
		char *buffer = 0;

		va_list args;
		va_start(args, fmt);

		while( result == -1 )
		{
			if( buffer )
				delete [] buffer;

			buffer = new char [length + 1];

			memset(buffer, 0, length + 1);

#if defined TA3D_PLATFORM_WINDOWS
			result = _vsnprintf_s( buffer, length, _TRUNCATE, fmt, args );
#else
			result = uszprintf(buffer, length, fmt, args);
#endif
			length *= 2;
		}

		String s( buffer );
		delete [] buffer;

		va_end(args);

		return s;
	}

#if defined TA3D_PLATFORM_WINDOWS
	void ExtractPathFile( const String &szFullFileName, String &szFile, String &szDir )
	{
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];

		// Below extracts our path where the application is being run.
		::_splitpath_s( szFullFileName.c_str(), drive, dir, fname, ext );
      
		szFile = fname;
		szDir = drive;
		szDir += dir;
   }
#endif
	String GetClientPath( void )
	{
		static bool bName = false;
		static String szName = "";
      
		if (!bName)
		{
#if defined TA3D_PLATFORM_WINDOWS
			char fPath[ MAX_PATH ];
			String Tmp;

			::GetModuleFileNameA( NULL, fPath, MAX_PATH );

			ExtractPathFile( fPath, Tmp, szName );
#endif
			bName = true;
		}
      
		return szName;
	}

	// Cire: Useful malloc operation.
	void *GetMem( sint32 size, sint32 zero )
	{
		void *result;

		if( zero )
			result = calloc( size, 1 );
		else
			result = malloc( size );

		/*
		TODO: Global DEBUGGER NOTIFICATION HERE.
		if( !result )
		GlobalDebugger->Failed to alloc memory for hpi
		*/
		return result;
	}

	FILE *TA3D_OpenFile( const String &FileName, const String Mode )
	{
		FILE *file;

#if defined TA3D_PLATFORM_WINDOWS
		errno_t err;
		if( ( err = fopen_s( &file, FileName.c_str(), Mode.c_str() )) == 0 )
			return file;
#else
		file=fopen( FileName.c_str(), Mode.c_str() );
		if( file )
			return file;
#endif
		return NULL;
	}
}
