/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2006  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*
**  File: stdafx.h
** Notes:  **** PLEASE SEE README.txt FOR LICENCE AGREEMENT ****
**  Cire: *stdafx.h and stdafx.cpp will generate a pch file
**         (pre compiled headers).
**        *All other cpp files MUST include this file as its first include.
**        *No .h files should include this file.
**        *The goal is to include everything that we need from system, and
**          game libiries, ie everything we need external to our project.
*/

#pragma once

// Cire:
//   Right off the bat, lets get everything we will need from STL includes.
#include <cstdio>
#include <cstdlib>
#include <map>
#include <list>
#include <vector>
#include <algorithm>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

// Now lets get platform specific stuff.
#include "TA3D_Platform.h"

#if defined TA3D_PLATFORM_WINDOWS
// Cire:
//   I had to setup a pragma on c4312 warnings, because algero include
//     was generating alot of compiler noise.
	#pragma warning( disable : 4312 )

// Cire:
//   The below definations, make it so that we do not need to run fix
//     on allegro, to set proper platform, I'am not entirely sure how
//     other platfroms wana handle this so I just left it as including
//     allgero directly.
	#define ALLEGRO_MSVC

// Cire:
//   Since algero include will include windows headers for us, we probably
//    should make sure that when it does it excludes rarely used crap
//    from windows headers.
	#define WIN32_LEAN_AND_MEAN

	#include <allegro.h> // in allegro

// Cire:
//   Since we disabled 4312, we probably should set it back to its
//     default state.
	#pragma warning( default : 4312 )



	#include "tools/win32/include/gl/glew.h"

	#pragma comment(lib, "tools/win32/libs/alleg.lib")
	#pragma comment(lib, "tools/win32/libs/agl.lib")
	#pragma comment( lib, "opengl32.lib" )
	#pragma comment( lib, "glu32.lib" )
	#pragma comment( lib, "glaux.lib" )
	#pragma comment( lib, "tools/win32/libs/glew32.lib" )

	#pragma warning( disable : 4005 )
	#include <alleggl.h>
	#pragma warning( default : 4005 )
#else
// Cire:
//   Other platfroms may wish to adjust how allegro is included, for
//   now its this way.
	#include <allegro.h>
	#include <alleggl.h>
#endif

// Now lets get gl/glu support.
#include <GL/glu.h>

#include "jpeg/ta3d_jpg.h"

// Math include.
#include <math.h>
// Cire:
//   Malloc should not be platform specfic right???
#include <malloc.h>


// Cire:
//   string and wstring typedefs to make life easier.
typedef std::string  String;
typedef std::wstring WString;

// Cire:
//   Since byte seems to be common throughout the project we'll typedef here.
typedef uint8         byte;
typedef unsigned char uchar;
typedef signed char   schar;

// Floating point types, defined for consistencies sakes.
typedef float  real32;
typedef double real64;


#if defined TA3D_PLATFORM_WINDOWS
// Cire:
//   The below definations are my way of dealing with 2005 deperation
//   by pretty much redirecting them to proper secure methods, kludgy i know but
//   it works.
	#define strcasecmp(x,xx) _stricmp( x, xx )
	inline char *strcat(char *a, char *b)
	{
		uint32 i = strlen(a);
		uint32 e = 0;
		while(b[e]) {
			a[i++]=b[e];
			e++;
			}
		a[i]=b[e];
		return a;
	}
	#define strdup(x) _strdup( x )
	#define vsprintf(x,xx,xxx) vsprintf_s( x, sizeof(x), xx, xxx )
	inline char *strlwr(char *text)
	{
		uint32 i=0;
		while(text[i]) {
			if(text[i]>='A' && text[i]<='Z')
				text[i]+='a'-'A';
			i++;
			}
		return text;
	}
	inline char *strupr(char *text)
	{
		uint32 i=0;
		while(text[i]) {
			if(text[i]>='a' && text[i]<='z')
				text[i]+='A'-'a';
			i++;
			}
		return text;
	}
	#define getcwd( x, xx ) _getcwd( x, xx )

    #define strncpy( x,xx,xxx ) strncpy_s( x, sizeof(x),xx,xxx )

	#define sscanf(x,xx,xxx ) sscanf_s(x,xx,xxx )
// Cire:
//  The below functions don't exists within windows math routines.
	static inline const real32 asinh( const real32 f )
	{
		return (real32)log( (real32)(f + sqrt( f * f + 1)) );
	}

	static inline const real32 acosh( const real32 f )
	{
		return (real32)log( (real32)( f + sqrt( f * f- 1)) );
	}

	static inline const real32 atanh( const real32 f)
	{
		return (real32)log( (real32)( (real32)(1.0f / f + 1.0f) /
			(real32)(1.0f / f - 1.0f))  ) / 2.0f;
	}
#endif

#if ((MAKE_VERSION(4, 2, 1) == MAKE_VERSION(ALLEGRO_VERSION, ALLEGRO_SUB_VERSION, ALLEGRO_WIP_VERSION)) && ALLEGRO_WIP_VERSION>0) \
	|| (MAKE_VERSION(4, 2, 1) < MAKE_VERSION(ALLEGRO_VERSION, ALLEGRO_SUB_VERSION, ALLEGRO_WIP_VERSION))
	#define FILE_SIZE	file_size_ex
#else
	#define FILE_SIZE	file_size
#endif

// zuzuf: to prevent some warnings
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_TARNAME
#undef PACKAGE_STRING
#undef PACKAGE_VERSION
