/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"

#include "../config.h"		// Configuration de TA3D à la compilation
#ifdef TA3D_PLATFORM_WINDOWS
	#undef USE_OGG
	#undef USE_MP3
#else
	#if not USE_OGG
		#undef USE_OGG
	#endif
	#if not USE_MP3
		#undef USE_MP3
	#endif
#endif

#include "music.h"
#include "intro.h"			// Introduction
#include "ta3d.h"			// Moteur
#include "menu.h"			// Menus du jeu
#include "taconfig.h"		// Configuration

FONT *TA3D::VARS::ta_alfont;
FONT *TA3D::VARS::ta_glfont;
FONT *TA3D::VARS::aglfont;
RGB  *TA3D::VARS::pal;

VECTOR cursor_on_map(CAMERA *cam,MAP *map);

int main(int argc,char *argv[])
{
allegro_init();

install_timer();			// Initialise allegro
install_keyboard();
install_mouse();
install_allegro_gl();

install_sound(DIGI_AUTODETECT, MIDI_NONE, 0);

install_int_ex(Timer,precision);

int		scr_w=640,scr_h=480;
bool	fullscreen=false;
int		FSAA=0;
TA3D_CONFIG	config_manager;
config_manager.set_config_file("ta3d.cfg");
config_manager.add_option("showfps",true,false,NULL);
config_manager.add_option("fps_limit",false,true,NULL);
config_manager.add_option("wireframe",true,false,NULL);
config_manager.add_option("particle",true,false,NULL);
config_manager.add_option("trees",true,false,NULL);
config_manager.add_option("shadow",true,false,NULL);
config_manager.add_option("shadow_quality",false,false,NULL);
config_manager.add_option("shadow_r",false,true,NULL);
config_manager.add_option("priority_level",false,false,NULL);
config_manager.add_option("timefactor",false,true,NULL);
config_manager.add_option("screen_width",false,false,&scr_w);
config_manager.add_option("screen_height",false,false,&scr_h);
config_manager.add_option("fullscreen",true,false,&fullscreen);
config_manager.add_option("fsaa",false,false,&FSAA);
config_manager.add_option("water_quality",false,false,NULL);
config_manager.add_option("LANG",false,false,&LANG);
config_manager.add_option("height_line",true,false,NULL);
if(config_manager.read_config_file())
	config_manager.write_config_file();

	allegro_gl_clear_settings();			// Initialise AllegroGL
	allegro_gl_set (AGL_STENCIL_DEPTH, 8);
	allegro_gl_set (AGL_SAMPLE_BUFFERS, 0);
	allegro_gl_set (AGL_SAMPLES, FSAA);
	allegro_gl_set (AGL_COLOR_DEPTH, 32);
	allegro_gl_set (AGL_Z_DEPTH, 32);
	allegro_gl_set (AGL_FULLSCREEN, TRUE);
	allegro_gl_set (AGL_DOUBLEBUFFER, 1);
	allegro_gl_set (AGL_RENDERMETHOD, 1);
	allegro_gl_set (AGL_SUGGEST, AGL_RENDERMETHOD | AGL_COLOR_DEPTH | AGL_Z_DEPTH | AGL_DOUBLEBUFFER | AGL_FULLSCREEN | AGL_SAMPLES | AGL_SAMPLE_BUFFERS | AGL_STENCIL_DEPTH);

	allegro_gl_use_mipmapping(TRUE);

	allegro_gl_flip_texture(false);

request_refresh_rate(85);

if(fullscreen)
	set_gfx_mode(GFX_OPENGL_FULLSCREEN, scr_w, scr_h, 0, 0);	// Entre en mode graphique OpenGL (plein écran)
else
	set_gfx_mode(GFX_OPENGL_WINDOWED, scr_w, scr_h, 0, 0);		// Entre en mode graphique OpenGL (fenêtré)

SCREEN_W_HALF = SCREEN_W / 2;
SCREEN_H_HALF = SCREEN_H / 2;
SCREEN_W_INV = 1.0f / SCREEN_W;
SCREEN_H_INV = 1.0f / SCREEN_H;

install_ext();		// Pour les fonctions OpenGl supplémentaires

if(g_useTextureCompression)			// Active la compression de texture
	allegro_gl_set_texture_format(GL_COMPRESSED_RGB_ARB);
else
	allegro_gl_set_texture_format(GL_RGB8);

set_uformat(U_ASCII);		// Juste histoire d'avoir un affichage correct des textes

set_window_title("Total Annihilation 3D");

start=Atimer;		// Pour l'animation du curseur

Console=new cConsole( "Console.log" );				// Initialise la console
Console->AddEntry("TA3D Engine");
Console->AddEntry(TA3D_ENGINE_VERSION);
translation_manager.load("ta3d.res");

if(g_useTextureCompression)
	Console->AddEntry(translation_manager.get(0));
if(g_useStencilTwoSide)
	Console->AddEntry(translation_manager.get(1));
if(g_useCopyDepthToColor)
	Console->AddEntry(translation_manager.get(2));
if(g_useProgram)
	Console->AddEntry(translation_manager.get(3));
if(g_useFBO)
	Console->AddEntry(translation_manager.get(4));

#ifdef USE_MP3
	Console->AddEntry(translation_manager.get(5));
#endif
#ifdef USE_OGG
	Console->AddEntry(translation_manager.get(6));
#endif

jpgalleg_init();
Console->AddEntry(translation_manager.get(7));		// JPG support

Console->AddEntry(translation_manager.get(8));

HPIManager=new cHPIHandler("");

pal=LoadPal("palettes\\palette.pal");			// Charge la palette graphique
set_palette(pal);		// Active la palette chargée

aglfont=allegro_gl_convert_allegro_font(font,AGL_FONT_TYPE_TEXTURED,-0.9f);							// Load fonts
ta_alfont=load_font("gfx/title_font.tga",NULL,NULL);												// TA's fonts
ta_glfont=allegro_gl_convert_allegro_font_ex(ta_alfont,AGL_FONT_TYPE_TEXTURED,-1.0f,GL_RGBA8);

Console->AddEntry(translation_manager.get(9));

play_intro();

Console->AddEntry(translation_manager.get(10));
model_manager.init();
unit_manager.init();
feature_manager.init();
weapon_manager.init();
sound_manager.init();
fx_manager.init();
Console->AddEntry(translation_manager.get(11));
init_rand_table();

ta3d_sidedata.init();
ta3d_sidedata.load_data();

{
	byte *data=HPIManager->PullFromHPI("anims\\CURSORS.GAF");
	cursor.load_gaf(data);
	cursor.convert();
	CURSOR_MOVE=cursor.find_entry("cursormove");
	CURSOR_GREEN=cursor.find_entry("cursorgrn");
	CURSOR_CROSS=cursor.find_entry("cursorselect");
	CURSOR_RED=cursor.find_entry("cursorred");
	CURSOR_LOAD=cursor.find_entry("cursorload");
	CURSOR_GUARD=cursor.find_entry("cursordefend");
	CURSOR_PATROL=cursor.find_entry("cursorpatrol");
	CURSOR_REPAIR=cursor.find_entry("cursorrepair");
	CURSOR_ATTACK=cursor.find_entry("cursorattack");
	CURSOR_BLUE=cursor.find_entry("cursornormal");
	CURSOR_AIR_LOAD=cursor.find_entry("cursorpickup");
	CURSOR_BOMB_ATTACK=cursor.find_entry("cursorairstrike");
	CURSOR_BALANCE=cursor.find_entry("cursorunload");
	CURSOR_RECLAIM=cursor.find_entry("cursorreclamate");
	CURSOR_WAIT=cursor.find_entry("cursorhourglass");
	CURSOR_CANT_ATTACK=cursor.find_entry("cursortoofar");
	CURSOR_CROSS_LINK=cursor.find_entry("pathicon");
//	for(int i=0;i<cursor.nb_anim;i++)
//		printf("%s\n",cursor.anm[i].name);
	free(data);
}

if(argc==1) {
	Console->AddEntry(translation_manager.get(12));
	main_menu();
	}
else
	if(HPIManager->Exists(argv[1])) {						// Check if this map exists
		Console->AddEntry(translation_manager.get(13));
		play(argv[1]);
		}

cursor.destroy();
ta3d_sidedata.destroy();

delete HPIManager;

Console->AddEntry(translation_manager.get(14));

allegro_gl_destroy_font(ta_glfont);
allegro_gl_destroy_font(aglfont);

destroy_font(ta_alfont);

delete Console;
Console=NULL;

allegro_exit();

return 0;
}
END_OF_MAIN();

/*--------------------------------------------------------------------\
|                            Moteur du jeu                            |
\--------------------------------------------------------------------*/
void play(char *filename)
{
#ifdef CWDEBUG
  Debug( libcw_do.on() );
  Debug( dc::malloc.on() );
  Debug( dc::notice.on() );
#endif
expected_players=2;

int start_time=Atimer;		// Pour la mesure du temps de chargement

loading(0.0f,translation_manager.get(22));

/*-----------------------charge les textures-------------------------*/

texture_manager.all_texture();

loading(100.0f/7.0f,translation_manager.get(23));
/*-----------------------charge les modèles 3D-----------------------*/

model_manager.init();
model_manager.load_all(loading);

loading(200.0f/7.0f,translation_manager.get(24));
/*-----------------------charge les éléments graphiques--------------*/

load_features();
feature_manager.clean();

loading(250.0f/7.0f,translation_manager.get(25));
/*-----------------------charge les armes----------------------------*/

load_weapons();

weapons.init();

loading(300.0f/7.0f,translation_manager.get(26));
/*-----------------------charge les unités---------------------------*/

load_all_units();

loading(400.0f/7.0f,translation_manager.get(27));
/*-----------------------libère la mémoire inutilisée----------------*/

texture_manager.destroy();

loading(500.0f/7.0f,translation_manager.get(10));
/*-----------------------initialise le moteur------------------------*/

SetDefOpenGlState();
particle_engine.init();

set_palette(pal);

players.init();													// Objet contenant les données sur les joueurs
players.add("joueur0","arm",PLAYER_CONTROL_LOCAL_HUMAN);		// Ajoute un joueur
players.add("joe(AI)","core",PLAYER_CONTROL_LOCAL_AI);			// Ajoute un joueur

units.init();

loading(600.0f/7.0f,translation_manager.get(28));
/*-----------------------charge la carte-----------------------------*/

byte *map_file=HPIManager->PullFromHPI(filename);

if(!map_file)	return;
MAP *map=load_tnt_map(map_file);
free(map_file);

replace_extension(filename, filename, "ota", strlen(filename)+1);

uint32 ota_size=0;
map_file=HPIManager->PullFromHPI(filename,&ota_size);
if(map_file) {
	map->ota_data.load((char*)map_file,ota_size);
	free(map_file);
	}

bool night = (map->ota_data.planet!=NULL && (strcmp(map->ota_data.planet,"lunar")==0 || strcmp(map->ota_data.planet,"metal")==0));

if(map->ota_data.maxwindspeed!=map->ota_data.minwindspeed)
	map->wind=(rand()%(map->ota_data.maxwindspeed-map->ota_data.minwindspeed))+map->ota_data.minwindspeed;

ASM_PROGRAM	game_script;		// Script qui va gérer la partie
game_script.compile("scripts/default.c");	// Compile le script

GLuint	sky;
if(night)	sky=LoadTex("gfx/sky-night.tga");
else sky=LoadTex("gfx/sky.jpg");
GLuint	water=LoadTex("gfx/water.jpg");
GLuint	glow=LoadTex("gfx/glow.tga");
GLuint	freecam_on=LoadTex("gfx/freecam_on.tga");
GLuint	freecam_off=LoadTex("gfx/freecam_off.tga");

loading(100.0f,translation_manager.get(29));

start_time=Atimer-start_time;
printf("temps de chargement: %f secondes\n",(float)start_time/SECS_TO_TIMER(1));

#if defined USE_MP3 || defined USE_OGG
music_manager.load_play_list();
#endif

float Conv=1.0f/SECS_TO_TIMER(1);
float dt=0.0f;
float t=0.0f;
int count=Atimer;
int nbfps=0;
int fpscount=Atimer;
int fps=0;
int i;

float r1,r2,r3,or1;
r1=r2=r3=0.0f;
r1=-63.44f;
CAMERA cam;
VECTOR cam_target;
bool cam_has_target=false;
bool freecam=false;
game_cam=&cam;

cam.RPos.x=cam.RPos.y=cam.RPos.z=0.0f;
cam.RPos.z+=150.0f;
cam.RPos.y=150.0f;
cam.zfar=500.0f;

float FogD=0.3f;
float FogNear=cam.zfar*0.5f;
float FogColor[]={0.8f,0.8f,0.8f,1.0f};

if(night) {	FogColor[0]=0.0f;	FogColor[1]=0.0f;	FogColor[2]=0.0f;	}

GLuint FogMode=GL_LINEAR;
	glFogi (GL_FOG_MODE, FogMode);
	glFogfv (GL_FOG_COLOR, FogColor);
	glFogf (GL_FOG_DENSITY, FogD);
	glHint (GL_FOG_HINT, GL_NICEST);
	glFogf (GL_FOG_START, FogNear);
	glFogf (GL_FOG_END, cam.zfar);
	
glEnable(GL_FOG);

HWLIGHT	sun;
sun.Att=0.0f;
sun.Dir.x=-1.0f;
sun.Dir.y=2.0f;
sun.Dir.z=1.0f;
sun.Dir.Unit();
sun.LightAmbient[0]=0.25f;
sun.LightAmbient[1]=0.25f;
sun.LightAmbient[2]=0.25f;
sun.LightAmbient[3]=0.25f;
sun.LightDiffuse[0]=1.0f;
sun.LightDiffuse[1]=1.0f;
sun.LightDiffuse[2]=1.0f;
sun.LightDiffuse[3]=1.0f;
sun.LightSpecular[0]=0.0f;
sun.LightSpecular[1]=0.0f;
sun.LightSpecular[2]=0.0f;
sun.LightSpecular[3]=0.0f;
sun.Directionnal=true;

float speed_limit=-1.0f;
float delay=(speed_limit==0.0f) ? 0.0f : 1.0f/speed_limit;
int nb_shoot=0;
bool shoot=false;

bool tilde=false;
bool showfps=false;
bool done=false;

#define EXIT_NONE		0x0
#define EXIT_VICTORY	0x1
#define EXIT_DEFEAT		0x2

int	exit_mode=EXIT_NONE;

fire=particle_engine.addtex("gfx/fire.tga","gfx/firea.tga");
build_part=particle_engine.addtex("gfx/part.tga",NULL);

int mx,my,omb=mouse_b,omb2=mouse_b,omb3=mouse_b;
int	sel_x[2],sel_y[2],selecting=false;
int cur_sel=-1;
bool selected=false;
int	build=-1;				// Indique si l'utilisateur veut construire quelque chose
int cur_sel_index=-1;
int omz=mouse_z;
float cam_h=150.0f;
float time_factor=1.0f;

bool wireframe=false;		// Mode fil de fer pour le terrain
bool shadow=false;			// Active les ombres
bool particle=true;
bool trees=true;
bool show_script=false;			// Affichage des scripts
bool show_model=false;			// Affichage les noms des sous objets du modèle 3D de l'unité sélectionnée
bool rotate_light=false;
float light_angle=0.0f;
int shadow_quality=1;
float shadow_r=0.02f;
int priority_level=0;
bool cheat_metal=false;
bool cheat_energy=false;
bool internal_name=false;
bool ia_debug=false;
bool view_dbg=false;

bool height_line=true;

int visibility_unit_id=0;

float unit_info=0.0f;
int	unit_info_id=-1;

int key_frame=0;

int fog_count=Atimer;

bool need_redraw=true;
GLint nb_aux_buffers=0;
rest(100);						// Juste pour éviter un "Xlib: ..." / To avoid "Xlib: ..." error

float wind_t=t;					// To handle wind variations
bool wind_change=false;

int video_timer=Atimer;			// To handle video
bool video_shoot=false;

int current_order=SIGNAL_ORDER_NONE;

WATER	water_obj;
water_obj.build(map->map_w,map->map_h,40,1000);

SKY		sky_obj;
sky_obj.build(10,400);

SHADER	water_shader, water_shader_reflec, water_pass1;
GLuint transtex,reflectex,first_pass,zbuf;
GLuint water_FBO;
int water_quality=0;

if(g_useProgram && g_useFBO && map->water) {
	glGenFramebuffersEXT(1,&water_FBO);

	water_pass1.load("shaders/water_pass1.frag","shaders/water_pass1.vert");
	water_shader.load("shaders/water.frag","shaders/water.vert");
	water_shader_reflec.load("shaders/water_reflec.frag","shaders/water.vert");

	BITMAP *tmp=create_bitmap_ex(32,512,512);	// On ne peut pas utiliser screen donc on crée un BITMAP temporaire
	transtex=allegro_gl_make_texture(tmp);		// pour les effets de transparence de l'eau
	reflectex=allegro_gl_make_texture(tmp);		// pour les effets de reflection sur l'eau
	first_pass=allegro_gl_make_texture(tmp);	// pour les effets de reflection/transparence sur l'eau

	glGenRenderbuffersEXT(1,&zbuf);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, zbuf);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT24, 512, 512);
	destroy_bitmap(tmp);
	glBindTexture(GL_TEXTURE_2D,transtex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D,reflectex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D,first_pass);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

	if(g_useTextureCompression)			// Active la compression de texture
		allegro_gl_set_texture_format(GL_COMPRESSED_RGB_ARB);
	else
		allegro_gl_set_texture_format(GL_RGB8);
	}

TA3D_CONFIG	config_manager;
config_manager.set_config_file("ta3d.cfg");
config_manager.add_option("showfps",true,false,&showfps);
config_manager.add_option("fps_limit",false,true,&speed_limit);
config_manager.add_option("wireframe",true,false,&wireframe);
config_manager.add_option("particle",true,false,&particle);
config_manager.add_option("trees",true,false,&trees);
config_manager.add_option("shadow",true,false,&shadow);
config_manager.add_option("shadow_quality",false,false,&shadow_quality);
config_manager.add_option("shadow_r",false,true,&shadow_r);
config_manager.add_option("priority_level",false,false,&priority_level);
config_manager.add_option("timefactor",false,true,&time_factor);
config_manager.add_option("screen_width",false,false,NULL);
config_manager.add_option("screen_height",false,false,NULL);
config_manager.add_option("fullscreen",true,false,NULL);
config_manager.add_option("fsaa",false,false,NULL);
config_manager.add_option("water_quality",false,false,&water_quality);
config_manager.add_option("LANG",false,false,&LANG);
config_manager.add_option("height_line",true,false,&height_line);
if(config_manager.read_config_file())
	config_manager.write_config_file();

delay=(speed_limit==0.0f) ? 0.0f : 1.0f/speed_limit;

do
{
	bool can_be_there=false;		// To tell if the unit the player wants to build can built where the cursor is

	if(video_shoot)
		if((Atimer-video_timer)*Conv>=1.0f/15.0f) {
			video_timer=Atimer;
			shoot=true;
			}

	need_redraw|=(nb_aux_buffers==0);

	uint16 nb_limit = max((units.index_list_size+fps)/(fps+1),(int)((units.index_list_size+fps)/(fps+1)*time_factor));
	for(i=0; i<nb_limit ;i++)
		if(visibility_unit_id<units.index_list_size) {			// met à jour progressivement le brouillard de guerre
			uint16 e = units.idx_list[visibility_unit_id];
			sint32 cur_px = units.unit[e].cur_px;
			sint32 cur_py = units.unit[e].cur_py;
			uint16 cur_sight = (int)units.unit[e].h+unit_manager.unit_type[units.unit[e].type_id].SightDistance>>3;
			if( units.unit[e].old_px!=cur_px || units.unit[e].old_py!=cur_py || units.unit[e].old_sight!=cur_sight ) {
				if(units.unit[e].old_sight>0)
					map->update_player_visibility(units.unit[e].owner_id,units.unit[e].old_px,units.unit[e].old_py,units.unit[e].old_sight,true);
				map->update_player_visibility(units.unit[e].owner_id,cur_px,cur_py,cur_sight);
				units.unit[e].old_px=cur_px;
				units.unit[e].old_py=cur_py;
				units.unit[e].old_sight=cur_sight;
				}
			visibility_unit_id++;
			}
		else {
			visibility_unit_id=0;
			break;
			}

	/*------------------------ handle wind speed and dir -----------------------*/

	if(t-wind_t>=10.0f) {		// Make a change every 10 sec.
		wind_t=t;
		map->wind+=(rand_from_table()%2001)-1000;
		if(map->wind<map->ota_data.minwindspeed)
			map->wind=map->ota_data.minwindspeed;
		else if(map->wind>map->ota_data.maxwindspeed)
			map->wind=map->ota_data.maxwindspeed;
		map->wind_dir+=(rand_from_table()%901)*0.1f-45.0f;
		if(map->wind_dir<0.0f)	map->wind_dir+=360.0f;
		else if(map->wind_dir>=360.0f)	map->wind_dir-=360.0f;
		map->wind_vec.y=0.0f;
		map->wind_vec.x=0.01f*map->wind*cos(map->wind_dir*DEG2RAD);
		map->wind_vec.z=0.01f*map->wind*sin(map->wind_dir*DEG2RAD);
		wind_change=true;
		}
	else
		wind_change=false;

	/*------------------- fin de la gestion des autres joueurs -----------------*/

	/*------------------------ gestion des autres joueurs ----------------------*/

	players.player_control(map);

	/*------------------- fin de la gestion des autres joueurs -----------------*/

	/*------------------------------ gestion du son ----------------------------*/

#if defined USE_MP3 || defined USE_OGG
	music_manager.poll();

	if(units.nb_attacked/(units.nb_attacked+units.nb_built+1)>=0.75f)
		music_manager.change_state(MUSIC_TYPE_BATTLE);
	else if(units.nb_attacked/(units.nb_attacked+units.nb_built+1)<=0.25f)
		music_manager.change_state(MUSIC_TYPE_NONE);
#endif

	sound_manager.clean(cam.RPos,64.0f*cam.Side);

	/*------------------------- fin de la gestion du son -----------------------*/

	VECTOR old_cam_pos=cam.RPos;
	float old_r1=r1,old_r2=r2,old_r3=r3;
	if(!freecam) {
		int delta=mouse_z-omz;
		r1+=delta*2.0f;
		if(r1>-45.0f) r1=-45.0f;
		else if(r1<-90.0f) r1=-90.0f;
		cam_h=151.0f+exp((-r1-45.0f)*0.15f)/1536.0f*max(map->map_w,map->map_h);
		if(delta && mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {
			cam_target=cam.RPos+cursor_on_map(&cam,map)-map->hit(cam.RPos,cam.Dir);
			cam_has_target=true;
			}
		}
	else {
		cam.RPos=cam.RPos-0.5f*(mouse_z-omz)*cam.Dir;
		cam_has_target=false;
		}
	omz=mouse_z;

	cam.zfar=500.0f+(cam_h-150.0f)*2.0f;

	cursor_type=CURSOR_DEFAULT;			// Revient au curseur par défaut

	if(mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {
		if(selected)
			for(i=0;i<units.index_list_size;i++) {
				uint32 e = units.idx_list[i];
				if(units.unit[e].flags!=0 && units.unit[e].owner_id==players.local_human_id && units.unit[e].sel && unit_manager.unit_type[units.unit[e].type_id].canmove) {
					cursor_type=CURSOR_MOVE;
					break;
					}
				}
		}

	if(build>=0)
		cursor_type=CURSOR_DEFAULT;

	dt=(Atimer-count)*Conv;			// Régulateur de vitesse d'affichage
	while(dt<delay) {
		switch(priority_level)
		{
		case 0:			rest(1);	break;
		case 1:			rest(0);	break;
		};
		dt=(Atimer-count)*Conv;
		}

	unit_info-=dt;
	light_angle+=dt*time_factor;

	t+=dt*time_factor;
	nbfps++;
	if(nbfps>=10 && (Atimer-fpscount)*Conv>=0.05f) {
		fps=(int)(nbfps/((Atimer-fpscount)*Conv));
		fpscount=Atimer;
		nbfps=0;
		}
	count=Atimer;

	if(freecam && cam.RPos.y<map->sealvl) {
		FogD=0.03f;
		FogNear=0.0f;
		FogMode=GL_EXP;

		FogColor[0]=0.0f;
		FogColor[1]=0.0f;
		FogColor[2]=0.3f;
		FogColor[3]=1.0f;
		}
	else {
		FogD=0.3f;
		FogNear=cam.zfar*0.5f;
		FogMode=GL_LINEAR;

		FogColor[0]=0.8f;
		FogColor[1]=0.8f;
		FogColor[2]=0.8f;
		FogColor[3]=1.0f;

		if(night) {	FogColor[0]=0.0f;	FogColor[1]=0.0f;	FogColor[2]=0.0f;	}
		}

	SetDefOpenGlState();
	glClearColor(FogColor[0],FogColor[1],FogColor[2],FogColor[3]);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

	glFogi (GL_FOG_MODE, FogMode);
	glFogfv (GL_FOG_COLOR, FogColor);
	glFogf (GL_FOG_DENSITY, FogD);
	glHint (GL_FOG_HINT, GL_NICEST);
	glFogf (GL_FOG_START, FogNear);
	glFogf (GL_FOG_END, cam.zfar);

/*------------bloc regroupant ce qui est relatif aux commandes----------------*/

	if(key[KEY_F1] && cur_sel>=0) {
		unit_info_id=cur_sel;
		unit_info=1.0f;
		}

	if(mouse_x<128.0f && mouse_y<128.0f && mouse_x>=0.0f && mouse_y>=0.0f && mouse_b==1) {
		cam.RPos.x=(mouse_x-64)*map->map_w/128.0f*252.0f/map->mini_w;
		cam.RPos.z=(mouse_y-64)*map->map_h/128.0f*252.0f/map->mini_h;
		cam_has_target=false;
		}
	if(mouse_x<10) {
		cam.RPos.x-=100.0f*dt;
		cam_has_target=false;
		}
	if(mouse_x>SCREEN_W-10) {
		cam.RPos.x+=100.0f*dt;
		cam_has_target=false;
		}
	if(mouse_y<10) {
		cam.RPos.z-=100.0f*dt;
		cam_has_target=false;
		}
	if(mouse_y>SCREEN_H-10) {
		cam.RPos.z+=100.0f*dt;
		cam_has_target=false;
		}

	if(freecam) {
		if(mouse_b==4) {
			get_mouse_mickeys(&mx,&my);
			if(omb==mouse_b) {
				r2-=mx;
				r1-=my;
				}
			position_mouse(SCREEN_W_HALF,SCREEN_H_HALF);
			}
		else if(omb==4)
			position_mouse(SCREEN_W_HALF,SCREEN_H_HALF);
		}
	else {
		if(mouse_b==4) {
			get_mouse_mickeys(&mx,&my);
			if(omb==mouse_b) {
				cam.RPos.x+=mx;
				cam.RPos.z+=my;
				cam_has_target=false;
				}
			position_mouse(SCREEN_W_HALF,SCREEN_H_HALF);
			}
		else if(omb==4) {
			position_mouse(SCREEN_W_HALF,SCREEN_H_HALF);
			cam_has_target=false;
			}
		}
	omb=mouse_b;

	if(!freecam)	{
		if(key[KEY_UP]) {
			cam.RPos.z-=100.0f*dt;
			cam_has_target=false;
			}
		if(key[KEY_DOWN]) {
			cam.RPos.z+=100.0f*dt;
			cam_has_target=false;
			}
		if(key[KEY_RIGHT]) {
			cam.RPos.x+=100.0f*dt;
			cam_has_target=false;
			}
		if(key[KEY_LEFT]) {
			cam.RPos.x-=100.0f*dt;
			cam_has_target=false;
			}

		float h=map->get_unit_h(cam.RPos.x,cam.RPos.z);
		if(h<map->sealvl)
			h=map->sealvl;
		cam.RPos.y+=5.0f*(h+cam_h-cam.RPos.y)*dt;
		}
	else {
		if(key[KEY_UP])
			cam.RPos=cam.RPos+100.0f*dt*cam.Dir;
		if(key[KEY_DOWN])
			cam.RPos=cam.RPos-100.0f*dt*cam.Dir;
		if(key[KEY_RIGHT])
			cam.RPos=cam.RPos+100.0f*dt*cam.Side;
		if(key[KEY_LEFT])
			cam.RPos=cam.RPos-100.0f*dt*cam.Side;
		}

	if(cam.RPos.x<-map->map_w_d) {
		cam.RPos.x=-map->map_w_d;
		cam_has_target=false;
		}
	if(cam.RPos.x>map->map_w_d)	{
		cam.RPos.x=map->map_w_d;
		cam_has_target=false;
		}
	if(cam.RPos.z<-map->map_h_d+200.0f)	{
		cam.RPos.z=-map->map_h_d+200.0f;
		cam_has_target=false;
		}
	if(cam.RPos.z>map->map_h_d)	{
		cam.RPos.z=map->map_h_d;
		cam_has_target=false;
		}

	MATRIX_4x4 Rotation;
	r1+=cam.Rotat.x*dt;
	r2+=cam.Rotat.y*dt;
	r3+=cam.Rotat.z*dt;
	Rotation=RotateX(r1*DEG2RAD)*RotateY(r2*DEG2RAD)*RotateZ(r3*DEG2RAD);

	if(cam_has_target) {
		cam_target.y=cam.RPos.y;
		cam.RPos=2.0f*dt*(cam_target-cam.RPos)+cam.RPos;
		if(((VECTOR)(cam_target-cam.RPos)).Sq()<10.0f)
			cam_has_target=false;
		}

	cam.SetMatrix(Rotation,dt);

	if(!need_redraw && ((VECTOR)(cam.RPos-old_cam_pos)).Sq()<=0.00001f && r1==old_r1 && r2==old_r2 && r3==old_r3)
		cam.RPos=old_cam_pos;
	else
		need_redraw=true;

	if(build>=0 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32)		// Check if this unit can be built there
		can_be_there=can_be_built(cursor_on_map(&cam,map),map,build);

	if(!selected)
		current_order=SIGNAL_ORDER_NONE;

	if(current_order!=SIGNAL_ORDER_NONE)
		selecting=false;

	if(selected) {
		bool builders=false;
		bool canattack=false;
		bool canreclamate=false;
		for(uint16 e=0;e<units.index_list_size;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel) {
				builders|=unit_manager.unit_type[units.unit[i].type_id].Builder;
				canattack|=unit_manager.unit_type[units.unit[i].type_id].canattack;
				canreclamate|=unit_manager.unit_type[units.unit[i].type_id].CanReclamate;
				}
			}
		int pointing=units.pick(&cam);		// Sur quoi le curseur est-il pointé??

		if(pointing>=0) {	// S'il y a quelque chose sous le curseur
			cursor_type=CURSOR_CROSS;
			if(units.unit[pointing].owner_id!=players.local_human_id) {
				if(canattack)
					cursor_type=CURSOR_ATTACK;
				else if(canreclamate)
					cursor_type=CURSOR_RECLAIM;
				else
					cursor_type=CURSOR_CANT_ATTACK;
				}
			else if(units.unit[pointing].port[BUILD_PERCENT_LEFT]>0.0f && builders)	cursor_type=CURSOR_REPAIR;

			switch(current_order)
			{
			case SIGNAL_ORDER_MOVE:		cursor_type=CURSOR_MOVE;	break;
			case SIGNAL_ORDER_PATROL:	cursor_type=CURSOR_PATROL;	break;
			case SIGNAL_ORDER_GUARD:	cursor_type=CURSOR_GUARD;	break;
			case SIGNAL_ORDER_ATTACK:	cursor_type=CURSOR_ATTACK;	break;
			case SIGNAL_ORDER_RECLAM:	cursor_type=CURSOR_RECLAIM;	break;
			case SIGNAL_ORDER_LOAD:		cursor_type=CURSOR_LOAD;	break;
			case SIGNAL_ORDER_UNLOAD:	cursor_type=CURSOR_LOAD;	break;
			};

			if(mouse_b!=1 && omb3==1) {
				if(cursor_type==CURSOR_ATTACK) {
					for(uint16 e=0;e<units.index_list_size;e++) {
						i = units.idx_list[e];
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel && unit_manager.unit_type[units.unit[i].type_id].canattack) {
							if(key[KEY_LSHIFT])
								units.unit[i].add_mission(MISSION_ATTACK,&(units.unit[pointing].Pos),false,0,&(units.unit[pointing]));
							else
								units.unit[i].set_mission(MISSION_ATTACK,&(units.unit[pointing].Pos),false,0,true,&(units.unit[pointing]));
							}
						}
					if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
					}
				else if(cursor_type==CURSOR_REPAIR) {
					for(uint16 e=0;e<units.index_list_size;e++) {
						i = units.idx_list[e];
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel
						&& unit_manager.unit_type[units.unit[i].type_id].Builder && unit_manager.unit_type[units.unit[i].type_id].TEDclass!=CLASS_PLANT) {
							sound_manager.play(unit_manager.unit_type[units.unit[i].type_id].repair,units.unit[i].Pos);
							if(key[KEY_LSHIFT])
								units.unit[i].add_mission(MISSION_REPAIR,&(units.unit[pointing].Pos),false,0,&(units.unit[pointing]));
							else
								units.unit[i].set_mission(MISSION_REPAIR,&(units.unit[pointing].Pos),false,0,true,&(units.unit[pointing]));
							}
						}
					if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
					}
				else if(cursor_type==CURSOR_RECLAIM) {
					for(uint16 e=0;e<units.index_list_size;e++) {
						i = units.idx_list[e];
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel
						&& unit_manager.unit_type[units.unit[i].type_id].CanReclamate) {
							sound_manager.play(unit_manager.unit_type[units.unit[i].type_id].working,units.unit[i].Pos);
							if(key[KEY_LSHIFT])
								units.unit[i].add_mission(MISSION_RECLAIM,&(units.unit[pointing].Pos),false,0,&(units.unit[pointing]));
							else
								units.unit[i].set_mission(MISSION_RECLAIM,&(units.unit[pointing].Pos),false,0,true,&(units.unit[pointing]));
							}
						}
					if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
					}
				else if(cursor_type==CURSOR_GUARD) {	// Le curseur donne un ordre
					units.give_order_guard(players.local_human_id,pointing,!key[KEY_LSHIFT]);
					if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
					}
				else if(cursor_type==CURSOR_LOAD) {	// Le curseur donne un ordre
					units.give_order_load(players.local_human_id,pointing,!key[KEY_LSHIFT]);
					if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
					}
				}
			}
		else
			switch(current_order)
			{
			case SIGNAL_ORDER_MOVE:		cursor_type=CURSOR_MOVE;	break;
			case SIGNAL_ORDER_PATROL:	cursor_type=CURSOR_PATROL;	break;
			case SIGNAL_ORDER_GUARD:	cursor_type=CURSOR_GUARD;	break;
			case SIGNAL_ORDER_ATTACK:	cursor_type=CURSOR_ATTACK;	break;
			case SIGNAL_ORDER_RECLAM:	cursor_type=CURSOR_RECLAIM;	break;
			case SIGNAL_ORDER_LOAD:		cursor_type=CURSOR_LOAD;	break;
			case SIGNAL_ORDER_UNLOAD:	cursor_type=CURSOR_LOAD;	break;
			};
		}

	if(cursor_type==CURSOR_RECLAIM && mouse_b!=1 && omb3 ==1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		VECTOR cur_pos=cursor_on_map(&cam,map);
		int px=((int)(cur_pos.x+map->map_w_d))>>3;
		int py=((int)(cur_pos.z+map->map_h_d))>>3;
		int idx=-1;
		if(px>=0 && px<map->bloc_w<<1 && py>=0 && py<map->bloc_h<<1) {
			idx=-map->map_data[py][px].unit_idx-2;
			if(idx>=0 && features.feature[idx].type>=0)
				if(!feature_manager.feature[features.feature[idx].type].reclaimable)
					idx=-1;
			}
		if(idx>=0)
			for(uint16 e=0;e<units.index_list_size;e++) {
				i = units.idx_list[e];
				if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel
				&& unit_manager.unit_type[units.unit[i].type_id].CanReclamate) {
					sound_manager.play(unit_manager.unit_type[units.unit[i].type_id].working,units.unit[i].Pos);
					if(key[KEY_LSHIFT])
						units.unit[i].add_mission(MISSION_RECLAIM,&cur_pos,false,idx,NULL);
					else
						units.unit[i].set_mission(MISSION_RECLAIM,&cur_pos,false,idx,true,NULL);
					}
				}
		if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
		}

	if(cursor_type==CURSOR_MOVE && mouse_b!=1 && omb3 ==1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		units.give_order_move(players.local_human_id,cursor_on_map(&cam,map),!key[KEY_LSHIFT]);
		if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
		}

	if(cursor_type==CURSOR_PATROL && mouse_b!=1 && omb3 ==1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		units.give_order_patrol(players.local_human_id,cursor_on_map(&cam,map),!key[KEY_LSHIFT]);
		if(!key[KEY_LSHIFT])	current_order=SIGNAL_ORDER_NONE;
		}

	if(build>=0 && can_be_there && cursor_type==CURSOR_DEFAULT && mouse_b!=1 && omb3 ==1 && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32) {	// Le curseur donne un ordre
		VECTOR target = cursor_on_map(&cam,map);
		target.x=((int)(target.x)+map->map_w_d)>>3;
		target.z=((int)(target.z)+map->map_h_d)>>3;
		target.x=target.x*8.0f-map->map_w_d;
		target.z=target.z*8.0f-map->map_h_d;
		target.y=max(map->get_unit_h(target.x,target.z),map->sealvl);
		units.give_order_build(players.local_human_id,build,target,!key[KEY_LSHIFT]);
		if(!key[KEY_LSHIFT])
			build=-1;
		}

	if(mouse_b!=1 && omb3==1 && !key[KEY_LSHIFT] && mouse_x>128.0f && mouse_y>32.0f && mouse_y<SCREEN_H-32)
		current_order=SIGNAL_ORDER_NONE;

//---------------------------------	Code de sélection d'unités

	if(mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {
		if(mouse_b==2 && omb3!=2) {					// Déselectionne les unités
			if(build>=0)
				build=-1;				// Sort du mode construction
			else {
				selected=false;
				cur_sel=-1;
				cur_sel_index=-1;
				for(uint16 e=0;e<units.index_list_size;e++) {
					i = units.idx_list[e];
					if(units.unit[i].owner_id==players.local_human_id)			// On peut désélectionner les morts, ça ne change rien :-)
						units.unit[i].sel=false;
					}
				}
			}
		}

	if(build==-1 && (cursor_type==CURSOR_DEFAULT || cursor_type==CURSOR_CROSS) && mouse_x>128 && (mouse_y>32 || selecting)
	   && (mouse_y<SCREEN_H-32 || selecting)) {		// Si le curseur est dans la zone de jeu
		if(mouse_b!=1 && selecting) {		// Récupère les unités présentes dans la sélection
			if(sel_x[0]==sel_x[1] && sel_y[0]==sel_y[1]) {
				int pointing=units.pick(&cam);		// Sélectionne une unité sur clic
				if(!key[KEY_LSHIFT])
					for(uint16 e=0;e<units.index_list_size;e++) {
						i = units.idx_list[e];
						if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id)
							units.unit[i].sel=false;
						}
				if(pointing>=0 && units.unit[pointing].port[BUILD_PERCENT_LEFT]==0.0f)		// On ne sélectionne pas les unités en construction
					units.unit[pointing].sel^=true;			// Sélectionne/Désélectionne si l'unité est déjà sélectionnée en appuyant sur SHIFT
				selected=false;
				for(uint16 e=0;e<units.index_list_size;e++) {
					i = units.idx_list[e];
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id)
						selected|=units.unit[i].sel;
					}
				}
			else
				selected=units.select(&cam,sel_x,sel_y);		// Séléction au lasso
			cur_sel=-1;
			cur_sel_index=-1;
			for(uint16 e=0;e<units.index_list_size && cur_sel!=-2;e++) {
				i = units.idx_list[e];
				if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
					cur_sel= (cur_sel==-1) ? i : -2;
				}
			if(cur_sel>=0) {
				cur_sel_index=cur_sel;
				cur_sel=units.unit[cur_sel].type_id;
						// Joue un son
				sound_manager.play(unit_manager.unit_type[cur_sel].select1,units.unit[i].Pos);
				}
			}
		selecting=false;
		if(mouse_b==1) {
			if(omb3!=1) {
				sel_x[0]=mouse_x;
				sel_y[0]=mouse_y;
				}
			sel_x[1]=mouse_x;
			sel_y[1]=mouse_y;
			selecting=true;
			}
		}
	else
		selecting=false;
	omb3=mouse_b;

	if(mouse_x<=128 || mouse_y<=32 || mouse_y>=SCREEN_H-32)	cursor_type=CURSOR_DEFAULT;

	if(key[KEY_LCONTROL] && key[KEY_C]) {
		for(uint16 e=0;e<units.index_list_size;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags && units.unit[i].owner_id==players.local_human_id) {
				if(unit_manager.unit_type[units.unit[i].type_id].Category & CTRL_C)
					units.unit[i].sel=true;
				else if(!key[KEY_LSHIFT])
					units.unit[i].sel=false;
				}
			}
		cur_sel=-1;
		cur_sel_index=-1;
		build=-1;
		for(uint16 e=0;e<units.index_list_size && cur_sel!=-2;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
				cur_sel= (cur_sel==-1) ? i : -2;
			}
		selected=(cur_sel!=-1);
		if(cur_sel>=0) {
			cur_sel_index=cur_sel;
			cur_sel=units.unit[cur_sel].type_id;
			}
		}

	if((key[KEY_LCONTROL] || key[KEY_RCONTROL]) && key[KEY_Z]) {		// Séletionne toutes les unités dont le type est déjà sélectionné
		bool *sel_type = new bool[unit_manager.nb_unit];
		for(i=0;i<unit_manager.nb_unit;i++)
			sel_type[i]=false;
		for(uint16 e=0;e<units.index_list_size;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
				sel_type[units.unit[i].type_id]=true;
			}
		for(uint16 e=0;e<units.index_list_size;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags && units.unit[i].port[BUILD_PERCENT_LEFT]==0.0f && units.unit[i].owner_id==players.local_human_id && sel_type[units.unit[i].type_id])
				units.unit[i].sel=true;
			}
		cur_sel=-1;
		cur_sel_index=-1;
		for(uint16 e=0;e<units.index_list_size && cur_sel!=-2;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
				cur_sel= (cur_sel==-1) ? i : -2;
			}
		selected=(cur_sel!=-1);
		build=-1;
		if(cur_sel>=0) {
			cur_sel_index=cur_sel;
			cur_sel=units.unit[cur_sel].type_id;
			}
		delete[] sel_type;
		}
	else if(key[KEY_LCONTROL] || key[KEY_RCONTROL]) {			// Formation de groupes d'unités
		int grpe=-1;
		if(key[KEY_0])	grpe=0;
		if(key[KEY_1])	grpe=1;
		if(key[KEY_2])	grpe=2;
		if(key[KEY_3])	grpe=3;
		if(key[KEY_4])	grpe=4;
		if(key[KEY_5])	grpe=5;
		if(key[KEY_6])	grpe=6;
		if(key[KEY_7])	grpe=7;
		if(key[KEY_8])	grpe=8;
		if(key[KEY_9])	grpe=9;

		if(grpe>=0) {
			grpe=1<<grpe;
			for(uint16 e=0;e<units.index_list_size;e++) {
				i = units.idx_list[e];
				if(units.unit[i].flags && units.unit[i].owner_id==players.local_human_id) {
					if(units.unit[i].sel)
						units.unit[i].groupe|=grpe;
					else if(!key[KEY_LSHIFT])
						units.unit[i].groupe&=~grpe;
					}
				}
			}
		}
	else if(key[KEY_ALT]) {			// Restauration de groupes d'unités
		int grpe=-1;
		if(key[KEY_0])	grpe=0;
		if(key[KEY_1])	grpe=1;
		if(key[KEY_2])	grpe=2;
		if(key[KEY_3])	grpe=3;
		if(key[KEY_4])	grpe=4;
		if(key[KEY_5])	grpe=5;
		if(key[KEY_6])	grpe=6;
		if(key[KEY_7])	grpe=7;
		if(key[KEY_8])	grpe=8;
		if(key[KEY_9])	grpe=9;

		if(grpe>=0) {
			build=-1;
			grpe=1<<grpe;
			for(uint16 e=0;e<units.index_list_size;e++) {
				i = units.idx_list[e];
				if(units.unit[i].flags && units.unit[i].owner_id==players.local_human_id) {
					if(units.unit[i].groupe&grpe)
						units.unit[i].sel=true;
					else if(!key[KEY_LSHIFT])
						units.unit[i].sel=false;
					}
				}
			}

		cur_sel=-1;
		cur_sel_index=-1;
		for(uint16 e=0;e<units.index_list_size && cur_sel!=-2;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
				cur_sel= (cur_sel==-1) ? i : -2;
			}
		selected=(cur_sel!=-1);
		if(cur_sel>=0) {
			cur_sel_index=cur_sel;
			cur_sel=units.unit[cur_sel].type_id;
			}
		}

/*--------------bloc regroupant ce qui est relatif au temps-------------------*/

	float timetosimulate=dt*time_factor;								// Moteur physique
	for(float dt_sim=0.0f;dt_sim<timetosimulate;dt_sim+=TIME_UNIT) {
		float dt_r=min(timetosimulate-dt_sim,TIME_UNIT);
		units.move(dt_r,map,key_frame,wind_change);					// Anime les unités
		wind_change=false;
		key_frame++;
		weapons.move(dt_r,map);							// Anime les armes
		}
	if(particle)
		particle_engine.move(timetosimulate,map->wind_vec,map->ota_data.gravity);					// Anime les particules
	if(trees)
		features.move(timetosimulate,need_redraw);					// Anime les objets
	fx_manager.move(timetosimulate);

/*----------------------------------------------------------------------------*/

	if(g_useProgram && g_useFBO && water_quality==3 && map->water && !map->ota_data.lavaworld) {		// Dessine les reflets sur l'eau

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,water_FBO);

		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,reflectex,0);

		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,zbuf);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		glViewport(0, 0, 512, 512);

		for(i=0;i<8;i++) {
			glActiveTextureARB(GL_TEXTURE0_ARB+i);
			ReInitTexSys();
			glDisable(GL_TEXTURE_2D);
			}
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glDisable(GL_BLEND);

		double eqn[4]= { 0.0f, 1.0f, 0.0f, -map->sealvl};

		CAMERA refcam=cam;
		refcam.zfar*=2.0f;
		refcam.mirror=true;
		refcam.mirror_pos=-2.0f*map->sealvl;

		refcam.SetView();
		glClipPlane(GL_CLIP_PLANE1, eqn);
		glEnable(GL_CLIP_PLANE1);

		sun.Set(refcam);
		sun.Enable();

		refcam.SetView();
		glRotatef(Atimer*Conv*RAD2DEG,0.0f,1.0f,0.0f);
		glTranslatef(0.0f,60.0f,0.0f);

		refcam.zfar*=100.0f;
		refcam.SetView();
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,sky);
		glDisable(GL_LIGHTING);
		glFogi (GL_FOG_MODE, FogMode);
		glFogfv (GL_FOG_COLOR, FogColor);
		glFogf (GL_FOG_DENSITY, FogD);
		glHint (GL_FOG_HINT, GL_NICEST);
		glFogf (GL_FOG_START, FogNear);
		glFogf (GL_FOG_END, refcam.zfar);
		if(night) {
			glCullFace(GL_FRONT);
			glTranslatef(cam.RPos.x,cam.RPos.y,cam.RPos.z);
			sky_obj.draw();
			}
		else {
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f,map->map_h*0.002f);				glVertex3f(-2.0f*map->map_w,300.0f,2.0f*map->map_h);
				glTexCoord2f(map->map_w*0.002f,map->map_h*0.002f);	glVertex3f(2.0f*map->map_w,300.0f,2.0f*map->map_h);
				glTexCoord2f(map->map_w*0.002f,0.0f);				glVertex3f(2.0f*map->map_w,300.0f,-2.0f*map->map_h);
				glTexCoord2f(0.0f,0.0f);							glVertex3f(-2.0f*map->map_w,300.0f,-2.0f*map->map_h);
			glEnd();
			}
		glFogi (GL_FOG_MODE, FogMode);
		glFogfv (GL_FOG_COLOR, FogColor);
		glFogf (GL_FOG_DENSITY, FogD);
		glHint (GL_FOG_HINT, GL_NICEST);
		glFogf (GL_FOG_START, FogNear);
		glFogf (GL_FOG_END, refcam.zfar);
		glEnable(GL_CULL_FACE);
		glEnable(GL_LIGHTING);
		glEnable(GL_FOG);
		glCullFace(GL_FRONT);
		refcam.zfar=(500.0f+(cam_h-150.0f)*2.0f)*2.0f;
		refcam.SetView();

		if(cam.RPos.y<=300.0f) {

			if(wireframe)
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);

			map->draw(&refcam,1,false,0.0f,t,dt*time_factor,false,false,false);

			if(wireframe)
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

			if(trees)
				features.draw(&refcam);		// Dessine les éléments "2D" / "sprites"

			need_redraw=true;

			units.draw(&refcam,map,false,true,false,height_line);			// Dessine les unités / draw units

			glDisable(GL_CULL_FACE);
			weapons.draw(&refcam,map);			// Dessine les objets produits par les armes / draw weapons

			fx_manager.draw(&refcam,map,map->sealvl);		// Effets spéciaux en surface / fx above water

			if(particle)
				particle_engine.draw(&refcam,map->map_w,map->map_h,map->bloc_w,map->bloc_h,map->view);	// Dessine les particules

			}

		glDisable(GL_CLIP_PLANE1);

		for(i=0;i<8;i++) {
			glActiveTextureARB(GL_TEXTURE0_ARB+i);
			ReInitTexSys();
			glDisable(GL_TEXTURE_2D);
			}
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glDisable(GL_BLEND);

		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,0,0);

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

		glViewport(0, 0, SCREEN_W, SCREEN_H);

		SetDefOpenGlState();

		glFogi (GL_FOG_MODE, FogMode);
		glFogfv (GL_FOG_COLOR, FogColor);
		glFogf (GL_FOG_DENSITY, FogD);
		glHint (GL_FOG_HINT, GL_NICEST);
		glFogf (GL_FOG_START, FogNear);
		glFogf (GL_FOG_END, cam.zfar);
		}

	cam.SetView();

	sun.Set(cam);
	sun.Enable();

	cam.SetView();
	glRotatef(Atimer*Conv*RAD2DEG,0.0f,1.0f,0.0f);
	glTranslatef(0.0f,60.0f,0.0f);

	cam.zfar*=100.0f;
	cam.SetView();
	glColor4f(1.0f,1.0f,1.0f,1.0f);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,sky);
	glDisable(GL_LIGHTING);
	if(night) {
		glTranslatef(cam.RPos.x,cam.RPos.y,cam.RPos.z);
		sky_obj.draw();
		}
	else {
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f,0.0f);							glVertex3f(-2.0f*map->map_w,300.0f,-2.0f*map->map_h);
			glTexCoord2f(map->map_w*0.002f,0.0f);				glVertex3f(2.0f*map->map_w,300.0f,-2.0f*map->map_h);
			glTexCoord2f(map->map_w*0.002f,map->map_h*0.002f);	glVertex3f(2.0f*map->map_w,300.0f,2.0f*map->map_h);
			glTexCoord2f(0.0f,map->map_h*0.002f);				glVertex3f(-2.0f*map->map_w,300.0f,2.0f*map->map_h);
		glEnd();
		}
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_FOG);
	cam.zfar=500.0f+(cam_h-150.0f)*2.0f;
	cam.SetView();

	if(wireframe)
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);

	map->draw(&cam,1,false,0.0f,t,dt*time_factor);

	if(wireframe)
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

	if(trees)
		features.draw(&cam);		// Dessine les éléments "2D"

	need_redraw=true;

/*----------------------------------------------------------------------------------------------*/

	if(build>=0 && mouse_x>128 && mouse_y>32 && mouse_y<SCREEN_H-32) {			// Affiche le bâtiment à construire
		VECTOR target=cursor_on_map(&cam,map);
		target.x=((int)(target.x)+map->map_w_d)>>3;
		target.z=((int)(target.z)+map->map_h_d)>>3;
		target.x=target.x*8.0f-map->map_w_d;
		target.z=target.z*8.0f-map->map_h_d;
		target.y=max(map->get_unit_h(target.x,target.z),map->sealvl);
		cam.SetView();
		glTranslatef(target.x,target.y,target.z);
		glScalef(unit_manager.unit_type[build].Scale,unit_manager.unit_type[build].Scale,unit_manager.unit_type[build].Scale);
		if(unit_manager.unit_type[build].model) {
			if(can_be_there)
				glColor4f(1.0f,1.0f,1.0f,1.0f);
			else
				glColor4f(1.0f,0.0f,0.0f,0.0f);
			unit_manager.unit_type[build].model->draw(NULL,false,false,false,0,NULL,NULL,NULL,0.0f,NULL,false,players.local_human_id,true,false);
			glColor4f(1.0f,1.0f,1.0f,1.0f);
			}
		cam.SetView();
		glTranslatef(target.x,target.y,target.z);
		float DX = (unit_manager.unit_type[build].FootprintX<<2);
		float DZ = (unit_manager.unit_type[build].FootprintZ<<2);
		float red=1.0f, green=0.0f;
		if(can_be_there) {
			green=1.0f;
			red=0.0f;
			}
		glDisable(GL_CULL_FACE);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glBegin(GL_QUADS);
			glColor4f(red,green,0.0f,1.0f);
			glVertex3f(-DX,0.0f,-DZ);			// First quad
			glVertex3f(DX,0.0f,-DZ);
			glColor4f(red,green,0.0f,0.0f);
			glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);
			glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

			glColor4f(red,green,0.0f,1.0f);
			glVertex3f(-DX,0.0f,-DZ);			// Second quad
			glVertex3f(-DX,0.0f,DZ);
			glColor4f(red,green,0.0f,0.0f);
			glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
			glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

			glColor4f(red,green,0.0f,1.0f);
			glVertex3f(DX,0.0f,-DZ);			// Third quad
			glVertex3f(DX,0.0f,DZ);
			glColor4f(red,green,0.0f,0.0f);
			glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
			glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);

			glColor4f(red,green,0.0f,1.0f);
			glVertex3f(-DX,0.0f,DZ);			// Fourth quad
			glVertex3f(DX,0.0f,DZ);
			glColor4f(red,green,0.0f,0.0f);
			glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
			glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
		glEnd();
		glDisable(GL_BLEND);
		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		}

	units.draw(&cam,map,true,false,true,height_line);			// Dessine les unités sous l'eau / Draw units which are under water

	weapons.draw(&cam,map,true);			// Dessine les objets produits par les armes sous l'eau / Draw weapons which are under water

	if(map->water)
		fx_manager.draw(&cam,map,map->sealvl,true);		// Effets spéciaux sous-marins / Draw fx which are under water

	if(map->water) {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		for(i=0;i<8;i++) {
			glActiveTextureARB(GL_TEXTURE0_ARB+i);
			ReInitTexSys();
			glDisable(GL_TEXTURE_2D);
			}

		if(!g_useProgram || !g_useFBO || water_quality<2) {
			if(water_quality==1) {					// water_quality=1
				glColor4f(1.0f,1.0f,1.0f,0.5f);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glClientActiveTextureARB(GL_TEXTURE0_ARB);

				map->draw(&cam,1,true,map->sealvl,t,dt*time_factor);
				}
			else {									// water_quality=0
				glColor4f(0.3f,0.3f,0.6f,0.75f);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
				glEnable(GL_LIGHTING);

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D,water);
				glClientActiveTextureARB(GL_TEXTURE0_ARB);

				glActiveTextureARB(GL_TEXTURE1_ARB);
				glBindTexture(GL_TEXTURE_2D,sky);
				glEnable(GL_TEXTURE_2D);
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
				glEnable(GL_TEXTURE_GEN_S);
				glEnable(GL_TEXTURE_GEN_T);
				glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE_EXT);
				glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT,GL_INTERPOLATE);

				glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB_EXT,GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB_EXT,GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB_EXT,GL_PREVIOUS_EXT);
				glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND1_RGB_EXT,GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB_EXT,GL_CONSTANT_EXT);
				glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND2_RGB_EXT,GL_SRC_COLOR);
				float col[4]={0.5f,0.5f,0.5f,0.0f};
				glTexEnvfv(GL_TEXTURE_ENV,GL_TEXTURE_ENV_COLOR,col);
				glClientActiveTextureARB(GL_TEXTURE1_ARB);

				cam.SetView();
				glTranslatef(0.0f,map->sealvl,0.0f);
				water_obj.draw(t,cam.RPos.x,cam.RPos.z,false);

				glActiveTextureARB(GL_TEXTURE1_ARB);
				ReInitTexSys();
				glDisable(GL_TEXTURE_2D);
				glClientActiveTextureARB(GL_TEXTURE1_ARB);

				glActiveTextureARB(GL_TEXTURE0_ARB);
				ReInitTexSys();
				glEnable(GL_TEXTURE_2D);
				glClientActiveTextureARB(GL_TEXTURE0_ARB);
				}
			}
		else {
			glColor4f(1.0f,1.0f,1.0f,1.0f);
			glDisable(GL_LIGHTING);

			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,water_FBO);					// First pass of water rendering, store reflection vector
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,first_pass,0);

			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface la texture tampon

			glViewport(0,0,512,512);

			glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,map->lava_map);
			glClientActiveTextureARB(GL_TEXTURE0_ARB);

			water_pass1.on();
			water_pass1.setvar1f("t",t);
			water_pass1.setvar2f("factor",water_obj.w/map->map_w,water_obj.w/map->map_h);

			cam.SetView();
			glTranslatef(0.0f,map->sealvl,0.0f);
			water_obj.draw(t,cam.RPos.x,cam.RPos.z,true);

			water_pass1.off();

			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,0,0);

			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

			glViewport(0, 0, SCREEN_W, SCREEN_H);

			float logw=log((float)SCREEN_W)/log(2.0f);
			float logh=log((float)SCREEN_H)/log(2.0f);
			int wx=logw>(int)logw ? (int)logw+1 : (int)logw;
			int wy=logh>(int)logh ? (int)logh+1 : (int)logh;
			wx=1<<wx;
			wy=1<<wy;
			glBindTexture(GL_TEXTURE_2D,transtex);								// Store what's on screen for transparency effect
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, wx, wy, 0);

			glEnable(GL_STENCIL_TEST);											// Draw basic water in order to have correct texture mapping
			glClear(GL_STENCIL_BUFFER_BIT);
			glStencilFunc(GL_ALWAYS,128, 0xffffffff);
			glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

			glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,water);
			glClientActiveTextureARB(GL_TEXTURE0_ARB);

			cam.SetView();
			glTranslatef(0.0f,map->sealvl,0.0f);
			water_obj.draw(t,cam.RPos.x,cam.RPos.z,true);

			glDisable(GL_STENCIL_TEST);

			glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);

			glEnable(GL_LIGHTING);

			glActiveTextureARB(GL_TEXTURE0_ARB);
			if(water_quality==2 || map->ota_data.lavaworld)
				glBindTexture(GL_TEXTURE_2D,sky);
			else
				glBindTexture(GL_TEXTURE_2D,reflectex);
			glEnable(GL_TEXTURE_2D);

			glActiveTextureARB(GL_TEXTURE1_ARB);
			glBindTexture(GL_TEXTURE_2D,transtex);
			glEnable(GL_TEXTURE_2D);

			glActiveTextureARB(GL_TEXTURE2_ARB);
			glBindTexture(GL_TEXTURE_2D,first_pass);
			glEnable(GL_TEXTURE_2D);

			if(water_quality==2) {
				water_shader.on();
				water_shader.setvar1i("sky",0);
				water_shader.setvar1i("rtex",1);
				water_shader.setvar1i("bump",2);
				water_shader_reflec.setvar2f("coef",(float)SCREEN_W/wx,(float)SCREEN_H/wy);
				}
			else {
				water_shader_reflec.on();
				water_shader_reflec.setvar1i("sky",0);
				water_shader_reflec.setvar1i("rtex",1);
				water_shader_reflec.setvar1i("bump",2);
				water_shader_reflec.setvar2f("coef",(float)SCREEN_W/wx,(float)SCREEN_H/wy);
				}

			glColor4f(1.0f,1.0f,1.0f,1.0f);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			glDisable(GL_DEPTH_TEST);
			cam.SetView();
			glEnable(GL_STENCIL_TEST);
			glStencilFunc(GL_NOTEQUAL,0, 0xffffffff);
			glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
			glBegin(GL_QUADS);
				VECTOR P=cam.RPos+1.1f*(cam.Dir+0.75f*cam.Up-cam.Side);
				glTexCoord2f(0.0f,1.0f);	glVertex3f(P.x,P.y,P.z);
				P=cam.RPos+1.1f*(cam.Dir+0.75f*cam.Up+cam.Side);
				glTexCoord2f(1.0f,1.0f);	glVertex3f(P.x,P.y,P.z);
				P=cam.RPos+1.1f*(cam.Dir-0.75f*cam.Up+cam.Side);
				glTexCoord2f(1.0f,0.0f);	glVertex3f(P.x,P.y,P.z);
				P=cam.RPos+1.1f*(cam.Dir-0.75f*cam.Up-cam.Side);
				glTexCoord2f(0.0f,0.0f);	glVertex3f(P.x,P.y,P.z);
			glEnd();
			glDisable(GL_STENCIL_TEST);
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_BLEND);

			if(water_quality==2)
				water_shader.off();
			else
				water_shader_reflec.off();
			}
		for(i=0;i<8;i++) {
			glActiveTextureARB(GL_TEXTURE0_ARB+i);
			ReInitTexSys();
			glDisable(GL_TEXTURE_2D);
			}
		glActiveTextureARB(GL_TEXTURE0_ARB);
		}

	units.draw(&cam,map,false,false,true,height_line);			// Dessine les unités non encore dessinées / Draw units which have not been drawn

	weapons.draw(&cam,map,false);			// Dessine les objets produits par les armes n'ayant pas été dessinés / Draw weapons which have not been drawn

	if(selected && key[KEY_LSHIFT]) {
		cam.SetView();
		bool builders = false;
		for(uint16 e=0;e<units.index_list_size;e++) {
			i = units.idx_list[e];
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel) {
				builders |= unit_manager.unit_type[units.unit[i].type_id].Builder;
				units.unit[i].show_orders();					// Dessine les ordres reçus par l'unité
				}
			}

		if(builders)
			for(uint16 e=0;e<units.index_list_size;e++) {
				i = units.idx_list[e];
				if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && !units.unit[i].sel
				&& unit_manager.unit_type[units.unit[i].type_id].Builder && (unit_manager.unit_type[units.unit[i].type_id].TEDclass&CLASS_PLANT)==0)
					units.unit[i].show_orders(true);					// Dessine les ordres reçus par l'unité
				}
		}

	if(shadow && cam.RPos.y<=300.0f)
		if(shadow_quality<=1) {
			if(rotate_light) {
				sun.Dir.x=-1.0f;
				sun.Dir.y=1.0f;
				sun.Dir.z=1.0f;
				sun.Dir.Unit();
				VECTOR Dir=-sun.Dir;
				Dir.x=cos(light_angle);
				Dir.z=sin(light_angle);
				Dir.Unit();
				sun.Dir=-Dir;
				units.draw_shadow(&cam,Dir,map);
				}
			else {
				sun.Dir.x=-1.0f;
				sun.Dir.y=1.0f;
				sun.Dir.z=1.0f;
				sun.Dir.Unit();
				units.draw_shadow(&cam,-sun.Dir,map);
				}
			}
		else {
			float alpha=1.0f-exp((1.0f/shadow_quality)*log(0.5f));
			VECTOR Dir;
			if(rotate_light) {
				sun.Dir.x=-1.0f;
				sun.Dir.y=1.0f;
				sun.Dir.z=1.0f;
				sun.Dir.Unit();
				Dir=-sun.Dir;
				Dir.x=cos(light_angle);
				Dir.z=sin(light_angle);
				Dir.Unit();
				sun.Dir=-Dir;
				}
			else {
				sun.Dir.x=-1.0f;
				sun.Dir.y=1.0f;
				sun.Dir.z=1.0f;
				sun.Dir.Unit();
				Dir=-sun.Dir;
				}
			for(int i=0;i<shadow_quality;i++) {
				VECTOR RDir;
				RDir=Dir;
				RDir.x+=cos(i*PI*2.0f/shadow_quality)*shadow_r;
				RDir.z+=sin(i*PI*2.0f/shadow_quality)*shadow_r;
				RDir.Unit();
				units.draw_shadow(&cam,RDir,map,alpha);
				}
			}

	if(!map->water)
		fx_manager.draw(&cam,map,map->sealvl,true);		// Effets spéciaux en surface
	fx_manager.draw(&cam,map,map->sealvl);		// Effets spéciaux en surface

	if(particle)
		particle_engine.draw(&cam,map->map_w,map->map_h,map->bloc_w,map->bloc_h,map->view);	// Dessine les particules

	if(key[KEY_ESC]) done=true;
	if(key[KEY_TILDE]) {
		if(!tilde)
			Console->ToggleShow();
		tilde=true;
		}
	else
		tilde=false;

	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	for(i=7;i>=0;i--) {
		glActiveTextureARB(GL_TEXTURE0_ARB+i);
		ReInitTexSys();
		glDisable(GL_TEXTURE_2D);
		}
	allegro_gl_set_allegro_mode();		// Affiche console, infos,...

	old_cam_pos=cam.RPos;
	int signal=game_script.run(map,timetosimulate,&cam,players.local_human_id);

	if(old_cam_pos.x!=cam.RPos.x || old_cam_pos.y!=cam.RPos.y || old_cam_pos.z!=cam.RPos.z)
		need_redraw=true;

	switch(signal)
	{
	case 0:				// Rien de spécial
		break;
	case -1:			// Fin du script
		game_script.stop();
		break;
	case -2:			// Pause
		break;
	case -3:			// Attente d'un évènement
		break;
	case 1:				// Fin de partie (match nul)
		done=true;
		exit_mode=EXIT_NONE;
		break;
	case 2:				// Fin de partie (victoire)
		done=true;
		exit_mode=EXIT_VICTORY;
		break;
	case 3:				// Fin de partie (défaite)
		done=true;
		exit_mode=EXIT_DEFEAT;
		break;
	case 4:				// Caméra en mode normal
		if(freecam) {
			freecam=false;
			r2=0.0f;
			r1=or1;
			need_redraw=true;
			}
		break;
	case 5:				// Caméra libre
		if(!freecam) {
			freecam=true;
			or1=r1;
			need_redraw=true;
			}
		break;
	};

	if(selecting) {						// Affiche le rectangle de selection
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_LINE_LOOP);
			glColor4f(0.0f,0.0f,0.0f,1.0f);
			glVertex2i(sel_x[0]+1,sel_y[0]+1);
			glVertex2i(sel_x[1]+1,sel_y[0]+1);
			glVertex2i(sel_x[1]+1,sel_y[1]+1);
			glVertex2i(sel_x[0]+1,sel_y[1]+1);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glColor4f(1.0f,1.0f,1.0f,1.0f);
			glVertex2i(sel_x[0],sel_y[0]);
			glVertex2i(sel_x[1],sel_y[0]);
			glVertex2i(sel_x[1],sel_y[1]);
			glVertex2i(sel_x[0],sel_y[1]);
		glEnd();
		}

	if(cur_sel_index>=0 && cur_sel_index<=units.last_index && units.unit[cur_sel_index].flags==0) {
		cur_sel=-1;
		cur_sel_index=-1;
		current_order=SIGNAL_ORDER_NONE;
		}

	int n=cur_sel;
	if(n>=0 && units.unit[cur_sel_index].port[BUILD_PERCENT_LEFT]>0.0f)		// Unité non terminée
		n=-1;
	int sel=unit_manager.unit_build_menu(n,omb2);		// Menu correspondant à l'unité / Unit's menu
	if(sel==-2) {										// Crée des armes / build weapons
		if(mouse_b==1 && omb2!=1) {
			if(key[KEY_LSHIFT])
				units.unit[cur_sel_index].planned_weapons+=5.0f;
			else
				units.unit[cur_sel_index].planned_weapons+=1.0f;
			}
		else if(mouse_b==2 && omb2!=2) {
			if(key[KEY_LSHIFT])
				units.unit[cur_sel_index].planned_weapons-=5.0f;
			else
				units.unit[cur_sel_index].planned_weapons-=1.0f;
			if(units.unit[cur_sel_index].planned_weapons<0.0f)
				units.unit[cur_sel_index].planned_weapons=0.0f;
			}
		sel=-1;
		}
	units.complete_menu(cur_sel_index);
	int signal_order=units.menu_order(omb2,players.local_human_id);
	players.show_resources();

	switch(signal_order)
	{
	case SIGNAL_ORDER_MOVE:
	case SIGNAL_ORDER_PATROL:
	case SIGNAL_ORDER_GUARD:
	case SIGNAL_ORDER_ATTACK:
	case SIGNAL_ORDER_RECLAM:
	case SIGNAL_ORDER_LOAD:
	case SIGNAL_ORDER_UNLOAD:
		current_order=signal_order;
		break;
	case SIGNAL_ORDER_STOP:
	case SIGNAL_ORDER_ONOFF:
		current_order=SIGNAL_ORDER_NONE;
		break;
	};
	if(sel>=0 || (mouse_b&2))
		current_order=SIGNAL_ORDER_NONE;

	if(sel>=0 && mouse_b==2 && omb2!=2) {
		MISSION *cur=units.unit[cur_sel_index].mission;
		MISSION **old=&(units.unit[cur_sel_index].mission);
		int nb=1;
		if(key[KEY_LSHIFT]) nb=5;
		if(cur)	{	old=&(cur->next);	cur=cur->next; }
		if(cur)
			while(cur) {
				if((cur->mission==MISSION_BUILD && cur->data==sel) || cur->mission==MISSION_STOP) {			// Efface un ordre
					if(cur->mission==MISSION_BUILD)
						nb--;
					*old=cur->next;
					free(cur);
					cur=*old;
					if(nb==0)
						break;
					continue;
					}
				old=&(cur->next);
				cur=cur->next;
				}
		cur=units.unit[cur_sel_index].mission;
		if(nb>0 && cur!=NULL && cur->mission==MISSION_BUILD_2 && cur->data==sel) {
			units.kill(((UNIT*)(units.unit[cur_sel_index].mission->p))->idx,map);
			units.unit[cur_sel_index].next_mission();
			}
		}
	if(sel>=0 && mouse_b==1 && omb2!=1) {
		build=sel;
		if(unit_manager.unit_type[cur_sel].TEDclass==CLASS_PLANT) {			// S'il s'agit d'un bâtiment
			if(key[KEY_LSHIFT])
				for(i=0;i<5;i++)
					units.give_order_build(players.local_human_id,build,units.unit[cur_sel_index].Pos,false);
			else
				units.give_order_build(players.local_human_id,build,units.unit[cur_sel_index].Pos,false);
			build=-1;
			}
		}

	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	PutTex(freecam ? freecam_on : freecam_off,32,SCREEN_H-64,95,SCREEN_H);
	glDisable(GL_BLEND);

	if(mouse_x>=32 && mouse_x<=95 && mouse_y>=SCREEN_H-64 && mouse_b==1 && omb2==0) {
		freecam^=true;
		if(!freecam)	r2=0.0f;
		if(freecam)		or1=r1;
		else
			r1=or1;
		}

	omb2=mouse_b;

	map->draw_mini(0,0,128,128,&cam);	// Mini-carte
	units.draw_mini(map->map_w,map->map_h,map->mini_w,map->mini_h,map->map_data);
	weapons.draw_mini(map->map_w,map->map_h,map->mini_w,map->mini_h);

	if(view_dbg) {
		glDisable(GL_TEXTURE_2D);
		glColor3f(1.0f,1.0f,1.0f);
		glBegin(GL_POINTS);
		float rw=128.0f*map->mini_w/252/map->bloc_w;
		float rh=128.0f*map->mini_h/252/map->bloc_h;
		float dw=64.0f-0.5f*map->bloc_w*rw;
		float dh=64.0f-0.5f*map->bloc_h*rh;
		for(int y=0;y<map->bloc_h;y++)
			for(int x=0;x<map->bloc_w;x++)
				if(map->view[y][x])
					glVertex2f(x*rw+dw,y*rh+dh);
		glEnd();
		glEnable(GL_TEXTURE_2D);
		}

	if(ia_debug)
		for(int i=0;i<players.nb_player;i++)
			if(players.control[i]==PLAYER_CONTROL_LOCAL_AI)
				players.ai_command[i].strategic_data.show_zones(map->map_w,map->map_h,map->mini_w,map->mini_h);

	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
	glEnable(GL_BLEND);
	if(show_script)
	if(cur_sel>=0 && unit_manager.unit_type[cur_sel].script) {
		float Y=32.0f;
		allegro_gl_printf(aglfont,128.0f,Y,0.0f,0xFFFFFF,"%d scripts",unit_manager.unit_type[cur_sel].script->nb_script);	Y+=9.0f;
		for(i=0;i<unit_manager.unit_type[cur_sel].script->nb_script;i++) {
			if(units.unit[cur_sel_index].is_running(i))
				allegro_gl_printf(aglfont,128.0f,Y,0.0f,0xFFFFFF,"%d %s (on)",i,unit_manager.unit_type[cur_sel].script->name[i]);
			else
				allegro_gl_printf(aglfont,128.0f,Y,0.0f,0xFFFFFF,"%d %s (off)",i,unit_manager.unit_type[cur_sel].script->name[i]);
			Y+=9.0f;
			}
		}

	if(show_model && cur_sel>=0 && unit_manager.unit_type[cur_sel].model)
		unit_manager.unit_type[cur_sel].model->print_struct(32.0f,128.0f,aglfont);

	if(internal_name && cur_sel>=0 && unit_manager.unit_type[cur_sel].Unitname!=NULL)
		allegro_gl_printf(aglfont,128.0f,32.0f,0.0f,0xFFFFFF,"nom interne %s",unit_manager.unit_type[cur_sel].Unitname);

	if(unit_info>0.0f && unit_info_id>=0)
		unit_manager.unit_type[unit_info_id].show_info(unit_info,ta_glfont);

/*	if(n>=0 && unit_manager.unit_type[cur_sel].weapon[0]!=NULL && unit_manager.unit_type[cur_sel].weapon[0]->name!=NULL)
		allegro_gl_printf(aglfont,128.0f,32.0f,0.0f,0xFFFFFF,"arme ok");
	else if(n>=0 && unit_manager.unit_type[cur_sel].weapon[0]!=NULL)
		allegro_gl_printf(aglfont,128.0f,32.0f,0.0f,0xFFFFFF,"arme sans nom");
	else
		allegro_gl_printf(aglfont,128.0f,32.0f,0.0f,0xFFFFFF,"pas d'arme");*/

/*	if(selected && cur_sel!=-1) {					// Sur les unités sélectionnées
		char *unit_info[]={"MISSION_STANDBY","MISSION_VTOL_STANDBY","MISSION_GUARD_NOMOVE","MISSION_MOVE","MISSION_BUILD","MISSION_BUILD_2","MISSION_STOP","MISSION_REPAIR","MISSION_ATTACK",
							"MISSION_AIM","MISSION_SHOOT"};
		for(i=0;i<=units.last_index;i++)
			if(units.unit[i].flags!=0 && units.unit[i].owner_id==0 && units.unit[i].sel) {
				if(units.unit[i].mission!=NULL && units.unit[i].mission->mission>=0 && units.unit[i].mission->mission<=10)
					printf("MISSION: %s\n",unit_info[units.unit[i].mission->mission]);
				else
					printf("MISSION: NONE\n");
				}
		}*/

	glDisable(GL_BLEND);

	char *cmd=NULL;
	if(!shoot || video_shoot)
		cmd=Console->draw(ta_glfont,dt,text_height(ta_alfont));			// Affiche la console

	float Y=0.0f;

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

	if(showfps) {
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		allegro_gl_printf(ta_glfont,0.0f,Y,0.0f,0xFFFFFF,"fps: %d",fps);
		Y+=9.0f;
		}

	glDisable(GL_BLEND);

	if(mouse_b!=4)
		draw_cursor();

	if(shoot) {
		BITMAP *shoot_bmp=create_bitmap_ex(24,SCREEN_W,SCREEN_H);
		blit(screen,shoot_bmp,0,0,0,0,SCREEN_W,SCREEN_H);
		char nom[100];
		nom[0]=0;
		strcat(nom,"shoot0000");
		nom[strlen(nom)-4]+=(nb_shoot/1000)%10;
		nom[strlen(nom)-3]+=(nb_shoot/100)%10;
		nom[strlen(nom)-2]+=(nb_shoot/10)%10;
		nom[strlen(nom)-1]+=nb_shoot%10;
		nb_shoot=++nb_shoot%10000;
		strcat(nom,".jpg");
		save_jpg_ex(nom,shoot_bmp,NULL,85,JPG_SAMPLING_411,NULL);
		destroy_bitmap(shoot_bmp);
		shoot=false;
		}

	allegro_gl_unset_allegro_mode();

	allegro_gl_flip();

	if(cmd) {		// Analyse les commandes tapées dans la console
		if(strstr(cmd,"fps_on")) showfps=true;				// Affiche le nombre d'images/seconde
		if(strstr(cmd,"fps_off")) showfps=false;				// Cache le nombre d'images/seconde
		if(strstr(cmd,"zshoot")) {									// Prend une capture d'écran(tampon de profondeur seulement)
			BITMAP *bmp=create_bitmap_ex(32,SCREEN_W,SCREEN_H);
			clear(bmp);
			glReadPixels(0,0,SCREEN_W,SCREEN_H,GL_DEPTH_COMPONENT,GL_INT,bmp->line[0]);
			save_bitmap("z.tga",bmp,NULL);
			destroy_bitmap(bmp);
			}
		else if(strstr(cmd,"video shoot")) video_shoot^=true;		// Capture video
		else if(strstr(cmd,"shoot")) shoot=true;					// Prend une capture d'écran
#if defined USE_MP3 || defined USE_OGG
		if(strstr(cmd,"change state"))
			music_manager.change_state(music_manager.game_state^1);
#endif
		if(strstr(cmd,"reload shaders") && g_useProgram) {
			water_shader.destroy();
			water_shader.load("shaders/water.frag","shaders/water.vert");
			water_shader_reflec.destroy();
			water_shader_reflec.load("shaders/water_reflec.frag","shaders/water.vert");
			water_pass1.destroy();
			water_pass1.load("shaders/water_pass1.frag","shaders/water_pass1.vert");
			}
		if(strstr(cmd,"view debug"))	view_dbg^=true;
		if(strstr(cmd,"ia debug"))	ia_debug^=true;
		if(strstr(cmd,"internal name")) internal_name^=true;		// Affiche/Cache le nom interne de l'unité
		if(strstr(cmd,"exit")) done=true;					// Quitte le programme
		if(strstr(cmd,"wireframe")) {	wireframe^=true;	need_redraw=true;	}
		if(strstr(cmd,"priority")) priority_level=atoi(strstr(cmd,"priority")+9);
		if(strstr(cmd,"water quality")) water_quality=atoi(strstr(cmd,"water quality")+14)%4;
		if(strstr(cmd,"shadow quality")) shadow_quality=atoi(strstr(cmd,"shadow quality")+15);
		else if(strstr(cmd,"shadow ray")) shadow_r=atof(strstr(cmd,"shadow ray")+11);
		else if(strstr(cmd,"shadow")) shadow^=true;
		if(strstr(cmd,"particle")) particle^=true;
		if(strstr(cmd,"trees")) trees^=true;
		if(strstr(cmd,"show script")) show_script^=true;
		if(strstr(cmd,"show model")) show_model^=true;
		if(strstr(cmd,"rotate light")) rotate_light^=true;
		if(strstr(cmd,"freecam")) {
			freecam^=true;
			if(!freecam)	r2=0.0f;
			if(freecam)		or1=r1;
			else
				r1=or1;
			}
		if(strstr(cmd,"fps_limit ")) {
			speed_limit=atoi(strstr(cmd,"fps_limit ")+10);
			if(speed_limit==0.0f) speed_limit=-1.0f;
			delay=1.0f/speed_limit;
			}
		if(strstr(cmd,"spawn ")) {
			char *cur_pos = cmd+6;
			int unit_type = -1;
			int player_id = players.local_human_id;
			int nb_to_spawn=1;
			char *f = cur_pos;
			char mem = 0;
			while(f[0]!=0 && f[0]!=32)	f++;
			mem = f[0];			f[0] = 0;
			unit_type = unit_manager.get_unit_index(cur_pos);
			f[0] = mem;
			if(mem) {
				f++;
				cur_pos=f;
				while(f[0]!=0 && f[0]!=32)	f++;
				mem = f[0];			f[0] = 0;
				player_id = atoi(cur_pos);
				f[0] = mem;
				if(mem) {
					f++;
					cur_pos=f;
					nb_to_spawn=atoi(cur_pos);
					}
				}
			for(i=0;i<nb_to_spawn;i++) {
				int id=0;
				if(unit_type==-1)
					id = units.create(rand()%unit_manager.nb_unit,player_id);
				else
					id = units.create(unit_type,player_id);
				int e=0;
				do {
					units.unit[id].Pos.x=(rand()%map->map_w)-map->map_w_d;
					units.unit[id].Pos.z=(rand()%map->map_h)-map->map_h_d;
					e++;
				}while(e<100 && !can_be_built(units.unit[id].Pos,map,units.unit[id].type_id));
				if(!can_be_built(units.unit[id].Pos,map,units.unit[id].type_id)) {
					units.unit[id].flags=0;
					units.kill(id,map);
					}
				}
			}
		if(strstr(cmd,"timefactor "))
			time_factor=atof(strstr(cmd,"timefactor ")+11);
		if(strstr(cmd,"addhp "))
			if(selected) {					// Sur les unités sélectionnées
				int value=atoi(strstr(cmd,"addhp ")+6);
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel) {
						units.unit[i].hp+=value;
						if(units.unit[i].hp<0)
							units.unit[i].hp=0;
						else if(units.unit[i].hp>unit_manager.unit_type[units.unit[i].type_id].MaxDamage)
							units.unit[i].hp=unit_manager.unit_type[units.unit[i].type_id].MaxDamage;
						}
				}
		if(strstr(cmd,"reset_script"))							// Réinitialise les scripts
			if(selected) {					// Sur les unités sélectionnées
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
						units.unit[i].reset_script();
				}
		if(strstr(cmd,"unitinfo"))
			if(selected && cur_sel!=-1) {					// Sur les unités sélectionnées
				char *unit_info[]={"ACTIVATION","STANDINGMOVEORDERS","STANDINGFIREORDERS","HEALTH","INBUILDSTANCE","BUSY","PIECE_XZ","PIECE_Y","UNIT_XZ","UNIT_Y","UNIT_HEIGHT","XZ_ATAN","XZ_HYPOT","ATAN",
								   "HYPOT","GROUND_HEIGHT","BUILD_PERCENT_LEFT","YARD_OPEN","BUGGER_OFF","ARMORED"};
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
						for(int e=1;e<21;e++)
							printf("%s=%f\n",unit_info[e-1],units.unit[i].port[e]);
				}
		if(strstr(cmd,"kill"))							// Détruit les unités sélectionnées
			if(selected) {					// Sur les unités sélectionnées
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
						units.kill(i,map);
				selected=false;
				cur_sel=-1;
				}
		if(strstr(cmd,"script "))							// Force l'éxecution d'un script
			if(selected) {					// Sur les unités sélectionnées
				int id=atoi(strstr(cmd,"script ")+7);
				for(i=0;i<=units.last_index;i++)
					if(units.unit[i].flags!=0 && units.unit[i].owner_id==players.local_human_id && units.unit[i].sel)
						units.unit[i].launch_script(id);
				}

		if(strstr(cmd,"metal")) cheat_metal^=true;				// cheat codes
		if(strstr(cmd,"energy")) cheat_energy^=true;			// cheat codes
		free(cmd);
		}
	if(cheat_metal) players.metal[players.local_human_id]=players.metal_s[players.local_human_id];					// cheat codes
	if(cheat_energy) players.energy[players.local_human_id]=players.energy_s[players.local_human_id];				// cheat codes
	if(key[KEY_F12])
		shoot=true;

/*------------ Code de gestion du déroulement de la partie -----------------------------------*/

	if(signal==-1) {			// Si le script est terminé, on reprend les règles standard
		bool win=true;
		for(i=0;i<players.nb_player;i++)
			if(!players.annihilated[i] && i!=players.local_human_id)	{
				win=false;
				break;
				}
		if(win) {						// Victoire
			done=true;
			exit_mode=EXIT_VICTORY;
			}
		if(players.annihilated[players.local_human_id]) {	// Défaite
			done=true;
			exit_mode=EXIT_DEFEAT;
			}
		}

}while(!done);

config_manager.write_config_file();		// Mémorise la configuration

#if defined USE_MP3 || defined USE_OGG
music_manager.save_play_list();
music_manager.destroy();
#endif

water_obj.destroy();
sky_obj.destroy();

if(g_useProgram && g_useFBO && map->water) {
	glDeleteFramebuffersEXT(1,&water_FBO);
	water_pass1.destroy();
	water_shader.destroy();
	water_shader_reflec.destroy();
	glDeleteRenderbuffersEXT(1,&zbuf);
	glDeleteTextures(1,&first_pass);
	glDeleteTextures(1,&reflectex);
	glDeleteTextures(1,&transtex);
	}

glDeleteTextures(1,&freecam_on);
glDeleteTextures(1,&freecam_off);
glDeleteTextures(1,&sky);
glDeleteTextures(1,&water);
glDeleteTextures(1,&glow);

printf("nombre de models chargés: %d \n",model_manager.nb_models);
printf("nombre d'unités chargées: %d \n",unit_manager.nb_unit);
printf("nombre de textures chargées: %d \n",texture_manager.nbtex);

switch(exit_mode)
{
case EXIT_NONE:
	break;
case EXIT_VICTORY:
	break;
case EXIT_DEFEAT:
	break;
};

Console->AddEntry("freeing memory used for the map");
map->destroy();
delete map;
Console->AddEntry("freeing memory used for 3d models");
model_manager.destroy();
Console->AddEntry("freeing memory used for units");
unit_manager.destroy();
Console->AddEntry("freeing memory used for particle engine");
particle_engine.destroy();
Console->AddEntry("freeing memory used for ingame units");
units.destroy();
Console->AddEntry("freeing memory used for weapons");
weapons.destroy();
Console->AddEntry("freeing memory used for sound");
sound_manager.destroy();
Console->AddEntry("freeing memory used for fx");
fx_manager.destroy();

Console->AddEntry("freeing memory used for weapon manager");
weapon_manager.destroy();

Console->AddEntry("freeing memory used for features");
feature_manager.destroy();

Console->AddEntry("freeing memory used for ingame features");
features.destroy();

Console->AddEntry("freeing memory used for textures");
texture_manager.destroy();

stats_menu();

Console->AddEntry("freeing memory used for players");
players.destroy();

}

int anim_cursor(int type)
{
	if(type==-1)
		return (Atimer-start>>17)%cursor.anm[cursor_type].nb_bmp;
	else
		return (Atimer-start>>17)%cursor.anm[type].nb_bmp;
}

void draw_cursor()
{
	int curseur=anim_cursor();
	if(curseur<0 || curseur>=cursor.anm[cursor_type].nb_bmp) {
		curseur=0;
		start=Atimer;
		}
	int dx=cursor.anm[cursor_type].ofs_x[curseur];
	int dy=cursor.anm[cursor_type].ofs_y[curseur];
	int sx=cursor.anm[cursor_type].bmp[curseur]->w-1;
	int sy=cursor.anm[cursor_type].bmp[curseur]->h-1;
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	PutTex(cursor.anm[cursor_type].glbmp[curseur],mouse_x-dx,mouse_y-dy,mouse_x-dx+sx,mouse_y-dy+sy);
	glDisable(GL_BLEND);
}

VECTOR cursor_on_map(CAMERA *cam,MAP *map)
{
	VECTOR cur_dir;
	cur_dir=cam->Dir+2.0f*(mouse_x-SCREEN_W_HALF)*SCREEN_W_INV*cam->Side-1.5f*(mouse_y-SCREEN_H_HALF)*SCREEN_H_INV*cam->Up;
	cur_dir.Unit();		// Direction pointée par le curseur
	return map->hit(cam->Pos,cur_dir);
}
